/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

/**
 * This file contains analysis of ATGC.
 */

#ifndef LERI_ATGC_H_
#define LERI_ATGC_H_

#include <cstdint>
#include "leri/common.h"
#include "leri/gio/javas.h"
#include "leri/atgc/atgc_constant.h"
// #include "leri/omics/kseq.h"
// #include <sys/types.h>
// #include <sys/stat.h>
// #include <fcntl.h>
// #include <map>
// #include <boost/assign/list_of.hpp>

#include <zlib.h>
// #include <signal.h>
// #include <gzstream.h>
// #include "leri/reads.h"
// #include "hdf5.h"
#include "leri/util/malloc.h"
// #include "leri/util/matrix.h"
// #include "leri/gio/json.h"
#include "leri/gio/figure.h"

namespace leri {

template <typename Dtype>
class ATGC {
public:
  explicit ATGC(string filename, bool hasQuality = true, bool phred64 = false);
  // virtual ~ATGC(){}
  ATGC(bool hasQuality = true, bool phred64 = false) {}
  ATGC(string output_dirt = "", string job_name = "")
  :output_dirt_(output_dirt),
   job_name_(job_name) {
    base_name_ = output_dirt_ + job_name_;
  }
  ~ATGC() {}




// public:// for synteny block
//   explicit ATGC(string & fileName): 
//       inputStream_(fileName.c_str()), fileName_(fileName) {} 
//   size_t  GetSequences(std::vector<FASTARecord>& record);
//   bool  IsOk() const  {
//     return inputStream_.good();
//   }
// private:// for synteny block
//   void ValidateSequence(std::string & sequence);
//   void ValidateHeader(std::string & header);

//   std::ifstream inputStream_;
//   std::string fileName_;



public:
  vector<string> reads;
  // Read and Base
  unsigned long int read_total_;
  unsigned long int base_total_;
  unsigned long int base_A, base_T, base_C, base_G, base_N;
  vector<vector<Dtype> > qual_per_base;
  vector<vector<int> > base_in_seqs;
  Dtype Nppm;
        
  // Length
  int len_min_;
  int len_max_;
  Dtype len_avg;
        
  // Qual stats
  int phred;
  unsigned long int min_phred;
  unsigned long int max_phred;
  vector<unsigned long> qual_vector;
  unsigned long int q20, q30;
  void init();
 
  void read_fastx(string file);
  void read_gzfile(vector<vector<char> >& lines, string file);

  void read_trim(string file, string ofile = "", int trim_quality = 13);
  void read_filter(string file, string ofile = "", int m = 0);
  void read_quality(string file, string ofile = "", int m = 0); // Transform Phred-scaled quality scores in a fastQ file
  void kmer_counter(string file);


public: // correction of errors in reads
  int64_t num_reads_;
  uint64_t bases_total_;
  uint64_t genome_len_;
  uint8_t* read_bin_; // Array to store read information in binary.
  uint8_t* counters_; // Array to store the counters for each witness.
  uint64_t* witness_; // Array to store the witnesses.
  uint64_t mem_current_ = 0;
  uint64_t mem_peak_    = 0;
  int* reads_good_;
  uint64_t* read_location_;
  uint64_t chunk_size_ = 0;
  uint64_t reads_size_ = 0;
  uint8_t* reverse_complements_;
  int64_t threshold_;


  /**
   * Read the input FASTA/FASTQ file and store in binary.
   */
  void error_correct(string fastx_file);
private:
  // void read_fastx(string fastx_file);
  Dtype calc_expected_number(int64_t witness_len, int64_t read_len);
  Dtype calc_incorrect_number(int64_t witness_len, int64_t read_len);
  void calc_reverse_complement();
  void calc_threshold(int64_t read_len, int64_t witness_len);
  void calc_witness_len(int64_t read_len, int64_t &witLength_small, int64_t &witLength_large);
  void read_fastx();
  void calc_witness_counter(int64_t* loc, 
       uint64_t witness_len_in_bits, int &hash_table_idx, uint64_t &hash_table_size);
  void correct_errors(int64_t* loc, uint64_t witness_len_in_bits, uint64_t &num_changed_pos, 
       uint64_t &hash_table_size, int &hash_table_idx);
  void write_fastx(string fastx_file);
public:
  void finish();
public:
  void stats_report();
private:
  void read_num_update(int n) {read_total_ += n;}
  void base_num_update(int n) {base_total_ += n;}
  void base_typ_counter(string seq);
  void quality_converter(string quality);
  void guess_phred();
  void quality_stats();
  void stats_print();
private:
  string output_dirt_;
  string job_name_; 
  string base_name_;
  string fastx_name_;
  char format_;

  Dtype** qual_per_base_;

  // string mFilename;
  // gzFile mZipFile;
  // ifstream mFile;
  // bool mZipped;
  // bool mHasQuality;
  // bool mPhred64;
  // Read<Dtype> reads_;
  int num_alw; // Number of IUPAC ambiguity characters (N)
  const static int min_read_len = 25;
  // const static int trim_quality    = 13;

protected:
  DISABLE_COPY_AND_ASSIGN(ATGC);
}; // class ATGC


} // namespace leri
#endif // LERI_ATGC_H_

