/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */


#ifndef LERI_COMMON_H_
#define LERI_COMMON_H_

// #include <limits>
#include <climits>
#include <cmath>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <utility>  // pair
#include <vector>
#include <cstddef>  // NULL
#include <algorithm>
#include <regex>
#include <random>
#include <stdlib.h> /* system, malloc， NULL, EXIT_FAILURE */
#include <chrono>
// Boost Library
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
// Google
// #include <gtest/gtest.h>
#include <gflags/gflags.h>
#include <glog/logging.h>
// TensorFlow
// #include <tensorflow/c/c_api.h>
// #include <tensorflow/core/platform/env.h>
// #include <tensorflow/core/public/session.h>

// #ifdef USE_EIGEN
// #include <eigen3/Eigen/Eigen>
// // #include <eigen3/Eigen/SVD> // PCA new
// #include <eigen3/Eigen/Sparse>
// #include <eigen3/Eigen/Eigenvalues>
// #endif // USE_EIGEN

// Intel
// TBB : TODO: delete or use it for future
#ifdef USE_TBB
#include <tbb/tbb.h>
#include <tbb/parallel_for.h>
#include <tbb/task_scheduler_init.h>
#include <tbb/blocked_range.h>
#endif // USE_TBB

#ifdef USE_OPENCV
#include <opencv2/opencv.hpp>
// #include <opencv/opencv.hpp>
#endif // USE_OPENCV

#ifdef USE_OPENMP
#include <omp.h>
#endif

#ifdef USE_GPU
#include <cuda.h>
#include <cuda_runtime.h>
#include <curand.h>
#include <driver_types.h>
#endif // USE_GPU

#include "leri/util/device_alternate.h"
#include "leri/util/timer.h"
#include "leri/util/random.h"

#define VERSION AS_STRING(LERI_VERSION)
#define VERS_INFO(str) str "LESRIA (Leri version " VERSION ") \n"
// #define HTTP_LERI "https://godzilla.uchicago.edu/pages/"
#define LHTTPS "https://kornmann.bioch.ox.ac.uk"
#define EMAIL "yaan.jang@gmail.com"

// Color printing
#define BLACK  -1
#define RED     1
#define GREEN   2
#define YELLOW  3
#define BLUE    4
#define MAGENTA 5
#define CYAN    6
#define WHITE   7

#ifndef USE_GPU
#define NO_GPU printf("-- \x1B[31m%s\x1B[0m\n", "I cannot use GPU!")
#endif

#ifndef DBL_EPSILON
#define DBL_EPSILON 2.2204460492503131e-16 
#endif
#ifndef DBL_MIN
#define DBL_MIN 2.2250738585072014e-308
#endif
#ifndef DBL_MAX
#define DBL_MAX 1.7976931348623157e308
#endif
// #define FLT_EPSILON 1.19209290e-7
// Minimum difference between two doubles to be considered different numbers.
static const double MIN_FLOAT_DIFF = 2.2204460492503131e-16; // DBL_EPSILON;
// static const double M_PI

// 180 / PI
#define RADIANS_TO_DEGREES 57.295779513082320876798154814105
// PI / 180
#define DEGREES_TO_RADIANS 0.017453292519943295769236907684886

#define N_COMMON_AA 21 // Number of ommon amino acids (including a gap)
// Sequence potential (TODO remove)
#define N_ALPHA 21
#define N_ALPHA_PAD 21

#define DEBUG 0

// Omics data
#define _A 0
#define _C 1
#define _G 2
#define _T 3
#define _N 4

// Convert macro to string
#define STRINGIFY(m) #m
#define AS_STRING(m) STRINGIFY(m)

#ifdef __APPLE__
#include <libkern/OSByteOrder.h>
#define htobe16(x) OSSwapHostToBigInt16(x)
#define htole16(x) OSSwapHostToLittleInt16(x)
#define be16toh(x) OSSwapBigToHostInt16(x)
#define le16toh(x) OSSwapLittleToHostInt16(x)

#define htobe32(x) OSSwapHostToBigInt32(x)
#define htole32(x) OSSwapHostToLittleInt32(x)
#define be32toh(x) OSSwapBigToHostInt32(x)
#define le32toh(x) OSSwapLittleToHostInt32(x)

#define htobe64(x) OSSwapHostToBigInt64(x)
#define htole64(x) OSSwapHostToLittleInt64(x)
#define be64toh(x) OSSwapBigToHostInt64(x)
#define le64toh(x) OSSwapLittleToHostInt64(x)
#endif


static time_t now = time(0);
static tm* ctm = localtime(&now);

// Disable the copy and assignment operator for a class.
#define DISABLE_COPY_AND_ASSIGN(classname) \
private:\
  classname(const classname&);\
  classname& operator=(const classname&)

// Instantiate a class with float and double specifications.
#define INSTANTIATE_CLASS(classname) \
  char gInstantiationGuard##classname; \
  template class classname<float>; \
  template class classname<double>

// Instantiate a gpu function with float and double specifications.
/*

// LBFGS optimizer
#define INSTANTIATE_LBFGS_GPU(classname) \
  template LBFGS<float>::LBFGS(); \
  template LBFGS<double>::LBFGS(); \
  template LBFGS<float>::~LBFGS(); \
  template LBFGS<double>::~LBFGS(); \
  template void LBFGS<float>::dispatch_dot(const size_t n, float *dst, const float *d_x, const float *d_y, bool dstDevicePointer) const; \
  template void LBFGS<double>::dispatch_dot(const size_t n, double *dst, const double *d_x, const double *d_y, bool dstDevicePointer) const; \
  template void LBFGS<float>::dispatch_scale(const size_t n, float *d_dst, const float *d_x, const float *a, bool aDevicePointer) const; \
  template void LBFGS<double>::dispatch_scale(const size_t n, double *d_dst, const double *d_x, const double *a, bool aDevicePointer) const; \
  template void LBFGS<float>::dispatch_axpy(const size_t n, float *d_dst, const float *d_y, const float *d_x, const float *a, bool aDevicePointer) const; \
  template void LBFGS<double>::dispatch_axpy(const size_t n, double *d_dst, const double *d_y, const double *d_x, const double *a, bool aDevicePointer) const; \
  template LBFGS<float>::status classname<float>::LBFGS_gpu(float *d_x, int n, leri_evaluate_t proc_evaluate, void* instance); \
  template LBFGS<double>::status classname<double>::LBFGS_gpu(double *d_x, int n, leri_evaluate_t proc_evaluate, void* instance); \
  template LBFGS<float>::status LBFGS<float>::LBFGS_minimize(float *d_x, int n, leri_evaluate_t proc_evaluate, void* instance); \
  template LBFGS<double>::status LBFGS<double>::LBFGS_minimize(double *d_x, int n, leri_evaluate_t proc_evaluate, void* instance)

*/


#ifndef _COLORS_
#define _COLORS_

/* FOREGROUND */
#define RST   "\x1B[0m"
#define KRED  "\x1B[31m"
#define KGRN  "\x1B[32m"
#define KYEL  "\x1B[33m"
#define KBLU  "\x1B[34m"
#define KMAG  "\x1B[35m"
#define KCYN  "\x1B[36m"
#define KWHT  "\x1B[37m"

#define FRED(x) KRED x RST
#define FGRN(x) KGRN x RST
#define FYEL(x) KYEL x RST
#define FBLU(x) KBLU x RST
#define FMAG(x) KMAG x RST
#define FCYN(x) KCYN x RST
#define FWHT(x) KWHT x RST

#define printf_red(x) printf("-- \x1B[31m%s\x1B[0m", string(x).c_str());
#define printf_grn(x) printf("-- \x1B[32m%s\x1B[0m", string(x).c_str());
#define printf_yel(x) printf("-- \x1B[33m%s\x1B[0m", string(x).c_str());
#define printf_blu(x) printf("-- \x1B[34m%s\x1B[0m", string(x).c_str());
#define printf_mag(x) printf("-- \x1B[35m%s\x1B[0m", string(x).c_str());
#define printf_cyn(x) printf("-- \x1B[36m%s\x1B[0m", string(x).c_str());
#define printf_wht(x) printf("-- \x1B[37m%s\x1B[0m", string(x).c_str());
#define printf_blk(x) printf("-- %s", string(x).c_str());

#define BOLD(x) "\x1B[1m" x RST
#define UNDL(x) "\x1B[4m" x RST

#endif  /* _COLORS_ */ 

namespace leri {

// Common functions and classes from std that caffe often uses.
using namespace std;
using std::fstream;
using std::ifstream;
using std::ofstream;
using std::ostream;
using std::ios;
using std::cerr;
using std::isnan;
using std::isinf;
using std::iterator;
using std::make_pair;
using std::map;
using std::ostringstream;
using std::pair;
using std::set;
using std::string;
using std::to_string;
using std::stringstream;
using std::numeric_limits;
using std::istreambuf_iterator;
using std::streambuf;
using std::istream;
using std::vector;
using std::swap;
using std::endl;
using std::max;
using std::min;
using std::sort;
using std::count_if;
using std::mt19937;
using namespace std::chrono;

using boost::shared_ptr;
using namespace boost::filesystem;

// #ifdef USE_EIGEN
// using Eigen::MatrixXf;
// using Eigen::MatrixXd;
// using Eigen::VectorXd;
// using Eigen::VectorXf;
// using Eigen::VectorXd;
// #endif // USE_EIGEN

#ifdef USE_TBB
// Intel's TBB
using namespace tbb;
#endif

#ifdef USE_OPENCV
using namespace cv;
#endif // USE_OPENCV

void GlobalInit(int* pargc, char*** pargv);

// A singleton class to hold common leri stuff
class Leri {
public:
  ~Leri() {}
  // static Leri& Get();
  enum Brew { CPU, GPU };

  // This random number generator facade hides boost and CUDA rng
  // implementation from one another (for cross-platform compatibility).
  // class RNG {
  // public:
  //   // RNG();
  //   // explicit RNG(unsigned int seed);
  //   // explicit RNG(const RNG&);
  //   // RNG& operator=(const RNG&);
  //   // void* generator();
  //  private:
  //   // class Generator;
  //   // shared_ptr<Generator> generator_;
  // };

  // Getters for boost rng, curand, and cublas handles
  // inline static RNG& rng_stream() {
  //   // if (!Get().random_generator_) {
  //   //   Get().random_generator_.reset(new RNG());
  //   // }
  //   // return *(Get().random_generator_);
  // }
// #ifdef USE_GPU
//   inline static cublasHandle_t cublas_handle() { return Get().cublas_handle_; }
//   inline static curandGenerator_t curand_generator() {
//     return Get().curand_generator_;
//   }
// #endif

  // // Returns the mode: running on CPU or GPU.
  // // inline static Brew mode() { return Get().mode_; }
  // // The setters for the variables
  // // Set the mode. It is recommended that you don't change the mode halfway
  // // into the program since that may cause allocation of pinned memory being
  // // freed in a non-pinned way, which may cause problems - I haven't verified
  // // it personally but better to note it here in the header file.
  // // inline static void set_mode(Brew mode) { Get().mode_ = mode; }
  // // Set the random seed of both boost and curand
  // // static void set_random_seed(const unsigned int seed);
  // // Set the device. Since we have cublas and curand stuff, set device also
  // // requires us to reset those values.
  // static void SetDevice(const int device_id);
  // // Prints the current GPU status.
  // static void DeviceQuery();
  // // Check if specified device is available
  // static bool CheckDevice(const int device_id);
  // // Search from start_id to the highest possible device ordinal,
  // // return the ordinal of the first available device.
  // static int FindDevice(const int start_id = 0);
  // // Parallel info
  // // inline static int solver_count() { return Get().solver_count_; }
  // // inline static void set_solver_count(int val) { Get().solver_count_ = val; }
  // // inline static bool root_solver() { return Get().root_solver_; }
  // // inline static void set_root_solver(bool val) { Get().root_solver_ = val; }

 protected:
// #ifdef USE_GPU
//   cublasHandle_t cublas_handle_;
//   curandGenerator_t curand_generator_;
// #endif
//   shared_ptr<RNG> random_generator_;

  Brew mode_;
  // int solver_count_;
  // bool root_solver_;

 private:
  // The private constructor to avoid duplicate instantiation.
  Leri() {};
  DISABLE_COPY_AND_ASSIGN(Leri);
};


#define WELCOME         0

#define ERROR_SINGLE    1
#define ERROR_MULTIPLE  2
#define ERROR_EXIST     3
#define ERROR_LOCATE    4
#define ERROR_OPEN      5
#define ERROR_READ      6
#define ERROR_WRITE     7
#define ERROR_FAILED    8
#define ERROR_OTHER     9

#define WARNING        11
#define USAGE          21
#define OPTION         31
#define INVOCATION     51

#define INFO_INFO      60
#define INFO_LOAD      61
#define INFO_OPEN      62
#define INFO_READ      63
#define INFO_INIT      64
#define INFO_ING       65


static inline void print_text(std::string str = "", int color = -1, char delimiter = '\n') {
  if      (color == RED)     { printf_red(str); } // Red
  else if (color == GREEN)   { printf_grn(str); } // Green
  else if (color == YELLOW)  { printf_yel(str); } // Yellow
  else if (color == BLUE)    { printf_blu(str); } // Blue
  else if (color == MAGENTA) { printf_mag(str); } // Magenta
  else if (color == CYAN)    { printf_cyn(str); } // Cyan
  else if (color == WHITE)   { printf_wht(str); } // White
  else                       { printf_blk(str); } // Black
  printf("%c", delimiter);
}
static inline void print_info(std::string str = "", int s = -1, char delimiter = '\n') {
  std::string str_ = "";
  int color = BLACK;
  switch(s) {
    case WELCOME:        str_ = "Hi there, I am Amo!"; color = GREEN; break;
    case ERROR_SINGLE:   str_ = "I found an error and list it below!"; color = RED; break;
    case ERROR_MULTIPLE: str_ = "I found errors and list them below!"; color = RED; break;
    case INVOCATION:     str_ = "Invocation:"; color = BLUE; break;
    case USAGE:          str_ = "Usage:"; color = GREEN; break;
    case OPTION:         str_ = "Options:"; color = GREEN; break;
    case WARNING:        str_ = "Warning: " + str; color = YELLOW; break;
    case ERROR_EXIST:    str_ = "Error: the file " + str + " does not exist!"; color = RED; break;
    case ERROR_LOCATE:   str_ = "Error: where to locate the " + str + "?"; color = RED; break;
    case ERROR_OPEN:     str_ = "Error: cannot open " + str; color = RED; break;
    case ERROR_READ:     str_ = "Error: cannot read " + str; color = RED; break;
    case ERROR_WRITE:    str_ = "Error: cannot write out " + str; color = RED; break;
    case ERROR_FAILED:   str_ = "Error: failed to open " + str; color = RED; break;
    case ERROR_OTHER:    str_ = "Error: " + str; color = RED; break;
    case INFO_LOAD:      str_ = "Loading " + str + " ..."; color = BLACK; break;
    case INFO_OPEN:      str_ = "Opening " + str + " ..."; color = BLACK; break;
    case INFO_READ:      str_ = "Reading " + str + " ..."; color = BLACK; break;
    case INFO_INIT:      str_ = "Initializing " + str + " ..."; color = BLACK; break;
    case INFO_ING:       str_ = str + " ..."; color = BLACK; break;
    case INFO_INFO:      str_ = str; color = BLACK; break;
    default:             str_ = str; color = BLACK; break;
  }
  print_text(str_, color, delimiter);
  // if (color == RED) { exit (EXIT_FAILURE); }
}

}  // namespace leri

#endif  // LERI_COMMON_H_
