/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright (C) 2015- All Rights Reserved - Yaan J. Jang.
 * Contact: yaan.jang 'AT' gmail.com
 *
 */

#ifndef LERI_DISTANCE_H_
#define LERI_DISTANCE_H_
#include "leri/common.h"

namespace leri {

// Some useful defines to access the components of an atom in a 1D array where the coordinates are stored one
// after another.
// #define xcoord_index(atom)	(atom * 4 + 0)
// #define ycoord_index(atom)	(atom * 4 + 1)
// #define zcoord_index(atom)	(atom * 4 + 2)
// #define wcoord_index(atom)	(atom * 4 + 3)

/**
 * Sets all the entries in distmask to FALSE.
 * @param distmask bool*
 * @param natoms int
 * @return void
 */
// void clear_distmask(bool *distmask, int natoms);

/**
 * Calculates the distances between all the natoms, whose xyz coordinates are stored in the array coords.
 * @param distances vector<vector<double> >&
 *@param coords vector<vector<double> >& 
 * @param natoms int
 * @return void
 */
template <typename Dtype>
void calculate_all_distances(vector<vector<Dtype> >& distances, vector<vector<Dtype> >coords, int natoms) {
  Dtype xi, yi, zi;
  Dtype xj, yj, zj;
  Dtype dx, dy, dz;
  Dtype dist;
  for (int i = 0; i < natoms; i++) {
    xi = coords[i][0];// xcoord_index(i)];
    yi = coords[i][1];// [ycoord_index(i)];
    zi = coords[i][2];// [zcoord_index(i)];
    for (int j = i + 1; j < natoms; j++) {
      xj = coords[j][0];// [xcoord_index(j)];
      yj = coords[j][1];// [ycoord_index(j)];
      zj = coords[j][2];// [zcoord_index(j)];
      
      dx = xi - xj;
      dy = yi - yj;
      dz = zi - zj;

      #ifdef __USE_SQUARED_DISTANCES__
        // dist = dx * dx + dy * dy + dz * dz;
        dist = sqrt(dx * dx + dy * dy + dz * dz);
      #else
        dist = sqrt(dx * dx + dy * dy + dz * dz);
        // LOG(INFO)<<dist;
      #endif
      
      // distances[distance_index(i, j, natoms)] = dist;
      // distances[distance_index(j, i, natoms)] = dist;
      distances[i][j] = dist;
      distances[j][i] = dist;
    }
  }
  // return NO_Error;
}
/**
 * Calculates the distances between the atom pairs that have TRUE in the correspoing entry in the distance mask distmask. 
 * @param distances Dtype*
 *@param coords vector<vector<Dtype> >& 
 * @param natoms int
 *@param distmask bool*
 * @return void
 */
template <typename Dtype>
void calculate_masked_distances(vector<vector<Dtype> >& distances, vector<vector<Dtype> >& coords, int natoms, vector<vector<bool> >& distmask){
  // LOG(INFO)<<"@calculate_masked_distances";
  // int i, j;
  Dtype xi, yi, zi;
  Dtype xj, yj, zj;
  Dtype dx, dy, dz;
  Dtype dist;
  for (int i = 0; i < natoms; i++) {
    xi = coords[i][0];// xcoord_index(i)];
    yi = coords[i][1];// ycoord_index(i)];
    zi = coords[i][2];// zcoord_index(i)];
    for (int j = i + 1; j < natoms; j++) {
      if (distmask[i][j] == 1) {
        xj = coords[j][0];// xcoord_index(j)];
        yj = coords[j][1];// ycoord_index(j)];
        zj = coords[j][2];// zcoord_index(j)];

        dx = xi - xj;
        dy = yi - yj;
        dz = zi - zj;

        #ifdef __USE_SQUARED_DISTANCES__
          dist = dx * dx + dy * dy + dz * dz;
        #else
          dist = sqrt(dx * dx + dy * dy + dz * dz);
        #endif
            
        // distances[distance_index(i, j, natoms)] = dist;
        // distances[distance_index(j, i, natoms)] = dist;
        distances[i][j] = dist;
        distances[j][i] = dist;
      }
    }
  }
    // return NO_Error;
}

/**
 * Calculates the inverse distances between all the natoms, whose xyz coordinates are stored in the array coords.
 * @param distances double*
 *@param coords vector<vector<double> >& 
 * @param natoms int
 * @return void
 */
template <typename Dtype>
void calculate_all_invdistances(double *invdistances, double *coords, int natoms){
  LOG(INFO)<<"@calculate_all_invdistances";
    // int i, j;
    // double xi, yi, zi;
    // double xj, yj, zj;
    // double dx, dy, dz;
    // double dist;
    // for (i = 0; i < natoms; i++)
    // {
    //     xi = coords[xcoord_index(i)];
    //     yi = coords[ycoord_index(i)];
    //     zi = coords[zcoord_index(i)];
    //     for (j = i + 1; j < natoms; j++)
    //     {
    //         xj = coords[xcoord_index(j)];
    //         yj = coords[ycoord_index(j)];
    //         zj = coords[zcoord_index(j)];
            
    //         dx = xi - xj;
    //         dy = yi - yj;
    //         dz = zi - zj;

    //         #ifdef __USE_SQUARED_DISTANCES__
    //         dist = dx * dx + dy * dy + dz * dz;
    //         #else
    //         dist = sqrt(dx * dx + dy * dy + dz * dz);
    //         #endif
            
    //         invdistances[distance_index(i, j, natoms)] = 1.0 / dist;
    //         invdistances[distance_index(j, i, natoms)] = 1.0 / dist;
    //     }
    // }
    // return NO_Error;
}
/**
 * Calculates the inverse distances between the atom pairs that have TRUE in the correspoing entry in the distance mask distmask. 
 * @param distances double*
 *@param coords vector<vector<double> >& 
 * @param natoms int
 *@param distmask bool*
 * @return void
 */
template <typename Dtype>
void calculate_masked_invdistances(double *invdistances, double *coords, int natoms, bool *distmask){
  LOG(INFO)<<"@calculate_masked_invdistances";
    // int i, j;
    // double xi, yi, zi;
    // double xj, yj, zj;
    // double dx, dy, dz;
    // double dist;
    // for (i = 0; i < natoms; i++)
    // {
    //     xi = coords[xcoord_index(i)];
    //     yi = coords[ycoord_index(i)];
    //     zi = coords[zcoord_index(i)];
    //     for (j = i + 1; j < natoms; j++)
    //         if (distmask[distance_index(i, j, natoms)] == 1)
    //         {
    //             xj = coords[xcoord_index(j)];
    //             yj = coords[ycoord_index(j)];
    //             zj = coords[zcoord_index(j)];

    //             dx = xi - xj;
    //             dy = yi - yj;
    //             dz = zi - zj;

    //             #ifdef __USE_SQUARED_DISTANCES__
    //             dist = dx * dx + dy * dy + dz * dz;
    //             #else
    //             dist = sqrt(dx * dx + dy * dy + dz * dz);
    //             #endif
            
    //             invdistances[distance_index(i, j, natoms)] = 1.0 / dist;
    //             invdistances[distance_index(j, i, natoms)] = 1.0 / dist;
    //         }
    // }
    // return NO_Error;
}

template <typename Dtype> 
void set_distmask_intervals_product(bool value, vector<vector<bool> >& mask, Atom<Dtype>* atoms, int natoms, int atom00, int atom01, int atom10, int atom11) {
  // LOG(INFO)<<"atom00 = "<<atom00<<", atom01 = "<<atom01<<", atom10 = "<<atom10<<", atom11 = "<<atom11;
  for (int i = atom00; i <= atom01; i++) {
    for (int j = atom10; j <= atom11; j++) {
  // LOG(INFO) <<atoms[i].res;
      #ifdef __MASK_INTER_RESIDUE_PAIRS__
        if (atoms[i].res != atoms[j].res) mask[i][j] = value;
      #else
        mask[i][j] = value;
      #endif
    }
  }
}
} // namespace leri 
#endif
