/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_MOLECULAR_ENERGY_H_
#define LERI_MOLECULAR_ENERGY_H_

#include "leri/common.h"
#include "leri/protein/chain.h"
#include "leri/folding/modules.h"

namespace leri {

template <typename Dtype>
class Energy {
public:
  explicit Energy(string output_dirt = "", string job_name = "")
  :output_dirt_(output_dirt),
   job_name_(job_name) {
   	quiet_ = false;
    base_name_ = output_dirt_ + job_name_;
  }
  ~Energy() {}
  
  void init();
  void actuate();
  void finish();
  
  void set_quiet(bool q = false) { quiet_ = q; }
  void set_system(Chain<Dtype>* chain);
  void set_cfg(const string cfg = "") { cfg_file_ = cfg; }

  Dtype get_energy() { return energy_; };
  
  void calc_energy_in_chain();

private:
  // Compute energy
  void init_energy();
  void read_energy_cfg();
  // void calc_energy_in_protein(int idx);

private:
  string output_dirt_;    // output directory
  string job_name_;       // job name
  string base_name_;      // basic name
  
  string cfg_file_;
  // string seq_file_;
  // string pdb_file_;
  
  bool quiet_;
  Dtype energy_;


  // Compute energy
  int num_energy_mods_;
  string file_dopepw_;
  string file_burialaa_;
  string file_hnce_;
  string file_cnce_;
  Dtype wgt_dopepw_;
  Dtype wgt_burialaa_;
  Dtype wgt_hnce_;
  Dtype wgt_cnce_;

  vector<int> energy_mods_;
  vector<string> energy_cfgs_;
  vector<Dtype> energy_current_;
  vector<Dtype> energy_coeffs_;
  Module<Dtype> module_;


  // Protein<Dtype>* protein_;
  Chain<Dtype>* chain_;
  int num_of_proteins_;
  int num_of_chains_;

private: 
  /////////////////////////////////
  // These functions are from leri/sampling/dsa.h
  int get_energy_module(const string name);
  void execute_set_system_quiet(int mod, bool quiet = false);
  /**
   * Executes the system.
   * @param protein Protein<Dtype>*
   * @return void
   */
  // void execute_set_system_model(int mod, Chain<Dtype>* chain = NULL);
  /**
   * Executes the registered initialization function for custom module mod.
   * @param mod int
   * @return void
   */
  void execute_init_func_energy_module(int mod, string cfg_name);
  /**
   * Executes the registered pre-calculation function for energy module mod.
   * @param mod int
   * @return void
   */
  void execute_prev_func_energy_module(int mod, Chain<Dtype>* chain = NULL);
  /**
   * Executes the registered energy calculation function for energy module mod.
   * @param mod int
   * @param energy Dtype*
   * @return void
   */
  void execute_calc_energy_func_energy_module(int mod, Dtype *energy);
  /**
   * Executes the registered gradient calculation function for energy module mod.
   * @param mod int
   * @param grad Dtype*
   * @return void
   */
  void execute_calc_gradient_func_energy_module(int mod, Dtype *grad);
  /**
   * Executes the registered post-calculation function for energy module mod.
   * @param mod int
   * @return void
   */
  void execute_post_func_energy_module(int mod);
  /**
   * Executes the registered finalization custom for driver module mod.
   * @param mod int
   * @return void
   */
  void execute_finish_func_energy_module(int mod);
protected:
  DISABLE_COPY_AND_ASSIGN(Energy);
}; // class MDConfig

} // namespace leri
#endif // LERI_MOLECULAR_COMPLEX_H_