/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_NETWORK_H_
#define LERI_NETWORK_H_

#include "leri/common.h"
#include "leri/util/matrix.h"
#include "leri/util/strutil.h"
#include "leri/util/fileutil.h"
#include "leri/solvers/solver.h"
#include "leri/layers/activation.h"
#include "leri/bias/loss.h"
#include "leri/layers/layer.h"
#include "leri/layers/new_layer.h"

namespace leri {

template <typename Dtype>
class Network {
public: 
  Network(string opt_name = "")
  :nthreads_(1),
   batch_size_(1), 
   skip_energy_level_(0.f) { 
    nthreads_internal_ = 1;
    n_output_ = 0;  
    // idx_layer_ = 0;
    solver_ = new_solver<Dtype>(opt_name);
    loss_func_ = NULL;
    //vector<base_layer<Dtype> *> layer_set;
    //layer_sets_.push_back(layer_set);
    layer_sets_.resize(1);
    dwgt_sets_.resize(batch_size_);
    dbias_sets_.resize(batch_size_);
    batch_open_.resize(batch_size_);
    e_sum_running_ = 0.;
    train_correct_ = 0;
    train_samples_ = 0;
    train_skipped_ = 0;
    idx_epoch_     = 0; 
    epoch_max_     = 1000;
    train_updates_ = 0;
    acc_current_   = 0;
    old_acc_current_ = 0;
    stuck_cnt_ = 0;
    acc_best_so_far_ = 0;
    acc_cnt_best_so_far_=0;
    use_augmentation_=0;
    augment_x_ = 0;
    augment_y_ = 0;
    augment_h_flip_ = 0;
    augment_v_flip_ = 0; 
    augment_theta_ = 0;
    augment_scale_ = 0;
    augment_pad_ = leri::edge; 
    init_lock();
  }
  
  ~Network() {
    clear();
    if (loss_func_) { delete loss_func_; }
    // if(solver_) { delete solver_; }
    destroy_lock();  
  }

  bool train_class(Dtype* in, int label_index, int _thread_number = -1);
  bool train_target(Dtype* in, Dtype* target, int positive=1, int _thread_number = -1);
  int  predict_class(const Dtype* in, int nthread = -1);

  string get_configuration();
  // inline void set_max_epochs(int max_e) { if (max_e <= 0) max_e = 1; epoch_max_ = max_e; }
  inline int  get_epoch() { return idx_epoch_; }
  inline bool get_input_size(int* w, int* h, int* c);
  inline int  get_output_size() {return n_output_;}
  inline int  get_thread_count() {return nthreads_;}
  inline int  get_layer_index(const char* name);

  void enable_external_threads(int nthreads = -1);
  void enable_internal_threads(int nthreads = -1);

  void set_mini_batch_size(int batch_cnt);
  int  get_mini_batch_size() { return batch_size_; }
  void reset_mini_batch() { memset(batch_open_.data(), BATCH_FREE, batch_open_.size()); }
  void sync_mini_batch();
  // return index of next free batch
  // or returns -2 (BATCH_FILLED_COMPLETE) if no free batches - all complete (need a sync call)
  // or returns -1 (BATCH_FILLED_IN_PROCESS) if no free batches - some still in progress (must wait to see if one frees)
  int get_next_open_batch();
  int reserve_next_batch();

  inline void  set_learning_rate(Dtype alpha) { solver_->learning_rate_ = alpha; }
  inline Dtype get_learning_rate() { return solver_->learning_rate_; }
  inline void  reset_solver() { solver_->reset(); }
  
  inline bool get_smart_training() {return train_smart_;}
  inline void set_smart_training(bool flag) { train_smart_ = flag; }
  inline Dtype get_smart_train_level() { return skip_energy_level_; }
  inline void set_smart_train_level(Dtype val) { skip_energy_level_ = val; }

  // Update the weights W. 
  // use w_new = w_old - alpha dE/dw
  // E = sum: 1/2*||y-target||^2
  // note y = f(x*w)
  // dE = (target-y)*dy/dw = (target-y)*df/dw = (target-y)*df/dx* dx/dw = (target-y) * df * y_prev  
  // similarly for cross entropy

  void epoch_start(string loss_function = "mse");
  bool elvis_left_the_building();
  bool epoch_end();

  void reset_smart_training();
  void update_smart_train(const Dtype E, bool correct);
  void backward_hidden(const int my_batch_index, const int thread_number);

// #else

//   Dtype get_learning_rate() {return 0;}
//   void set_learning_rate(Dtype alpha) {}
//   void train(Dtype *in, Dtype *target){}
//   void reset() {}
//   Dtype get_smart_train_level() {return 0;}
//   void set_smart_train_level(Dtype _level) {}
//   bool get_smart_train() { return false; }
//   void set_smart_train(bool _use) {}

// #endif
  void set_random_augmentation(int translate_x, int translate_y, int flip_h, int flip_v, leri::pad_type padding = leri::edge);
  void set_random_augmentation(int translate_x, int translate_y, int flip_h, int flip_v, Dtype rotation_deg, Dtype scale, leri::pad_type padding = leri::edge);
  void clear();

  static string read_lines(const string str, void* instance);
  void set_graph(vector<vector<string> > str_graph, int depth, int width = 1);
  void read(string fname, string fmode);
  void write(string fname, string postfix = ".csv");

  // TODO: replace with functions in fileutil.h
  // bool write(ofstream& ofs, bool binary = false, bool final = false);
  // bool write0(string &filename, bool binary = false, bool final = false);
  // bool write(char *filename, bool binary = false, bool final = false);
  // string getcleanline(istream& ifs);
  // bool read(istream &ifs);
  // bool read(string filename);
  // // bool read(const char *filename);

private:
  LMatrix<Dtype> make_input(Dtype* in, const int nthreads);// bool layer_push_back(const char* layer_name, const char* layer_config); // layer_name must be unique
  bool layer_push_back(const string layer_name, const string layer_config); // layer_name must be unique
  void build_layer_sets();
  void sync_layer_sets();
  void heat_weights();
  void remove_means();
  void connect(const char* layer_name_top, const char* layer_name_bottom);
  // do not delete or modify the returned pointer. it is a live pointer to the last layer in the network
  // if calling over multiple threads, provide the thread index since the interal data is not otherwise thread safe
  void connect_all(); // TODO: not support branch/highway/resnet/inception types of architectures
  Dtype* forward(const Dtype* in, int _thread_number = -1, int _train = 0);

private:
  #ifdef mUSE_OPENMP
  omp_lock_t lock_batch_;
  void lock_batch()    { omp_set_lock(&lock_batch_); }
  void unlock_batch()  { omp_unset_lock(&lock_batch_); }
  void init_lock()     { omp_init_lock(&lock_batch_); }
  void destroy_lock()  { omp_destroy_lock(&lock_batch_); }
  int get_thread_num() { return omp_get_thread_num(); }
  #else
  void lock_batch() {}
  void unlock_batch() {}
  void init_lock(){}
  void destroy_lock() {}
  int get_thread_num() {return 0;}
  #endif

public:
  int train_correct_;
  int train_skipped_;
  int stuck_cnt_;
  int train_updates_;
  int train_samples_;
  int idx_epoch_;
  int epoch_max_;
  int acc_cnt_best_so_far_;
  Dtype acc_best_so_far_;
  Dtype old_acc_current_;
  Dtype acc_current_;
  int use_augmentation_; // 0 = off, 1 = leri, 2 = opencv
  int augment_x_;
  int augment_y_;
  int augment_h_flip_;
  int augment_v_flip_;
  Dtype augment_theta_;
  Dtype augment_scale_;
  leri::pad_type augment_pad_;
  vector< vector<base_layer<Dtype> *>> layer_sets_;
  map<string, int> layer_map_;
  vector<pair<string, string>> layer_graph_;
  vector<LMatrix<Dtype> *> wgt_;

  vector< vector<LMatrix<Dtype>>> dwgt_sets_;
  vector< vector<LMatrix<Dtype>>> dbias_sets_;
  vector< unsigned char > batch_open_; 
  
private:
  int n_output_;
  int nthreads_;
  int nthreads_internal_;
  static const int MAIN_LAYER_SET = 0; // TODO: should relate to nthreads_
  int batch_size_;
  Dtype skip_energy_level_;
  bool train_smart_;
  vector <Dtype> e_running_;
  Dtype e_sum_running_;
  loss_func* loss_func_;
  Solver<Dtype>* solver_;
  static const unsigned char BATCH_RESERVED = 1, BATCH_FREE = 0, BATCH_COMPLETE = 2;
  static const int BATCH_FILLED_COMPLETE = -2, BATCH_FILLED_IN_PROCESS = -1;
protected:
  DISABLE_COPY_AND_ASSIGN(Network);
};

} // namespace leri
#endif // LERI_NETWORK_H_
