/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_LAYER_NEW_LAYER_H_
#define LERI_LAYER_NEW_LAYER_H_

#include "leri/common.h"
#include "leri/util/strutil.h"
#include "leri/layers/layer.h"
#include "leri/layers/input_layer.h"  
#include "leri/layers/shuffle_layer.h"  
#include "leri/layers/fully_connected_layer.h"  
#include "leri/layers/max_pooling_layer.h"  
#include "leri/layers/semi_stochastic_pooling_layer.h"  
#include "leri/layers/dropout.h"  
#include "leri/layers/maxout.h"  
#include "leri/layers/deepcnet.h"  
#include "leri/layers/convolution_layer.h"
#include "leri/layers/concatenation_layer.h"

namespace leri {

template <typename Dtype>
base_layer<Dtype>* new_layer(const string layer_name, const string config) {
  base_layer<Dtype>* ptr = NULL;
  vector<string> str;
  str = string_split(config, ' ');
  string tstr = str[0];
  str2lower(tstr);
  if (tstr.compare("input") == 0) {
    ptr = new input_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]), stoi(str[3]));
  } else if (tstr.compare("convolution") == 0) {
    // ptr = new convolution_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]), stoi(str[3]), new_activation_function(str[4]));
    ptr = new convolution_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]), stoi(str[3]), new_activation_func<Dtype>(str[4]));
  } else if (tstr.compare("group_convolution") == 0) {
    // ptr = new convolution_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]), stoi(str[3]), stoi(str[4]), new_activation_function(str[5]));
    ptr = new convolution_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]), stoi(str[3]), stoi(str[4]), new_activation_func<Dtype>(str[5]));
  } else if (tstr.compare("fully_connected") == 0) {
    if (stoi(str[1]) <= 0) { print_info("fully connected layer has invalid output channels", RED); exit(EXIT_FAILURE); }
    // ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_function(str[2]));
    ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_func<Dtype>(str[2]));
  } else if (tstr.compare("softmax") == 0) {
    if (stoi(str[1]) <= 0) { print_info("softmax layer has invalid output channels", RED); exit(EXIT_FAILURE); }
    // ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_function(str[0]));
    ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_func<Dtype>(str[0]));
  } else if (tstr.compare("brokemax") == 0) {
    // ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_function("brokemax"));
    ptr = new fully_connected_layer<Dtype>(layer_name, stoi(str[1]), new_activation_func<Dtype>("brokemax"));
  } else if (tstr.compare("max_pool") == 0) {
    if(stoi(str[2]) > 0 && stoi(str[2]) <= stoi(str[1])) {
      ptr = new max_pooling_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]));
    } else {
      ptr = new max_pooling_layer<Dtype>(layer_name, stoi(str[1]));
    }
  } else if (tstr.compare("mfm") == 0) {
    ptr = new maxout_layer<Dtype>(layer_name, stoi(str[1]));
  } else if (tstr.compare("semi_stochastic_pool") == 0) {
    if(stoi(str[2]) > 0 && stoi(str[2]) <= stoi(str[1])) {
      ptr = new semi_stochastic_pooling_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[2]));
    } else {
      ptr = new semi_stochastic_pooling_layer<Dtype>(layer_name, stoi(str[1]));
    }
  } else if (tstr.compare("deepcnet") == 0) {
    // ptr = new deepcnet_layer<Dtype>(layer_name, stoi(str[1]), new_activation_function(str[2]));
    ptr = new deepcnet_layer<Dtype>(layer_name, stoi(str[1]), new_activation_func<Dtype>(str[2]));
  } else if (tstr.compare("shuffle") == 0) {
    ptr = new shuffle_layer<Dtype>(layer_name, stoi(str[1]));
  } else if (tstr.compare("dropout") == 0) {
    ptr = new dropout_layer<Dtype>(layer_name, (Dtype)stof(str[1]));
  } else if ((tstr.compare("resize") == 0) || (tstr.compare("concatenate") == 0)) {
    string pad = str[2];
    leri::pad_type p = leri::zero;
    if (pad.compare("median") == 0) { p = leri::median_edge; }
    else if (pad.compare("median_edge") == 0) { p = leri::median_edge; }
    else if (pad.compare("edge") == 0) { p = leri::edge; }
    ptr = new concatenation_layer<Dtype>(layer_name, stoi(str[1]), stoi(str[1]), p);
  } else { ptr = NULL; print_info("No valid layers found", RED); exit(EXIT_FAILURE); }
  return ptr;
}

} // namespace leri
#endif // LERI_LAYER_NEW_LAYER_H_
