/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_LSTM_H_
#define LERI_LSTM_H_

#include "leri/common.h"
#include "leri/util/fileutil.h"

namespace leri {

template <typename Dtype>
class LSTM {
public:
  LSTM(int memCells, int inputVecSize, string train_file = "", string test_file = "")
  :train_file_(train_file),
   test_file_(test_file) {
    memCells_ = memCells;
    inputVectDim = inputVecSize;
    noOfIns = 0;
  }
  LSTM(const LSTM& orig) {};
  virtual ~LSTM() {};

  void init();
  void finish() {};
  
  /**
  * 
  * @param input: training data set
  * @param output: target values
  * @param trainDataSize: training data size
  * @param timeSteps: unfolding time steps
  * @param learningRate
  * @param iterations: training iterations
  * @return 0
  */
  // int train(vector<Dtype> * input, vector<Dtype> output, int trainDataSize, int timeSteps, float learningRate, int iterations);
  int train(int trainDataSize, int timeSteps, float learningRate, int iterations);
  /**
  * predict a future point using a input vector of size n
  * 
  * input: {t-n,...,t-2,t-1,t}
  * result: {t+1} 
  * 
  * @param input: input vector for he prediction
  * @return result: predicted value
  */
  // Dtype predict(vector<Dtype> * input);
  Dtype predict();

  void read_file();
  
private:
  /**
   * Forward Propagation of the network
   *  
   * @param input: input vector
   * @param timeSteps: unfolded time steps in the input vector
   * @return 0
   */
  int forward(vector<Dtype> * input, int timeSteps);
  /**
   * Backward Propagation of the network
   * 
   * @param output: output from the forward propagation
   * @param timeSteps: unfolded time steps
   * @return 0
   */
  int backward(vector<Dtype> output, int timeSteps);
  /**
   * Initialize the weights and bias values for the gates
   * Random initialization
   * 
   * @return 0
   */
  int initWeights();
  /**
   * Clear Vectors
   * 
   * @return 0
   */
  int clearVectors();
  /**
   * print the given vector
   * 
   * @param vec: vector
   * @return 0 
   */
  // int printVector(vector<Dtype> vec);
  /**
   * Sigmoid function
   * 
   * @param x
   * @return value
   */
  Dtype sigmoid(Dtype x);
  /**
   * Sigmoid function
   * 
   * @param x: input vector
   * @return vector
   */
  vector<Dtype> sigmoid(vector<Dtype> x);
  
private:
  int memCells_;
  int inputVectDim;
  int timeSteps;

  string train_file_;
  string test_file_;
  
  // weight vector arrays
  vector<Dtype> * aWeightVecArr;
  vector<Dtype> * iWeightVecArr;
  vector<Dtype> * fWeightVecArr;
  vector<Dtype> * oWeightVecArr;
  
  // bias value arrays
  Dtype * aBiasArr;
  Dtype * iBiasArr;
  Dtype * fBiasArr;
  Dtype * oBiasArr;
  
  vector<Dtype> * memCellOutArr;
  vector<Dtype> * memCellStateArr;
  
  // gate output value arrays
  vector<Dtype> * aGateVecArr;
  vector<Dtype> * iGateVecArr;
  vector<Dtype> * fGateVecArr;
  vector<Dtype> * oGateVecArr;
  
  vector<Dtype> * aGateDeltaVecArr;
  vector<Dtype> * iGateDeltaVecArr;
  vector<Dtype> * fGateDeltaVecArr;
  vector<Dtype> * oGateDeltaVecArr;
  
  vector<Dtype> * xDeltaVecArr;
  vector<Dtype> DeltaOutVec;
  
  vector<Dtype> * aDeltaWeightVecArr;
  vector<Dtype> * iDeltaWeightVecArr;
  vector<Dtype> * fDeltaWeightVecArr;
  vector<Dtype> * oDeltaWeightVecArr;
  
  /**
   * data structures used for the prediction
   */
  int noOfIns; // number of inputs given to the trained net to predict
  vector<Dtype> * input2; // vector array created using the prediction inputs
  vector<Dtype> output2; // target vector of the inputs given to predict
  
	
}; // class LSTM

} // namespace leri
#endif // LERI_LSTM_H_
