/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

/**
 * Discrete simulated annealing sampler based in the cooling schedule proposed by 
 * AArts and Van Laarhoven (E. Aarts & & J. Korst: "Simulated annealing and 
 * Boltzmann Machines". Wiley & Sons, 1989. Also in the chapter 4 of N. Ansari, E. Hou:
 * "Computational Intelligence for Optimization". Kluwer, 1997).
 *
 */

#ifndef LERI_DSA_H_
#define LERI_DSA_H_

#include "leri/common.h" 
#include "leri/util/timer.h"
#include "leri/util/random.h"
#include "leri/gio/figure.h"
#include "leri/protein/chain.h"
#include "leri/protein/protein.h" // TODO: delete
#include "leri/folding/modules.h"

namespace leri {

#define rama_index_pos(res, bin)((res)* total_num_bins_ + bin)
// #define psi_phi_index_pos(aa0, aa1, bin)(aa0 + 20 * aa1 + 20 * 20 * bin)
#define aa_index_pos(aa0, aa1)(aa0 + 20 * aa1)

template <typename Dtype>
class DSA {
public:
  DSA(int max_num_steps = 2e3)
  :max_num_steps_(max_num_steps) {}
  ~DSA() {}

  void init(string cfg, vector<string> dirt);
  void prev_step(void);
  void actuate_step(void);
  void post_step(void);
  void finish(void);

  void set_system(Chain<Dtype>* chain) { chain_ = chain; }

private:
  void init_system();
  void set_random_seed(void);
  void read_cfg(string cfg);
  void load_par_file(void);

  void reset_status_parameters(void);
  void select_aa_from_list(void);
  void init_conformation(void);
  bool change_rama(int res);
  /**
   * Generate new conformation according to the movesets.
   * @param aa int*
   * @return bool
   */
  bool generate_new_conf(int *aa);
private:
  string output_dirt_;    // output directory
  string job_name_;       // job name
  string base_name_;
  string trajct_name_;
  // Protein<Dtype> protein_;
  Chain<Dtype>* chain_;
  Chain<Dtype>* chain_cur_;
  Chain<Dtype>* chain_opt_;
  Protein<Dtype> protein_cur_;
  Protein<Dtype> protein_opt_;


  Dtype bin_size_;


  int idx_chain_;
  int num_of_amino_acids_;
  string model_name_;
  string param_dirt_;
  string rama_dirt_;
  string aa_sample_list_;
  string aa_wgt_sample_list_;
  bool seed_flag_;
  string seed_file_;
  int num_energy_mods_;
  vector<string> energy_cfgs_;
  // int *energy_mods;
  vector<int> energy_mods_;
  // Dtype *energy_coeffs_;
  vector<Dtype> energy_coeffs_;
  int total_num_bins_;
  int total_num_phipsi_bins_;
  // vector<vector<Dtype> > omega_probs_;
  Dtype** omega_probs_;
  // vector<Dtype> psi_phi_;
  // vector<Dtype> psi_psi_;
  // vector<Dtype> phi_phi_;
  // vector<Dtype> phi_psi_long_;
  Dtype* phi_psi_;
  Dtype* psi_phi_;
  Dtype* psi_psi_;
  Dtype* phi_phi_;
  Dtype* phi_psi_long_;
  // int** rama_index_;
  // Dtype** rama_distribution_;
  vector<vector<int> > rama_index_;
  vector<vector<Dtype> > rama_distribution_;
  Dtype T; // Temperature

  struct P {
    int binsize;
    bool countermoves;
    bool sampleomega;
    Dtype deftemp0;
    Dtype defenergydev0;
    int updtempint;
    Dtype updtempd;
    Dtype logupdtempd;
    Dtype energyrmsdsmooth;
    bool checktempconv;
    Dtype convepsilon;
    int max_num_steps;
    bool randominit;
  } param_, followuppar;
  // DSAparam followuppar;
  // DSAparam param;




  int max_num_steps_;
  bool write_trajct_flag_;
  bool calc_init_temp_flag_;
























public:


  // #ifndef min
  // #define min(x, y) ((x) <(y) ?(x) :(y))
  // #endif
  // #ifndef max
  // #define  max(x, y) ((x) >(y) ?(x) :(y))
  // #endif



  // In common.h 
  // // Local copy of system variables.

  // struct ChainData *chains;        // Array of chain data.
  int nchains_;              // Total number of chains.
  // struct ResidueData *residues;    // Array of residue data.
  int nres_;                 // Total number of residues.
  // struct AtomData *atoms;          // Array of atom data.
  // int natoms;               // Total number of atoms.
  int natoms_;               // Total number of atoms.


  // Dtype *distances;           // Array of atom pair-wise distances.
  vector<vector<Dtype> > distances;
  Dtype *invdistances;        // Array of atom pair-wise 1/distances.
  bool *atommask;             // Array of atom masking values.
  // bool *distmask;             // Array of distance masking values.  
  vector<vector<bool> > distmask;





  // struct AtomData test_atom[natoms];

  // Dtype a;
  // Dtype b;
  // Dtype c;
  // Dtype d;


    







  int maxnumenergy_mods_; // TODO: remove

  // time_t starttime;
  // Timer tim;

  


  //int rbasinsmod;
  bool preround;
  int roundnum;
  int preroundnum;
  int npreround;

  // Dtype phi_psi[2];
  Dtype psi_phi_store[2];


  // Dtype *omega_probs;
    // vector<vector<Dtype> > omega_probs;

  // Dtype *psi_phi;
  // Dtype *psi_psi;
  // Dtype *phi_phi;
  // Dtype *phi_psi_long;
  // vector<Dtype> psi_phi;
  // vector<Dtype> psi_psi;
  // vector<Dtype> phi_phi;
  // vector<Dtype> phi_psi_long;





  //Tinit Options
  int MaxNInitTLoops;
  int MaxNInitTSteps;
  Dtype InitMaxAccRatio;
  Dtype MaxAccRatio;
  int DefInitTemp;
  int steps_double_crank_;

  // char* trajct_name;
  string trajct_name;
  bool convergence,finished;
  int step_;
  int ndecsteps, nincsteps, nrejsteps;
  int nramachanges;
  Dtype temp, prevtemp; 
  Dtype curenergy;
  // Dtype *curenergies;
  vector<Dtype> curenergies;
  Dtype optenergy;
  Dtype energysdevsmooth;
  int nsolutions;
  Dtype energysum;
  Dtype energysumsq;

  Dtype *curconf;
  Dtype *optconf;


  //residues sample value - No sample = 0, Pivot = 1, DC = 2 
  bool selected_res_;
  vector<int> samplevalues;
  // int *samplevalues;
  int *to_sample_, *sample_pivot_, *sample_double_crank_;
  int nP, nD, nS;
  // vector<int> to_sample_;
  // vector<int> sample_pivot_;
  // vector<int> sample_double_crank_;

  // Dtype * ResWeights;
  vector<Dtype> ResWeights;
  bool selectedwts;

  int dsaOptStep;


  Module<Dtype> module_;
  // DOPE_PW<Dtype> dope_pw;



  // void update_model(ProteinStruct<Dtype>* mdl, vector<vector<Dtype> > coords);
  void update_model(Protein<Dtype>* mdl, vector<vector<Dtype> > coords);


  // void load_omega_probs_from_file(const char* file, vector<vector<Dtype> >& omega);//Dtype *omega);
  // void load_phi_psi_dist_from_file(const char* file, Dtype* vec, const string str);//Dtype * array);
//   void load_rama_map_from_file(const char* file, vector<vector<int> >& idx_matrix, vector<vector<Dtype> >& ramadis_map, int nres, bool load_index);
// template <typename Dtype>
  // void load_rama_map_from_file(const char* file, vector<vector<Dtype> >& rama_matrix, bool load_index);

  void write_current_conf(FILE* fid_in);

  void calculate_mean_and_sdev(Dtype *energymean, Dtype *energysdev);

  int choose_residue();
  void choose_phi_psi(int res);
  void choose_psi_phi(int res);
  void set_phi_psi(int res, Dtype phi, Dtype psi, int idx_chain);
  void set_psi_phi(int res, Dtype psi, Dtype phi);

  int  general_bin(Dtype ang1, Dtype ang2);
  void generate_rand_conf();
  void generate_rand_selected();
  void restore_prev_conf(int res);
  void save_opt_conf();
  void restore_opt_conf();
  void save_cur_conf();
  void save_current_conf(int res);
  void restore_cur_conf();




private:
/**
   * Returns in *num the number of registered energy modules.
   * @param num int*
   * @return void
   */
  // void get_num_energy_modules(int *num);
  /**
   * Returns in *mod the index of the first energy module in the registry array with the specified name.
   * @param int mod*
   * @param name char*
   * @return void
   */
  void get_energy_module(int* mod, string name);
  /**
   * Returns in *name the name of energy module mod. The pointer *name should be freed after its use in the invoking code.
   * @param name char**
   * @param mod int
   * @return void
   */
  void get_energy_module_name(char **name, int mod);
  /**
   * Returns in *doc the documentation string of energy module mod.
   * @param doc char**
   * @param mod int
   * @return void
   */
  void get_energy_module_doc(char **doc, int mod);
  /**
   * Executes the system.
   * @param protein Protein<Dtype>*
   * @return void
   */
  // void execute_set_system_model(int mod, Protein<Dtype>* protein = NULL);
  void execute_set_system_model(int mod,  Chain<Dtype>* chain = NULL);
  /**
   * Executes the registered initialization function for custom module mod.
   * @param mod int
   * @return void
   */
  void execute_init_func_energy_module(int mod, string cfg_name);
  /**
   * Executes the registered pre-calculation function for energy module mod.
   * @param mod int
   * @return void
   */
  // void execute_prev_func_energy_module(int mod, Protein<Dtype>* protein = NULL);
  void execute_prev_func_energy_module(int mod, Chain<Dtype>* chain = NULL);
  /**
   * Executes the registered energy calculation function for energy module mod.
   * @param mod int
   * @param energy Dtype*
   * @return void
   */
  // void execute_calc_energy_func_energy_module(int mod, Dtype *energy, ProteinStruct<Dtype> mdl);
  // void execute_calc_energy_func_energy_module(int mod, Dtype *energy, Protein<Dtype>* protein);
  void execute_calc_energy_func_energy_module(int mod, Dtype *energy, Chain<Dtype>* chain);
  /**
   * Executes the registered gradient calculation function for energy module mod.
   * @param mod int
   * @param grad Dtype*
   * @return void
   */
  void execute_calc_gradient_func_energy_module(int mod, Dtype *grad);
  /**
   * Executes the registered post-calculation function for energy module mod.
   * @param mod int
   * @return void
   */
  void execute_post_func_energy_module(int mod);
  /**
   * Executes the registered finalization custom for driver module mod.
   * @param mod int
   * @return void
   */
  void execute_finish_func_energy_module(int mod);



private:
  Atom<Dtype>* atoms_new;
  AminoAcid<Dtype>* residues_new;
  Chain<Dtype>* chains_new;
  vector<vector<Dtype> > coords; // Array of atom coordinates.
  vector<vector<Dtype> > coords_saved;
  vector<vector<Dtype> > coords_opt;


protected:
  // DISABLE_COPY_AND_ASSIGN(DSA);
};









} // namespace leri
#endif // LERI_DSA_H_
