/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

/**
 * This file contains analysis of Nested Sampling.
 */

#ifndef LERI_NESTED_SAMPLING_H_
#define LERI_NESTED_SAMPLING_H_

#include "leri/common.h"
#include "leri/util/strutil.h"
#include "leri/gio/figure.h"
#include "leri/sampling/model.h"

namespace leri {

template <typename Dtype>
class NestedSampler {
public:
  // virtual ~NestedSampler(){}
  NestedSampler() {}
  NestedSampler(int num_particles, int mcmc_steps, ToyModel<Dtype> model);
  ~NestedSampler() {}

public:
  // Set RNG seed
  void set_rng_seed(unsigned int seed);

  // Call from_prior on all the particles
  void init();

  // Do an iteration of Nested Sampling
  void do_iteration();

  // Do an indefinite number of iterations
  void run();

  // Do a specified number of iterations
  Dtype run(int iterations);

  // Open and close (i.e. clear the contents of)
  // the output files. If not called, it will append
  // to whatever's already in them
  void clear_output_files();

  // Setter and getter for the RNG
  // void set_rng(const RNG& rng) { this->rng = rng; }
  // RNG get_rng() const { return rng; }

  // More getters
  Dtype get_logZ() const { return logZ_; }
  Dtype get_H() const { return H_; }
private:
  // MODEL model_;
  // The random number generator to use
  // RNG rng;

  // The particles, and information about them
  int num_particles;
  std::vector<ToyModel<Dtype> > particles;
  std::vector<Dtype> log_likelihoods_;
  std::vector<Dtype> tiebreakers;
  Dtype log_likelihoods_val_;

  // Whether from_prior has been called on all the particles
  bool initialised;

  // Count number of iterations done
  int iteration;

  // Number of equilibration steps
  int max_iter_;

  // Results based on deterministic approximation
  // log(X)
  // log prior mass (Skilling version),
  // log prior mass (Walter version),
  // log evidence, and information (Skilling versions)
  Dtype logX_, log_prior_mass1, log_prior_mass2, logZ_, H_;

  // Output file streams
  std::ofstream sample_file, sample_info_file;

  // Method to write a particular particle (and its info) to disk
  // void write_output(int index);

  // Function to determine whether one (likelihood, tiebreaker)
  // pair is below another
  static bool is_below(Dtype log_likelihood1, Dtype tiebreaker1,
          Dtype log_likelihood2, Dtype tiebreaker2) {
  if(log_likelihood1 < log_likelihood2)
    return true;

  if(log_likelihood1 == log_likelihood2)
  {
    if(tiebreaker1 < tiebreaker2)
      return true;
  }

  return false;
}
protected:
  DISABLE_COPY_AND_ASSIGN(NestedSampler);
}; // class NestedSampler


} // namespace leri
#endif // LERI_NESTED_SAMPLING_H_

