/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_GRADIENT_DESCENT_SOLVER_H_
#define LERI_GRADIENT_DESCENT_SOLVER_H_

#include "leri/common.h"
#include "leri/util/strutil.h"
#include "leri/util/matrix.h"
#include "leri/solvers/solver.h"

namespace leri {

/**
 * Optimizes the parameters of a given objective using
 * the Broyden–Fletcher–Goldfarb–Shanno (BFGS) algorithmwith limited-memory.
 */
template <typename Dtype>
class Gradient_Descent_Solver { // : public Solver<Dtype> {
public:
  // explicit Gradient_Descent_Solver(const SolverParam<Dtype>& param) 
  //   : Solver<Dtype>(param), gd_param(param) {
  //     // init(param);
  //   }
  Gradient_Descent_Solver() {}
  ~Gradient_Descent_Solver(){}

  virtual inline const char* type() const { return "Gradient_Descent_Solver"; }

public: 
  int nvar; // Number of varialbes
  int mrcd; // points in the latest past m iterations 
  Dtype tau;
  Dtype threshold;

  void init(int n, int m);
  void minimize(int max_iter, Dtype epsilon) ;
  void calc_bfgs_direction(         // L-BFGS two-loop recursion
    vector<Dtype>& vect,            // the bfgs direciton
    vector<vector<Dtype> >& xdiff_, // the difference of Xk+1 and Xk for k=0 to m-1
    vector<vector<Dtype> >& gdiff_, // the difference of Gk+1 and Gk for k=0 to m-1
    vector<Dtype>& rho_,            // 1.0/(Gdiff'*Xdiff)
    vector<Dtype>& alpha_,
    vector<Dtype>& grad_,           // gradient at current iteration
    int k, 
    Dtype gamma);                   // the initail approximation of inverse Hessian using gamma*I
  void backtracking_line_search(    // reference: Numerical Optimization(Jorge Nocedal,Stephen J.Wright,1999), Chapter 3, Page41.
  	Dtype* alfa, 
    vector<Dtype>& curt_point_,     // current point
    vector<Dtype>& prev_point_,     // previous point
    vector<Dtype>& curt_grad_,      // current gradient
    vector<Dtype>& ngtv_dirct_,     // negative search direction, Warning, p should not equal to g, because the procedure require p unchanged but change g when call proc_evaluate
    Dtype* fval,                    // current function value at curt_point
    // Dtype threshold_,            // sufficient decrease condition threshold
    Dtype init_step,                // initial step length
    // Dtype tau,                   // scale factor in backtracking
    int*   eval_cnt);               // fitness evaluation
  void guess_init_step(Dtype* val, vector<Dtype>& g, int iter);
  void finish();

private:
  // const SolverParam<Dtype>& gd_param;  // Parameters to control the Gradient_Descent algorithm

  vector<vector<Dtype> > xdiff;
  vector<vector<Dtype> > gdiff;
  vector<Dtype> rho;
  vector<Dtype> alpha;

  vector<Dtype> curt_point;
  vector<Dtype> prev_point;
  vector<Dtype> curt_grad;
  vector<Dtype> prev_grad;
  vector<Dtype> ngtv_dirct;

protected:
  DISABLE_COPY_AND_ASSIGN(Gradient_Descent_Solver);
};

} //namespace leri

#endif  // LERI_GRADIENT_DESCENT_SOLVER_H_
