/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_LBFGS_H_
#define LERI_LBFGS_H_

#include "leri/util/matrix.h"
#include "leri/util/fileutil.h"
#include "leri/gio/figure.h"

namespace leri {

template <typename Dtype>
class LBFGS {
public:
  LBFGS(string base_name, int n, int max_iter = -1, int m = 6) // 30)
  :base_name_(base_name),
   max_iter_(max_iter),
   num_(n),
   nmem_(m) {
    sdct_ = 1e-4;
    scale_ = 0.9;
    epsilon_ = 1e-5;
   }
  ~LBFGS() {}

  typedef Dtype (*evaluate_func)(const int n, const Dtype* xvar, Dtype* gvar, void* instance);
  struct tag_iteration_data {
    Dtype alpha;
    Dtype* s;     /* [n] */
    Dtype* y;     /* [n] */
    Dtype ys;     /* vecdot(y, s) */
  };
  typedef struct tag_iteration_data iteration_data_t;

  void init();
  void actuate(Dtype* xvar, Dtype* ptr_fx, evaluate_func proc_evaluate, void* instance);
  void finish() {};

private:
  int line_search_morethuente(int n, Dtype* x, Dtype* f, Dtype* g, Dtype* s, Dtype* stp, const Dtype* xp, const Dtype* gp, Dtype* wa, evaluate_func proc_evaluate, void* instance);
  Dtype owlqn_x1norm(const Dtype* x, const int start, const int n);
  void owlqn_pseudo_gradient(Dtype* pg, const Dtype* x, const Dtype* g, const int n, const Dtype c, const int start, const int end); 
  int update_trial_interval(Dtype* x, Dtype* fx, Dtype* dx, Dtype* y, Dtype* fy, Dtype* dy, Dtype* t, Dtype* ft, Dtype* dt, const Dtype tmin, const Dtype tmax, int* brackt);
  static inline void cubic_minimizer(Dtype* cm, Dtype* u, Dtype* fu, Dtype* du, Dtype* v, Dtype* fv, Dtype* dv);
  static inline void cubic_minimizer(Dtype* cm, Dtype* u, Dtype* fu, Dtype* du, Dtype* v, Dtype* fv, Dtype* dv, const Dtype xmin, const Dtype xmax);

private:
  string base_name_;
  int max_iter_;
  int num_;       // number of varialbes
  int nmem_;      // number of limited memory
  int eval_cnt_;  // counter of evaluations
  Dtype sdct_;    // sufficient decrease condition threshold
  Dtype scale_;   // scale factor in backtracking
  // Dtype epsilon_;

  int m_;                 // default: 6, The number of corrections to approximate the inverse hessian matrix.
  Dtype epsilon_;         // default: 1e-5, epsilon for convergence, ||g|| < \ref epsilon * max(1, ||x||)
  int past_;              // default: 0, distance for delta-based convergence
  Dtype delta_;           // default: 0, Delta for convergence, (f' - f) / f <  delta
  // int  max_iter_;      // default: 100, The maximum number of iterations.
  int  linesearch_;
  int  max_linesearch_;   // default: 20, The maximum number of trials for the line search
  Dtype min_step_;        // default: 1e-20, The minimum step of the line search routine
  Dtype max_step_;        // default: 1e+20, The maximum step of the line search.
  Dtype ftol_;            // default: 1e-4, in (0, 0.5), A parameter to control the accuracy of the line search routine.
  Dtype wolfe_;           // default: 0.9, (ftol, 1.0), A coefficient for the Wolfe condition.
  Dtype gtol_;            // default: 0.9, (ftol, 1.0)
  Dtype xtol_;            // default: positive,  The machine precision for floating-point values. 

protected:
  DISABLE_COPY_AND_ASSIGN(LBFGS);
};

} // namespace leri

#endif // LERI_LBFGS_H_
