/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_FILEUTILS_H_
#define LERI_FILEUTILS_H_

#include <bzlib.h>
#include "leri/common.h"
#include "leri/util/strutil.h"

namespace leri {

static inline void copyright(string &str, string csymbol = "#") {
  csymbol += " ";
  str  = csymbol + str + "\n";
  str += csymbol + VERS_INFO("");
  str += csymbol + EMAIL + "\n";
}

static inline bool exist_file(string fname) {
  ifstream fid_in(fname);
  return fid_in.good();
}

static inline bool is_file_empty(string fname) {
  bool ret = true;
  if (exist_file(fname)) {
    ifstream fid_in(fname, ifstream::in);
    if (fid_in.is_open()) {
      ret = fid_in.peek() == ifstream::traits_type::eof();
      fid_in.close();
    }
  }
  return ret;
}
static inline bool exist_dirt(string dname) {
  bool ret = false;
  boost::filesystem::path pdirt(dname);
  if (boost::filesystem::exists(pdirt) && boost::filesystem::is_directory(pdirt)) {
    ret = true;
  }
  return ret;
}

// Basic methods
void dirt_check(string& dirt);
void get_file_name(string& pfnam, const string original);
void get_file_list(vector<string>& list, const string& dirt);
void get_num_of_lines_in_file(int* n, const string file, int nskip = 0);
/**
 * File name definition depending on the directory and postfix.
 * @param pfnam string&
 * @return void
 */
void define_output_file_name(string& pfnam, const string original, const string path, const string postfix);


template <typename Dtype> // C++11
using def_func = string(*)(const string str, void *instance);
string read_lines(const string str, void* instance);
template <typename Dtype> 
vector<string> read_file(string fname, def_func<Dtype> ptr_func, void *instance = NULL) {
  // ifstream fid_in;
  // fid_in.open(fname, ifstream::in);
  ifstream fid_in(fname, ifstream::in);
  if (!fid_in.is_open()) {
    print_info(fname, ERROR_OPEN);
    exit(EXIT_FAILURE);
  }
  vector<string> vstr; 
  if (fid_in.is_open()) {
    string line = "";
    // getline(fid_in, line); // Skip the first line
    while (fid_in) {
      if (!getline(fid_in, line)) { break; }
      trim_all(&line);
      if (line.length() < 1) { continue; }
      string tstr = ptr_func(line, instance);
      if (!tstr.empty()) {
        vstr.push_back(ptr_func(line, instance));
      }
      // string x = ptr_func(line, instance);
    }
    fid_in.close();
  }
  return vstr;
}




// Reading methods
template <typename Dtype> 
void read_matrix_from_atgc_file(vector<string>& name, vector<vector<Dtype> >& mat, const string file, char delimiter = ',');
template <typename Dtype> 
void read_eca_matrix_from_file(const string file, int num_site, Dtype** site_mtx = NULL, Dtype**** cpld_mtx = NULL, Dtype scale = -1.0, char delimiter = ' ');
vector<string> read_matrix_from_bz2file(string file, char delimiter);
template <typename Dtype> 
void read_matrix_from_bz2file(vector<vector<Dtype> >& mat, string file_name, char delimiter);
template <typename Dtype> 
void read_matrix_from_file_3column(Dtype** mtx, string file, char delimiter);
template <typename Dtype> 
void read_matrix_from_file(vector<vector<Dtype> >& mat, string file, char delimiter = ',');
template <typename Dtype> 
void read_matrix_from_file(Dtype* vec, const string file, const string str);
template <typename Dtype> 
void read_matrix_w_index(vector<vector<Dtype> >& mat, vector<string>& name_row, const string file, int nskip = 0, char delimiter = ',');


// Writting methods
void write_string(string fnam, string content, string tstr = "", string mode = "w", string postfix = "", string csymbol = "#", string lb = "\n");


template <typename Dtype> 
void write_vector(string out, Dtype* vec, int ncol, 
                  string delimiter = ",", string postfix = ".csv", 
                  string str = "", string txt = "", 
                  string fmt = "f", int fixp = 5, string mode = "w");

template <typename Dtype> 
void write_matrix(string out, Dtype** mat, int nrow, int ncol, 
                  int type = 1, // 1: nrow by ncol, 2: three columns
                  string str = "", string delimiter = ",", string postfix = ".csv", //  separator, postfix, text for copyright
                  string fmt = "f", int fixp = 5,
                  int* xlabel = NULL, int* ylabel = NULL);
template <typename Dtype> 
void write_matrix_w_index(string out,
                          Dtype** mat, int num[], 
                          vector<string> ytickers = vector<string>(), 
                          vector<string> xtickers = vector<string>(), 
                          bool yflag = true, 
                          bool gfflag = false, 
                          vector<string> glabels = vector<string>(), 
                          vector<string> flabels = vector<string>(), 
                          bool copyright = true,
                          int psize = -1, 
                          // string str = "", 
                          string delimiter = ",", 
                          string postfix = ".csv");

// Read bz2 file
BZFILE* bz2_open(const char* filename);
void bz2_read(string* str, void* bzfile);
void bz2_close(void* bzfile);
void read_bz2_file(string* str, const string filename);

template <typename Dtype> 
void read_matrix_from_atgc_file(vector<string>& name, vector<vector<Dtype> >& mat, const string file, char delimiter) {
  // ifstream fid_in;
  string line;
  int nrow = 0;
  int ncol = 0;
  vector<Dtype> vect;
  // fid_in.open(file, ifstream::in);
  ifstream fid_in(file, ifstream::in);
  if (fid_in.is_open()) {
    getline(fid_in, line); // Skip first line
    while (fid_in) {
      if (!getline(fid_in, line)) break;
      trim_all(&line);
      trim_all(&line);
      // LOG(INFO)<<line<<"=";
      if (line[0] == '#' || line.length() < 1) continue;
      line = replace_space(line);
      stringstream ss(line);
      vector <string> str;
      string subline;
      while (ss) {
        if (!getline(ss, subline, delimiter)) break;
        // LOG(INFO)<<"="<<subline<<"=";
        trim_all(&subline);
        trim_all(&subline);
        str.push_back(subline);
      }
      ncol = str.size();
      if (ncol < 1) continue;
      name.push_back(str[0]);
      if (ncol < 3) { // bad data points
        vect.push_back(0);
        vect.push_back(0);
      } else {
        for (int i = 1; i < ncol; i++) {
          vect.push_back(atof(str[i].c_str()));
        }
      }
      mat.push_back(vect);
      vect.clear();
      ++nrow;
    }
    fid_in.close();
  } else {
    printf("\n-- \033[1;31m");
    printf("Error: I cannot load data from: <%s>", file.c_str());
    printf("\033[0m\n");
    exit (EXIT_FAILURE);
  }
}

template <typename Dtype> 
void read_eca_matrix_from_file(const string file, int num_site, Dtype** site_mtx, Dtype**** cpld_mtx, Dtype scale, char delimiter) {
  // ifstream fid_in;
  string line;
  int row = 0;
  int idx   = 0;
  int idx_0 = 0;
  int idx_1 = 0;
  vector<string> vec_str;
  // fid_in.open(file, ifstream::in);
  ifstream fid_in(file, ifstream::in);
  if (fid_in.is_open()) {
    getline(fid_in, line); // Skip the first line
    while (fid_in) {
      if (!getline(fid_in, line)) break;
      trim_all(&line);
      if (line.length() < 1) continue;
      if (row < num_site) {
        if (line[0] == '#') continue;
        vec_str = string_split(line, delimiter);
        for (int k = 0; k < vec_str.size(); k++) {
          site_mtx[row][k] = scale*atof(vec_str[k].c_str());
        }
        row++;
      } else {
        vec_str = string_split(line, delimiter);
        if (line[1] == '#') {
          idx   = 0;
          idx_0 = atoi(vec_str[2].c_str())-1;
          idx_1 = atoi(vec_str[4].c_str())-1;
        } else {
          for (int k = 0; k < vec_str.size(); k++) {
            Dtype val = atof(vec_str[k].c_str());
            cpld_mtx[idx_0][idx_1][idx][k] = scale*val;
            cpld_mtx[idx_1][idx_0][k][idx] = scale*val;
          }
          idx++;
        }
      }
    }
    fid_in.close();
  } else {
    printf("\n-- \033[1;31m");
    printf("Error: I cannot load data from: <%s>", file.c_str());
    printf("\033[0m\n");
    exit (EXIT_FAILURE);
  }
}
// TODO: remove
template <typename Dtype> 
void compress_matrix_to_bz2file(string file_name, char delimiter) {
  vector<vector<Dtype> > mat;
  FILE* file = fopen(file_name.c_str(), "rb");
  if (file != NULL) {
    string out_file_name = file_name;
    out_file_name += "2.bz2";
    FILE* out_file = fopen(out_file_name.c_str(), "wb");
    if (0 == out_file) {
      fclose(file);
      exit (EXIT_FAILURE);
    }
    int bz2error;
    BZFILE* out_bz2 = BZ2_bzWriteOpen(&bz2error, out_file, 9, 0, 0);
    if (BZ_OK != bz2error) {
      fclose(file);
      fclose(out_file);
      exit (EXIT_FAILURE);
    }
    // load file data into buffer, and store them to bzip2-out stream
    char buffer[1024];
    uint loaded_len;
    while ((loaded_len = fread(buffer, 1, 1024, file)) > 0) {
      BZ2_bzWrite(&bz2error, out_bz2, buffer, loaded_len);
      if (BZ_OK != bz2error) {
        fclose(file);
        BZ2_bzWriteClose(&bz2error, out_bz2, 1, 0, 0);
        fclose(out_file);
        exit (EXIT_FAILURE);
      }
    }
    fclose(file);
    uint total_in, total_cmp;   // how many bytes we read, and resulting compressed size
    BZ2_bzWriteClose(&bz2error, out_bz2, 0, &total_in, &total_cmp);
    fclose(out_file);
  }
}

template <typename Dtype> 
void read_matrix_from_bz2file(vector<vector<Dtype> >& mat, string filename, char delimiter) {
  mat.clear();
  // BZFILE* bzfile = bz2_open(file_name.c_str());
  string str = "";
  // bz2_read(&str, bzfile);
  read_bz2_file(&str, filename); 
  // bz2_close(bzfile);
  vector<string> lines = vector<string>();
  lines = string_split(str, '\n');
  for (int i = 0; i < lines.size(); i++) {
    string line = lines[i];
    trim_all(&line);
    if (line[0] != '#' && line.length() > 0) {
      vector<string> str_vec = string_split(line, delimiter);
      vector<Dtype> vec = vector<Dtype>();
      for (int j = 0; j < str_vec.size(); j++) {
        vec.push_back(atof(str_vec[j].c_str()));
      }
      mat.push_back(vec);
      str_vec.clear();
      vec.clear();
      line = "";
    }
  }
  lines.clear();
}
template <typename Dtype> 
void read_matrix_from_file_3column(Dtype** mtx, string file, char delimiter) {
  // ifstream fid_in;
  string line;
  int nrow = 0;
  vector<vector<Dtype> > mat;
  vector<Dtype> vect;
  // fid_in.open(file, ifstream::in);
  ifstream fid_in(file, ifstream::in);
  if (fid_in.is_open()) {
    while (fid_in) {
      if (!getline(fid_in, line)) break;
      trim_all(&line);
      if (line[0] == '#' || line.length() < 1) continue;
      stringstream ss(line);
      vector <string> str;
      string subline;
      while (ss) {
        if (!getline(ss, subline, delimiter)) break;
        str.push_back(subline);
      }
      for (int i = 0; i < str.size(); i++) {
        vect.push_back(atof(str[i].c_str()));
      }
      mat.push_back(vect);
      vect.clear();
      ++nrow;
    }
    fid_in.close();
    #ifdef USE_OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < nrow; i++) {
      int m = mat[i][0];
      int n = mat[i][1];
      mtx[m][n] = mat[i][2];
    }
  } else {
    printf("\n-- \033[1;31m");
    printf("Error: I cannot load data from: <%s>", file.c_str());
    printf("\033[0m\n");
    exit (EXIT_FAILURE);
  }
}
template <typename Dtype> 
void read_matrix_from_file(vector<vector<Dtype> >& mat, string file, char delimiter) {
  mat.clear();
  // ifstream fid_in;
  string line;
  int row = 0;
  vector<Dtype> vect;
  // fid_in.open(file, ifstream::in);
  ifstream fid_in(file, ifstream::in);
  if (fid_in.is_open()) {
    while (fid_in) {
      if (!getline(fid_in, line)) break;
      trim_all(&line);
      if (line[0] == '#' || line.length() < 1) continue;
      stringstream ss(line);
      vector <string> str;
      string subline;
      while (ss) {
        if (!getline(ss, subline, delimiter)) break;
        str.push_back(subline);
      }
      for (int i = 0; i < str.size(); i++) {
        vect.push_back(atof(str[i].c_str()));
      }
      mat.push_back(vect);
      vect.clear();
      ++row;
    }
    fid_in.close();
  } else {
    printf("\n-- \033[1;31m");
    printf("Error: I cannot load data from: <%s>", file.c_str());
    printf("\033[0m\n");
    exit (EXIT_FAILURE);
  }
}
// TODO: merge
#define psi_phi_index_pos(aa0, aa1, bin)(aa0 + 20 * aa1 + 20 * 20 * bin)
template <typename Dtype> 
void read_matrix_from_file(Dtype* vec, const string file, const string str) {
// load_phi_psi_dist_from_file(const char* file, Dtype* vec, const string str);
  int tocount = 18*18+2, index, currentindex = 0;
  int indices[2];
  int cnt = 0;
  // Read data
  string line;
  // ifstream fid_in;
  // fid_in.open(file, ios::in);
  ifstream fid_in(file, ios::in);
  if(fid_in.is_open()) {
    printf("-- Loading %s probilities ... ", str.c_str());
    while(getline(fid_in, line)) {
      trim_all(&line);
      if(line[0] == '#' || line.length() == 0) { continue; }
      Dtype currentvalue;
      int modvalue = cnt % tocount;
      if(modvalue == 0 || modvalue==1) {
        index = atoi(line.c_str());
        indices[modvalue] = index;
      } else {
        currentvalue = atof(line.c_str());
        currentindex = psi_phi_index_pos(indices[0], indices[1], (modvalue-2)); // WARNING: parenthases here very important because a macro is used, not a function
        vec[currentindex] = currentvalue;
      }
      cnt++;
    }
    fid_in.close();
    printf("done.\n");
  } else {
    printf("\n-- \033[1;31m");
    printf( "Error: I cannot open file: %s\n", file.c_str());
    printf("\033[0m\n");
  }
}

template <typename Dtype> 
void read_matrix_w_index(vector<vector<Dtype> >& mat, vector<string>& name_row, const string file, int nskip, char delimiter) {
  vector<string> name_col;
  // ifstream fid_in;
  string line;
  int nrow = 0;
  vector<Dtype> vect;
  // fid_in.open(file, ifstream::in);
  ifstream fid_in(file, ifstream::in);
  if (fid_in.is_open()) {
    for (int i = 0; i < nskip; i++) {
      getline(fid_in, line);
    }
    while(getline(fid_in, line)) {
      trim_all(&line);
      // if(line[0] == '#' || line.length() == 0) { continue; }
      if (line[0] == '#' || line.length() < 1) continue;
      if (line.length() < 1) continue;
      stringstream ss(line);
      vector <string> str;
      string subline;
      while (ss) {
        if (!getline(ss, subline, delimiter)) break;
        str.push_back(subline);
      }
      for (int i = 0; i < str.size(); i++) {
        if (line[0] == '#') {
          name_col.push_back(str[i]);
        } else {
          if (i) {
            vect.push_back(atof(str[i].c_str()));
          } else {
            name_row.push_back(str[i]);
          }
        }
      }
      if (line[0] != '#') {
        mat.push_back(vect);
        vect.clear();
      }
      ++nrow;
    }
    fid_in.close();
  } else {
    printf("\n-- \033[1;31m");
    printf( "Error: I cannot open file: %s\n", file.c_str());
    printf("\033[0m\n");
  }
}




template <typename Dtype> 
void write_vector(string out, Dtype* vec, int ncol, 
                  string delimiter, string postfix, 
                  string str, string txt, 
                  string fmt, int fixp, string mode) {
  out += postfix;
  FILE* fid = fopen(out.c_str(), mode.c_str());
  if (!str.empty()) {
    copyright(str);
    fprintf(fid, "%s\n", str.c_str());
  }
  if (txt.compare("")) {
    fprintf(fid, "%s\n", txt.c_str());
  }
  for (int j = 0; j < ncol; j++) {
    string pstr = "%." + num2string(fixp) + fmt;
    if (vec[j] == 0) { pstr = "%1." + fmt;}
    fprintf(fid, pstr.c_str(), vec[j]);
    if (j < ncol-1) { fprintf(fid, "%s", delimiter.c_str()); }
    else { fprintf(fid, "\n"); }
  }
  fclose(fid);
}
template <typename Dtype> 
void write_matrix(string out, Dtype** mat, int nrow, int ncol, 
                  int type, 
                  string str, string delimiter, string postfix, 
                  string fmt, int fixp,
                  int* xlabel, int* ylabel) {
  out += postfix;
  FILE* fid = fopen(out.c_str(), "w");
  if (!str.empty()) {
    copyright(str);
    fprintf(fid, "%s\n", str.c_str());
  }
  for (int i = 0; i < nrow; i++) {
    for (int j = 0; j < ncol; j++) {
      string pstr = "";
      if (type == 0) {
        pstr = "%." + num2string(fixp) + fmt;
        if (mat[i][j] == 0 || j == 0) { pstr = "%1." + fmt; }
        fprintf(fid, pstr.c_str(), mat[i][j]);
        if (j < ncol-1) { fprintf(fid, "%s", delimiter.c_str()); }
        else { fprintf(fid, "\n"); }
      }
      if (type == 1) {
        if (!fmt.compare("i")) {
          pstr = "%" + fmt;
        } else {
          pstr = "%." + num2string(fixp) + fmt;
          if (mat[i][j] == 0) { pstr = "%1." + fmt;}
        }
        fprintf(fid, pstr.c_str(), mat[i][j]);
        if (j < ncol-1) { fprintf(fid, "%s", delimiter.c_str()); }
        else { fprintf(fid, "\n"); }
      }
      if ((type == 2 || type == 4) && mat[i][j] != 0) {
        pstr = "%i" + delimiter + "%i" + delimiter;
        if (!fmt.compare("i")) {
          pstr += "%" + fmt;
        } else {
          pstr += "%." + num2string(fixp) + fmt;
        }        
        pstr += "\n";
        if (type == 2) {
          int idx_1 = i+1;
          int idx_2 = j+1;
          if (xlabel != NULL) { idx_1 = xlabel[i]; }
          if (ylabel != NULL) { idx_2 = ylabel[j]; }
          fprintf(fid, pstr.c_str(), idx_1, idx_2, mat[i][j]);
        }
        if (type == 4) { 
          fprintf(fid, pstr.c_str(), i, j, mat[i][j]);
        }
      }
    }
  }
  fprintf(fid, "\n");

  fclose(fid);
}

template <typename Dtype> 
void write_matrix_w_index(string out, 
                          Dtype** mat, 
                          int num[], 
                          vector<string> ytickers, 
                          vector<string> xtickers, 
                          bool yflag,  
                          bool gfflag,  
                          vector<string> glabels, 
                          vector<string> flabels, 
                          bool copyright,
                          int psize, 
                          // string str, 
                          string delimiter, 
                          string postfix) {
  int nrow = num[0];
  int ncol = num[1];
  int nint = num[2];  // How many columns will be written in INT
  int fixp = num[3];  // Precision 
  out += postfix;
  FILE* fid = fopen(out.c_str(), "w");
  if (xtickers.size() > 0) {
    for (int j = 0; j < xtickers.size(); j++) {
      fprintf(fid, "%s", xtickers[j].c_str());
      if (j < xtickers.size()-1) { fprintf(fid, "%s", delimiter.c_str()); }
      else { fprintf(fid, "\n"); }
    }
  }

  for (int i = 0; i < nrow; i++) {
    if (yflag && ytickers.size() > 0) {
      fprintf(fid, "%s,", ytickers[i].c_str());
    }
    for (int j = 0; j < ncol; j++) {
      if (j < nint && nint > 0) {
        fprintf(fid, "%i", (int)mat[i][j]);
      } else {
        string pstr = "%." + num2string(fixp) + "f";
        if (mat[i][j] == 0) { pstr = "%1.f";}
        fprintf(fid, pstr.c_str(), mat[i][j]);
      }
      if (j < ncol-1) {
        fprintf(fid, ",");
      }
    }
    if (psize != -1) { fprintf(fid, ",%i,", psize);}
    if (gfflag) {
      // fprintf(fid, ",");
      if (!glabels.empty()) {
        fprintf(fid, "%s,", glabels[i].c_str());
      } else {
        fprintf(fid, "Data,");
      }
      if (!flabels.empty()) {
        fprintf(fid, "\"%s\"\n", flabels[i].c_str());
      } else {
        fprintf(fid, "\"NULL\"\n");
      }
    } else {
      fprintf(fid, "\n");
    }
  }
  fclose(fid);
}




} // namespace leri
#endif // LERI_FILEUTILS_H_

