/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_MALLOC_H_
#define LERI_MALLOC_H_

#include "leri/common.h"
#include "leri/util/gpu_detector.h"

namespace leri {

inline static void* vec_malloc(size_t size) {
  void * memblock = malloc(size);
  if (memblock) {
    memset(memblock, 0, size);
  }
  return memblock;
}

template <typename Dtype> 
Dtype* create_vec_malloc(int ncol, Dtype val = 0) {
  Dtype* vect = new Dtype [ncol];
  for (int i = 0; i < ncol; i++) { vect[i] = val; }
  return vect;
}
template <typename Dtype> 
Dtype* create_str_vec_malloc(int ncol, string val = "") {
  string* vect = new Dtype [ncol];
  for (int i = 0; i < ncol; i++) { vect[i] = val; }
  return vect;
}
template <typename Dtype> 
void free_vec_malloc(Dtype* vect) { 
  delete [] vect;
}
template <typename Dtype> 
Dtype** create_eye_mtx_malloc(int nrow, int ncol) {
  Dtype** mtx = new Dtype* [nrow];
  for (int i = 0; i < nrow; i++) {
    mtx[i] = new Dtype [ncol];
    for (int j = 0; j < ncol; j++) {
      if (i == j) {
        mtx[i][j] = 1.0; 
      } else {
        mtx[i][j] = 0.0; 
      }
    }
  }
  return mtx;
}
template <typename Dtype> 
Dtype** create_mtx_malloc(int nrow, int ncol, Dtype val = 0) {
  Dtype** mtx = new Dtype* [nrow];
  for (int i = 0; i < nrow; i++) {
    mtx[i] = new Dtype [ncol];
    for (int j = 0; j < ncol; j++) { 
      mtx[i][j] = val; 
    }
  }
  //// As follows, it is continuous, may be better
  // mtx[0] = new Dtype[nrow * ncol];
  // for (int i = 1; i < nrow; ++i) {
  //   mtx[i] = mtx[i-1] + ncol;
  // }
  // #ifdef USE_OPENMP
  // #pragma omp parallel for
  // #endif
  // for (int i = 0; i < nrow; i++) {
  //   for (int j = 0; j < ncol; j++) {
  //     mtx[i][j] = val;
  //   }
  // }
  return mtx;
}
template <typename Dtype> 
string** create_str_mtx_malloc(int nrow, int ncol, string val = "") {
  string** mtx = new string* [nrow];
  for(int i = 0; i < nrow; i++){
    mtx[i] = new string [ncol];
    for(int j = 0; j < ncol; j++){
      mtx[i][j] = val;
    }
  }
  return mtx;
}
template <typename Dtype> 
void free_mtx_malloc(Dtype** mtx, int nrow) {
  for (int i = 0; i < nrow; i++) {
    delete [] mtx[i];
  }
  delete [] mtx;
}
template <typename Dtype> 
Dtype*** create_3D_mtx_malloc(int n1, int n2, int n3, Dtype val = 0) {
  Dtype*** mtx = new Dtype** [n1];
  for (int i = 0; i < n1; i++) {
    mtx[i] = new Dtype* [n2];
    for (int j = 0; j < n2; j++) {
      mtx[i][j] = new Dtype [n3];
      for (int k = 0; k < n3; k++) {
        mtx[i][j][k] = val;
      }
    }
  }
  return mtx;
}
template <typename Dtype> 
void free_3D_mtx_malloc(Dtype*** mtx, int n1, int n2) {
  for (int i = 0; i < n1; i++) {
    for (int j = 0; j < n2; j++) {
      delete [] mtx[i][j];
    }
    delete [] mtx[i];
  }
  delete [] mtx;
}
template <typename Dtype> 
Dtype**** create_4D_mtx_malloc(int n1, int n2, int n3 = -1, int n4 = -1, Dtype val = 0) {
  Dtype**** mtx = new Dtype*** [n1];
  for (int i = 0; i < n1; i++) {
    mtx[i] = new Dtype** [n2];
    for (int j = 0; j < n2; j++) {
      if (n3 != -1) {
        mtx[i][j] = new Dtype* [n3];
        for (int k = 0; k < n3; k++) {
          mtx[i][j][k] = new Dtype [n4];
          for (int t = 0; t < n4; t++) {
            mtx[i][j][k][t] = val;
          }
        }
      } else {
        mtx[i][j] = NULL;
      }
    }
  }
  return mtx;
}

template <typename Dtype> 
void free_4D_mtx_malloc(Dtype**** mtx, int n1, int n2, int n3 = -1) {
  for (int i = 0; i < n1; i++) {
    for (int j = 0; j < n2; j++) {
      if (n3 != -1) {
        for (int k = 0; k < n3; k++) {
          delete [] mtx[i][j][k];
        }
        delete [] mtx[i][j];
      } else {
        mtx[i][j] = NULL;
      }
    }
    delete [] mtx[i];
  }
  delete [] mtx;
} 

template <typename Dtype> 
Dtype***** create_5D_mtx_malloc(int n1, int n2, int n3 = -1, int n4 = -1, int n5 = -1, Dtype val = 0) {
  Dtype***** mtx = new Dtype**** [n1];
  for (int i = 0; i < n1; i++) {
    mtx[i] = new Dtype*** [n2];
    for (int j = 0; j < n2; j++) {
      if (n3 != -1) {
        mtx[i][j] = new Dtype** [n3];
        for (int k = 0; k < n3; k++) {
          mtx[i][j][k] = new Dtype* [n4];
          for (int t = 0; t < n4; t++) {
            mtx[i][j][k][t] = new Dtype [n5];
            for (int p = 0; p < n5; p++) {
              mtx[i][j][k][t][p] = val;
            }
          }
        }
      } else {
        mtx[i][j] = NULL;
      }
    }
  }
  return mtx;
}

template <typename Dtype> 
void free_5D_mtx_malloc(Dtype**** mtx, int n1, int n2, int n3 = -1, int n4 = -1) {
  for (int i = 0; i < n1; i++) {
    for (int j = 0; j < n2; j++) {
      if (n3 != -1) {
        for (int k = 0; k < n3; k++) {
          for (int t = 0; t < n4; t++) {
            delete [] mtx[i][j][k][t];
          }
          delete [] mtx[i][j][k];
        }
        delete [] mtx[i][j];
      } else {
        mtx[i][j] = NULL;
      }
    }
    delete [] mtx[i];
  }
  delete [] mtx;
} 


/**
 *  GPU-based methods
 */
#ifdef USE_GPU


template <typename Dtype> 
Dtype* create_vec_cudamalloc(int n) {
  Dtype* vec_dev;
  cudaMalloc((void **)&vec_dev, n*sizeof(Dtype));
  return vec_dev;
}
template <typename Dtype> 
void host2dev_vec_memcpy(Dtype* vec_dev, Dtype* vec_host, int n) {
  cudaMemcpy(vec_dev, vec_host, n*sizeof(Dtype), cudaMemcpyHostToDevice);
}
template <typename Dtype> 
void dev2host_vec_memcpy(Dtype* vec_host, Dtype* vec_dev, int n) {
  cudaMemcpy(vec_host, vec_dev, n*sizeof(Dtype), cudaMemcpyDeviceToHost);
}
template <typename Dtype> 
void free_vec_cudamalloc(Dtype* vec_dev) {
  cudaFree(vec_dev);
}




#endif // USE_GPU























}// namespace leri
#endif // LERI_MALLOC_H_
