/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */
 
#ifndef LERI_MATH_H_
#define LERI_MATH_H_

#include "leri/common.h"
#include "leri/util/malloc.h"
#include "leri/util/matrix.h"

namespace leri {

unsigned n_choose_k(unsigned n, unsigned k);
void n_choose_k_elements(int n, int k);

int generate_all_combinations(int *a,int n, int m);
void get_all_combinations(int** mtx, const vector<int> vec, int n, int m);

template <typename Dtype>
static inline Dtype sign(Dtype v) {
  // Dtype val = 0;
  // if (v > 0) { val = 1; }
  // else if (v < 0) { val = -1; }
  // else { val = 0.0; }
  return (v > 0) ? 1 : ((v < 0) ? -1 : 0);
}

template <typename Dtype>
Dtype mod(Dtype y, Dtype x) {
  if(x <= 0) {
    print_info("Warning: in mod(Dtype, Dtype)", YELLOW);
  }
  return (y/x - floor(y/x))*x;
}
template <typename Dtype>
void wrap(Dtype& x, Dtype min, Dtype max) {
  x = mod(x - min, max - min) + min;
}

template <typename Dtype>
static inline Dtype logistic(Dtype x) {
  return ((Dtype)1/((Dtype)1+exp(-(x))));
}
template <typename Dtype>
static inline Dtype anti_logistic(Dtype x) {
  return -log((Dtype)1/x-1);
}
template <typename Dtype>
Dtype logsumexp(Dtype* logv, int n) {
  if(n <= 1) {
    print_info("Warning: in logsumexp", WARNING);
  }
  Dtype max = logv[0];
  for (int i=1; i<n; i++) {
    if (logv[i] > max) {
      max = logv[i];
    }
  }
  Dtype answer = 0;
  // log(sum(exp(logf))   = log(sum(exp(logf - max(logf) + max(logf)))
  //      = max(logf) + log(sum(exp(logf - max(logf)))
  for(int i=0; i<n; i++)
    answer += exp(logv[i] - max);
  answer = max + log(answer);
  return answer;
}

template <typename Dtype>
Dtype logsumexp(const vector<Dtype>& logv) {
  int n = static_cast<int>(logv.size());
  //if(n<=1)
  //  cout<<"Warning in logsumexp"<<endl;
  Dtype max = *max_element(logv.begin(), logv.end());
  Dtype answer = 0;
  // log(sum(exp(logf))   = log(sum(exp(logf - max(logf) + max(logf)))
  //      = max(logf) + log(sum(exp(logf - max(logf)))
  for(int i=0; i<n; i++)
    answer += exp(logv[i] - max);
  answer = max + log(answer);
  return answer;
}

template <typename Dtype>
Dtype logsumexp(Dtype a, Dtype b) {
  Dtype x[2] = {a,b};
  return logsumexp(x, 2);
}
template <typename Dtype>
Dtype logdiffexp(Dtype a, Dtype b) {
  if(a <= b) {
    print_info("in logdiffexp!", ERROR_OTHER);
    exit (EXIT_FAILURE);
  }
  Dtype biggest = a;
  a -= biggest;
  b -= biggest;
  return log(exp(a) - exp(b)) + biggest;
}



/**
 * Compute cosine of angle of three vectors
 */
template <typename Dtype>
static inline void vec_cosine_angle(Dtype* val, Dtype* vect_a, Dtype* vect_b, int n) {
  Dtype a = vec_dot_prod(vect_a, vect_a, n);
  Dtype b = vec_dot_prod(vect_b, vect_b, n);
  Dtype c = vec_dot_prod(vect_a, vect_b, n);
  *val    = c / (sqrt(a*b)+DBL_EPSILON);
}
/**
 * Given three vectors a, b, and c, it makes vectors ba-> and bc-> and calculates the angle between them
 */
template <typename Dtype>
static inline void three_vec_angle(Dtype* angle, Dtype* vect_a, Dtype* vect_b, Dtype* vect_c, int dim = 3) {
  Dtype* vect_1 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_2 = create_vec_malloc<Dtype>(dim);
  vec_sub<Dtype>(vect_1, vect_b, vect_a, dim);
  vec_sub<Dtype>(vect_2, vect_b, vect_c, dim);
  vec_cosine_angle(angle, vect_1, vect_2, dim);
  if (vect_1 != NULL) { free_vec_malloc(vect_1); }
  if (vect_2 != NULL) { free_vec_malloc(vect_2); }
}
template <typename Dtype>
static inline void angle_four_points(Dtype* angle, Dtype* vect_ha, Dtype* vect_ta, Dtype* vect_hb, Dtype* vect_tb, int dim = 3) {
  Dtype* vect_1 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_2 = create_vec_malloc<Dtype>(dim);
  vec_sub<Dtype>(vect_1, vect_ha, vect_ta, dim);
  vec_sub<Dtype>(vect_2, vect_hb, vect_tb, dim);
  vec_cosine_angle(angle, vect_1, vect_2, dim);
  if (vect_1 != NULL) { free_vec_malloc(vect_1); }
  if (vect_2 != NULL) { free_vec_malloc(vect_2); }
}
// TODO: 
template <typename Dtype>
static inline Dtype angle_four_points(Dtype* vect_ha, Dtype* vect_ta, Dtype* vect_hb, Dtype* vect_tb, int dim = 3) {
  Dtype val = 0.0;
  Dtype* vect_1 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_2 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_3 = create_vec_malloc<Dtype>(dim);
  vec_sub<Dtype>(vect_1, vect_ha, vect_ta, dim);
  vec_sub<Dtype>(vect_2, vect_ta, vect_hb, dim);
  vec_sub<Dtype>(vect_3, vect_hb, vect_tb, dim);
  Dtype* vect_4 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_5 = create_vec_malloc<Dtype>(dim);
  Dtype* vect_6 = create_vec_malloc<Dtype>(dim);
  vec_cross_prod(vect_4, vect_2, vect_1);
  vec_cross_prod(vect_5, vect_3, vect_2);
  vec_cross_prod(vect_6, vect_5, vect_4);
  vec_cosine_angle(&val, vect_4, vect_5, dim);
  val = acos(val)*RADIANS_TO_DEGREES;
  val = sign(vec_dot_prod(vect_6, vect_2, dim))*val;
  if (vect_1 != NULL) { free_vec_malloc(vect_1); }
  if (vect_2 != NULL) { free_vec_malloc(vect_2); }
  if (vect_3 != NULL) { free_vec_malloc(vect_3); }
  if (vect_4 != NULL) { free_vec_malloc(vect_4); }
  if (vect_5 != NULL) { free_vec_malloc(vect_5); }
  if (vect_6 != NULL) { free_vec_malloc(vect_6); }
  return val;
}

template <typename Dtype>
void internal2cartesian(Dtype* cc, Dtype*c1, Dtype* c2, Dtype* c3, Dtype* p, int n = 3) { 
  Dtype* vec  = create_vec_malloc<Dtype>(n);
  Dtype** mtx = create_mtx_malloc<Dtype>(n, n);
  vec[0] = p[0]*cos(DEGREES_TO_RADIANS*p[1]);
  vec[1] = p[0]*cos(DEGREES_TO_RADIANS*p[2])*sin(DEGREES_TO_RADIANS*p[1]);
  vec[2] = p[0]*sin(DEGREES_TO_RADIANS*p[2])*sin(DEGREES_TO_RADIANS*p[1]);
  Dtype* vec_a = create_vec_malloc<Dtype>(n);
  Dtype* vec_b = create_vec_malloc<Dtype>(n);
  Dtype* vec_c = create_vec_malloc<Dtype>(n);
  vec_sub(vec_a, c2, c1, n);
  vec_sub(vec_b, c3, c2, n);
  vec_normalize(vec_b, n);
  vec_cross_prod(vec_c, vec_a, vec_b);
  vec_normalize(vec_c, n);
  vec_cross_prod(vec_a, vec_c, vec_b);
  for (int i = 0; i < n; i++) {
    mtx[i][0] = -vec_b[i];
    mtx[i][1] =  vec_a[i];
    mtx[i][2] =  vec_c[i];
  }
  mat_x_vec(vec_b, mtx, vec, n);
  vec_add(cc, c3, vec_b, n);
  if (vec   != NULL) { free_vec_malloc(vec); }
  if (vec_a != NULL) { free_vec_malloc(vec_a); }
  if (vec_b != NULL) { free_vec_malloc(vec_b); }
  if (vec_c != NULL) { free_vec_malloc(vec_c); }
  if (mtx   != NULL) { free_mtx_malloc(mtx, n); }
}

template <typename Dtype>
void internal2cartesian(Dtype* coord, Dtype* p1, Dtype dist, Dtype* p2, Dtype angle, Dtype* p3, Dtype torsion) {
  /* Compute coordinates for vector v, given its 'dis'tance
     from vector v1, 'ang'le v-v1-v2 and 'tors'ion v-v1-v2-v3.'
     Both ang and tors should be in degrees.
  */
  angle *= DEGREES_TO_RADIANS;
  Dtype sina = sin(angle);
  Dtype cosa = cos(angle);
  torsion *= DEGREES_TO_RADIANS;
  Dtype sint = sina * sin(torsion);
  Dtype cost = sina * cos(torsion);
  Dtype x1 = p1[0], y1 = p1[1], z1 = p1[2];
  Dtype x2 = p2[0], y2 = p2[1], z2 = p2[2];
  Dtype x3 = p3[0], y3 = p3[1], z3 = p3[2];

  Dtype u1x = x2 - x3;
  Dtype u1y = y2 - y3;
  Dtype u1z = z2 - z3;
  Dtype d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z);
  u1x = u1x*d;
  u1y = u1y*d;
  u1z = u1z*d;

  Dtype u2x = x1 - x2;
  Dtype u2y = y1 - y2;
  Dtype u2z = z1 - z2;
  d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z);
  u2x = u2x*d;
  u2y = u2y*d;
  u2z = u2z*d;

  Dtype cosine = u1x*u2x + u1y*u2y + u1z*u2z;
  Dtype sine;
  if (abs(cosine) < 1.0) {
    sine = 1.0/sqrt(1.0 - cosine*cosine);
  } else {
    sine = 1.0/sqrt(cosine*cosine - 1.0);
  }
  Dtype u3x = sine * (u1y*u2z - u1z*u2y);
  Dtype u3y = sine * (u1z*u2x - u1x*u2z);
  Dtype u3z = sine * (u1x*u2y - u1y*u2x);
  Dtype u4x = cost * (u3y*u2z - u3z*u2y);
  Dtype u4y = cost * (u3z*u2x - u3x*u2z);
  Dtype u4z = cost * (u3x*u2y - u3y*u2x);

  coord[0] = x1 + dist*(-u2x*cosa + u4x + u3x*sint);
  coord[1] = y1 + dist*(-u2y*cosa + u4y + u3y*sint);
  coord[2] = z1 + dist*(-u2z*cosa + u4z + u3z*sint);
}

/**
 * @param a  Real symmetric matrix
 * @param e  Eigenvalue
 * @param ev Each column is an eigenvector
 */
template <typename Dtype>
void jacobi(Dtype* covar, Dtype* e, Dtype* ev, int n) {
  int dim = n;
  vec_copy(e,  covar, n*n);
  vec_copy(ev, covar, n*n);
  // the identity matrix ev
  for (int j = 0; j < dim; j++) {
    for (int i = 0; i < dim; i++) {
      ev[i + j * dim] = (i == j) ? 1.0f : 0.0f;
    }
  }
  int loop_count = 0;
  while (true) {
    // Find the element with the largest absolute value among the off-diagonal components
    int   p = 0;
    int   q = 0;
    Dtype m = 0;
    for (int r = 0; r < dim; r++) {
      for (int c = 0; c < dim; c++) {
        if (r >= c) continue;
        Dtype val = abs(e[c + r * dim]);
        if (val > m) {
          p = r;
          q = c;
          m = val;
        }
      }
    }
    // Terminate when the element with the largest absolute value is within the tolerance
    const Dtype eps = 0.001f;
    if (m <= eps) { break; }
    // End if fall into an infinite loop
    if (++loop_count > 1000) { break; }
    Dtype app = e[p + p * dim];
    Dtype aqq = e[q + q * dim];
    Dtype apq = e[q + p * dim];

    Dtype alpha = (app - aqq) / 2.0f;
    Dtype beta  = -apq;
    Dtype gamma = abs(alpha) / sqrt(alpha * alpha + beta * beta);
    Dtype ct = sqrt((1.0f + gamma) / 2.0f);
    Dtype st = sqrt((1.0f - gamma) / 2.0f);
    if (alpha * beta < 0) st = -st;
    for (int r = 0; r < dim; r++) {
      Dtype tmp_p = ct * e[p + r * dim] - st * e[q + r * dim];
      Dtype tmp_q = st * e[p + r * dim] + ct * e[q + r * dim];
      e[p + r * dim] = tmp_p;
      e[q + r * dim] = tmp_q;
    }
    // Make e symmetric
    for (int r = 0; r < dim; r++) {
      e[r + p * dim] = e[p + r * dim];
      e[r + q * dim] = e[q + r * dim];
    }

    // Update pp, qq, pq, qp
    e[p + p * dim] = ct * ct * app + st * st * aqq - 2 * st * ct * apq;
    e[q + q * dim] = st * st * app + ct * ct * aqq + 2 * st * ct * apq;
    e[p + q * dim] = ct * st * (app - aqq) + (ct * ct - st * st) * apq;
    e[q + p * dim] = ct * st * (app - aqq) + (ct * ct - st * st) * apq;

    // Update diagonalization matrix
    for (int r = 0; r < dim; r++) {
      Dtype tmp_p = ct * ev[p + r * dim] - st * ev[q + r * dim];
      Dtype tmp_q = st * ev[p + r * dim] + ct * ev[q + r * dim];
      ev[p + r * dim] = tmp_p;
      ev[q + r * dim] = tmp_q;
    }
  }
  // Since the diagonal component of e is an eigenvalue, cut out only that
  for (int i = 0; i < dim; i++) {
    e[i] = e[i + i * dim];
  }
}

/**
 * Modified from https://people.sc.fsu.edu/~jburkardt/cpp_src/jacobi_eigenvalue/jacobi_eigenvalue.html
 * This function computes the eigenvalues and eigenvectors of a
 *  real symmetric matrix, using Rutishauser's modfications of the classical
 *  Jacobi rotation method with threshold pivoting. 
 */
template <typename Dtype>
void eig_calc(Dtype** eig_vec, Dtype* eig_val, int n, Dtype** mtx, int max_iter = 1000) {
  Dtype*  bw = create_vec_malloc<Dtype>(n);
  Dtype*  zw = create_vec_malloc<Dtype>(n);
  Dtype t, c, s, g, h;
  Dtype tau, theta, thresh;
  Dtype gap_i, term, term_k, term_i;

  for (int i = 0; i < n; i++) {
    eig_val[i] = mtx[i][i];
    bw[i] = eig_val[i];
  }
  int cnt = 0;
  // int rot_num = 0;

  while (cnt < max_iter) {
    cnt += 1;
    thresh = 0.0; // from upper triangle of the matrix
    for (int j = 0; j < n; j++) {
      for (int i = 0; i < j; i++) {
        thresh = thresh + mtx[j][i]*mtx[j][i];
      }
    }
    thresh = sqrt(thresh)/(Dtype)(4*n);
    if (thresh == 0.0) { break; }

    for (int k = 0; k < n; k++) {
      for (int i = k + 1; i < n; i++) {
        gap_i  = 10.0 * fabs (mtx[i][k]);
        term_k = gap_i + fabs (eig_val[k]);
        term_i = gap_i + fabs (eig_val[i]);
        //  Annihilate tiny offdiagonal elements.
        if (4 < cnt && term_k == fabs (eig_val[k]) && term_i == fabs (eig_val[i])) {
          mtx[i][k] = 0.0;
        } else if (thresh <= fabs (mtx[i][k])) {
          h = eig_val[i] - eig_val[k];
          term = fabs (h) + gap_i;
          if (term == fabs (h)) {
            t = mtx[i][k] / h;
          } else {
            theta = 0.5 * h / mtx[i][k];
            t = 1.0 / (fabs (theta) + sqrt (1.0 + theta * theta));
            if (theta < 0.0) {
              t = - t;
            }
          }
          c = 1.0 / sqrt (1.0 + t * t);
          s = t * c;
          tau = s / (1.0 + c);
          h = t * mtx[i][k];
          //  Accumulate corrections to diagonal elements.
          zw[k] = zw[k] - h;                 
          zw[i] = zw[i] + h;
          eig_val[k] = eig_val[k] - h;
          eig_val[i] = eig_val[i] + h;
          mtx[i][k] = 0.0;
          // Rotate using information from the upper triangle of the matrix only.
          for (int j = 0; j < k; j++) {
            g = mtx[k][j];
            h = mtx[i][j];
            mtx[k][j] = g - s * (h + g * tau);
            mtx[i][j] = h + s * (g - h * tau);
          }
          for (int j = k + 1; j < i; j++) {
            g = mtx[j][k];
            h = mtx[i][j];
            mtx[j][k] = g - s * (h + g * tau);
            mtx[i][j] = h + s * (g - h * tau);
          }
          for (int j = i + 1; j < n; j++) {
            g = mtx[j][k];
            h = mtx[j][i];
            mtx[j][k] = g - s * (h + g * tau);
            mtx[j][i] = h + s * (g - h * tau);
          }
          //  Accumulate information in the eigenvector matrix.
          for (int j = 0; j < n; j++) {
            g = eig_vec[j][k];// eig_vec[k][j];
            h = eig_vec[j][i];// eig_vec[i][j];
            eig_vec[j][k] = g - s * (h + g * tau);
            eig_vec[j][i] = h + s * (g - h * tau);
          }
          // rot_num = rot_num + 1;
        }
      }
    }

    for (int i = 0; i < n; i++) {
      bw[i] = bw[i] + zw[i];
      eig_val[i] = bw[i];
      zw[i] = 0.0;
    }
  }
  for (int j = 0; j < n; j++) {
    for (int i = 0; i < j; i++) {
      mtx[j][i] = mtx[i][j];
    }
  }
  //  Descending sort the eigenvalues and eigenvectors.
  for (int k = 0; k < n - 1; k++) {
    int m = k;
    for (int i = k + 1; i < n; i++) {
      if (eig_val[i] > eig_val[m]) {
        m = i;
      }
    }
    if (m != k) {
      t    = eig_val[m];
      eig_val[m] = eig_val[k];
      eig_val[k] = t;
      for (int i = 0; i < n; i++){
        t        = eig_vec[i][m];// [m][i];
        eig_vec[i][m] = eig_vec[i][k];
        eig_vec[i][k] = t;
      }
    }
  }
  if (bw != NULL) { free_vec_malloc(bw); }
  if (zw != NULL) { free_vec_malloc(zw); }
}



}  // namespace leri
#endif // LERI_MATH_H_
