/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_MATRIX_H_
#define LERI_MATRIX_H_

#include "leri/common.h"
#include "leri/util/malloc.h"

namespace leri {

// TODO:
template <typename Dtype>
void mat2sincosmat(Dtype** mtx, Dtype** mat, int nrow, int ncol) {
  for (int i = 0; i < nrow; i++) {
    for (int j = 0; j < ncol; j++) {
      mtx[j][j*2+0] = sin(mat[i][j]);
      mtx[j][j*2+1] = cos(mat[i][j]);
    }
  }
}




/**
 * Methods on vector
 */
template <typename Dtype> 
void vec_init(Dtype* vec, int n, Dtype val = 0, bool rand_flag = false) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; i++) {
    if (rand_flag) {
      vec[i] = get_frandom<Dtype>();
    } else {
      vec[i] = val;
    }
  }
}
template <typename Dtype> 
int vec_find(const vector<Dtype> vec, Dtype key) {
  int idx = -1;
  for (int i = 0; i < vec.size(); i++) {
    if (vec[i] == key) {
      idx = i;
      break;
    }
  }
  return idx;
}
template <typename Dtype> 
int vec_min_arg(const Dtype* vec, const int n) {
  int idx = 0;
  for (int i = 0; i < n; i++) {
    if(vec[idx] > vec[i]) { idx = i; }
  }
  return idx;
}
template <typename Dtype> 
int vec_max_arg(const Dtype* vec, const int n) {
  int idx = 0;
  for (int i = 0; i < n; i++) {
    if(vec[idx] < vec[i]) { idx = i; }
  }
  return idx;
}

// template <typename Dtype> 
// void vec_scale(vector<Dtype>& vec, Dtype s) {
//   int n = vec.size();
//   for(int i = 0; i < n; i++) { vec[i] *= s; }
// }

template <typename Dtype> 
void vec_boundary(Dtype* vec, int n, Dtype s = 1e20) {
  // int n = vec.size();
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { 
    if (vec[i] > s) {
      vec[i] = s;
    }
    if (vec[i] < -s) {
      vec[i] = -s;
    }
    // vec[i] *= s; 
  }
}
template <typename Dtype> 
void vec_scale(Dtype* vec, int n, Dtype s = 1) {
  // int n = vec.size();
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { vec[i] *= s; }
}
template <typename Dtype> 
void vec_scale_inv_sqrt(Dtype* vec, int n) {
  for(int i = 0; i < n; i++) { vec[i] = (Dtype)1/sqrt(vec[i]); }
}
template <typename Dtype> 
void vec_scale_inv(Dtype* vec, int n, Dtype s = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { vec[i] = s*1/vec[i]; }
}
template <typename Dtype> 
void vec_copy(Dtype* vec_a, const Dtype* vec_b, int n, Dtype s = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { vec_a[i] = s*vec_b[i]; }
}
template <typename Dtype> 
void vec_copy(Dtype* vec_a, const vector<Dtype>& vec_b, int n, Dtype s = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { vec_a[i] = s*vec_b[i]; }
}
template <typename Dtype> 
void vec_copy_back(vector<Dtype>& vec_a, const Dtype* vec_b, int n, Dtype s = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { vec_a[i] = s*vec_b[i]; }
}
template <typename string> 
void vec_copy(vector<string>& vec_a, const vector<string>& vec_b) {
  int n = vec_b.size();
  if (n > 0) {
    vec_a.reserve(n);
    for(int i = 0; i < n; i++) { vec_a[i] = vec_b[i]; }
  }
}
// // TODO: remove
// template <typename Dtype> 
// void vec_copy(vector<Dtype>& vec_a, const vector<Dtype>& vec_b, Dtype s = 1) {
//   int n = vec_a.size();
//   for(int i = 0; i < n; i++) { vec_a[i] = s*vec_b[i]; }
// }
template <typename Dtype>
void mat2vec_copy(Dtype* vec, const vector<vector<Dtype> >& mat, int m, int n) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      vec[i*n+j] = mat[i][j];
    }
  }
}
template <typename Dtype> 
void vec_sum(Dtype* val, const Dtype* vec, int n, Dtype scale = 1) {
  *val = 0;
  for(int i = 0; i < n; i++) { *val += scale*vec[i]; }
}
template <typename Dtype> 
void vec_add(Dtype* vec, const Dtype* vec_a, const Dtype* vec_b, int n, Dtype a = 1, Dtype b = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { 
    vec[i] = a*vec_a[i] + b*vec_b[i]; 
  }
}
template <typename Dtype> 
void vec_add(Dtype* vec, const Dtype* vec_a, int n, Dtype a = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for(int i = 0; i < n; i++) { 
    vec[i] += a*vec_a[i]; 
  }
}

// TODO: remove
template <typename Dtype> 
void vec_add(vector<Dtype>& vec, vector<Dtype>& vec_a, vector<Dtype>& vec_b, Dtype factor_a, Dtype factor_b){
  int n = vec_a.size();
  for(int i = 0; i < n; i++) {
    vec[i] = factor_a*vec_a[i]+factor_b*vec_b[i];
  }
}
// template <typename Dtype> 
// void vec_subtract(Dtype* vec, Dtype* vec_a, Dtype* vec_b, int n, Dtype a = 1, Dtype b = 1) {
//   #ifdef USE_OPENMP
//   #pragma omp parallel for
//   #endif
//   for(int i = 0; i < n; i++) { 
//     vec[i] = a*vec_a[i] - b*vec_b[i]; 
//   }
// }
template <typename Dtype>
static inline void vec_sub(Dtype* vec, Dtype* vec_a, Dtype* vec_b, int n, Dtype a = 1, Dtype b = 1) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; i++) {
    vec[i] = a*vec_a[i] - b*vec_b[i];
  }
}
// TODO: delete
template <typename Dtype> 
void vec_dot(Dtype* val, vector<Dtype>& vec_a, vector<Dtype>& vec_b) {
  int n = vec_a.size();
  *val = 0;
  for(int i = 0; i < n; i++) { *val += vec_a[i] * vec_b[i]; }
}
// template <typename Dtype>
// Dtype vec_dot_prod(const Dtype* vect_a, const Dtype* vect_b, int n) {
//   Dtype val = 0;
//   #ifdef USE_OPENMP
//   #pragma omp parallel for reduction (+:val)
//   #endif
//   for (int i = 0; i < n; i++) { 
//     val += vect_a[i]*vect_b[i];
//   }
//   return val;
// }
template <typename Dtype>
inline Dtype vec_dot_prod(const Dtype* x1, const Dtype* x2, const int n) {
  switch (n) {
    case 1: return x1[0] * x2[0];
    case 2: return x1[0] * x2[0] + x1[1] * x2[1];
    case 3: return x1[0] * x2[0] + x1[1] * x2[1] + x1[2] * x2[2];
    case 4: return x1[0] * x2[0] + x1[1] * x2[1] + x1[2] * x2[2] + x1[3] * x2[3];
    case 5: return x1[0] * x2[0] + x1[1] * x2[1] + x1[2] * x2[2] + x1[3] * x2[3] + x1[4] * x2[4];
  default:
    Dtype val = 0;
    #ifdef USE_OPENMP
    #pragma omp parallel for reduction (+:val)
    #endif
    for (int i = 0; i < n; i++) {
      val += x1[i] * x2[i];
    }
    return val;
  };
}



template <typename Dtype>
inline Dtype unwrap_2d_dot(const Dtype* x1, const Dtype* x2, const int n, int stride1, int stride2) {  
  Dtype val = 0;
  for(int j = 0; j < n; j++) {
    val += vec_dot_prod<Dtype>(&x1[stride1*j], &x2[stride2*j], n);
  }
  return val;
}
template <typename Dtype>
inline Dtype dot_rot180(const Dtype* x1, const Dtype* x2, const int n) {  
  switch(n) {
    case 1:  return x1[0]*x2[0]; 
    case 2:  return x1[0]*x2[1]+x1[1]*x2[0]; 
    case 3:  return x1[0]*x2[2]+x1[1]*x2[1]+x1[2]*x2[0]; 
    case 4:  return x1[0]*x2[3]+x1[1]*x2[2]+x1[2]*x2[1]+x1[3]*x2[0]; 
    case 5:  return x1[0]*x2[4]+x1[1]*x2[3]+x1[2]*x2[2]+x1[3]*x2[1]+x1[4]*x2[0]; 
  default: 
    Dtype val = 0;
    for(int i = 0; i < n; i++) {
      val += x1[i]*x2[n-i-1];
    }
    return val;  
  };
}
template <typename Dtype>
inline Dtype unwrap_2d_dot_rot180(const Dtype* x1, const Dtype* x2, const int n, int stride1, int stride2) {  
  Dtype val = 0;  
  for(int i = 0; i < n; i++) {
    val += dot_rot180<Dtype>(&x1[stride1*i], &x2[stride2*(n-i-1)], n);
  }
  return val;
}
template <typename Dtype>
inline void unwrap_aligned_NxN(const int N, Dtype* aligned_out,  const Dtype* in, const int in_size, const int stride = 1) {
  const int node_size = (in_size - N)/stride + 1;
  int c1 = 0;
  int off = 0;
  const int inc_off = N*N*8;

  for (int j = 0; j < node_size; j += 1) { // height
    for (int i = 0; i < node_size; i += 1) { // width
      const Dtype* tn = in + j*in_size + i;
      if(N == 5) {
        for (int k = 0; k < 5; k++) {
          aligned_out[c1 + 0 +  k * 40 + off] = tn[0 + 0 + in_size*k];
          aligned_out[c1 + 8 +  k * 40 + off] = tn[0 + 1 + in_size*k];
          aligned_out[c1 + 16 + k * 40 + off] = tn[0 + 2 + in_size*k];
          aligned_out[c1 + 24 + k * 40 + off] = tn[0 + 3 + in_size*k];
          aligned_out[c1 + 32 + k * 40 + off] = tn[0 + 4 + in_size*k];
        }
      } else if(N == 3) {
        aligned_out[c1 + off] = tn[0];
        aligned_out[c1 + 8 + off] = tn[0 + 1];
        aligned_out[c1 + 16 + off] = tn[0 + 2];

        aligned_out[c1 + 24 + off] = tn[0 +     in_size];
        aligned_out[c1 + 32 + off] = tn[0 + 1 + in_size];
        aligned_out[c1 + 40 + off] = tn[0 + 2 + in_size];

        aligned_out[c1 + 48 + off] = tn[0 +     2 * in_size];
        aligned_out[c1 + 56 + off] = tn[0 + 1 + 2 * in_size];
        aligned_out[c1 + 64 + off] = tn[0 + 2 + 2 * in_size];
      } else {
        int cnt=0;
        for (int k = 0; k < N; k++) {
          for (int m = 0; m < N; m++) {
            aligned_out[c1 + cnt*8 + off] = tn[0 + m + in_size*k];
            cnt++;
          }
        }
      }
      off++;
      if (off > 7) { off = 0; c1 += inc_off; }
    }
  }
}
template <typename Dtype>
inline void dotsum_unwrapped_NxN(const int N, const Dtype* im, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  const int NN = N*N;
  for (int j = 0; j < outsize; j += 8) {
    Dtype* c = out+j;
    for(int i = 0; i < NN; i++) { 
      const Dtype f = filter_ptr[i];
      c[0]+=im[0]*f;
      c[1]+=im[1]*f;
      c[2]+=im[2]*f;
      c[3]+=im[3]*f;
      c[4]+=im[4]*f;
      c[5]+=im[5]*f;
      c[6]+=im[6]*f;
      c[7]+=im[7]*f;
      im+=8; 
    }
  }
}

#ifdef USE_AVX
template <typename Dtype>
inline void dotsum_unwrapped_2x2(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  _mm256_zeroupper();
  const __m256 f0 = _mm256_broadcast_ss(&filter_ptr[0]); const __m256 f1 = _mm256_broadcast_ss(&filter_ptr[1]);
  const __m256 f2 = _mm256_broadcast_ss(&filter_ptr[2]); const __m256 f3 = _mm256_broadcast_ss(&filter_ptr[3]);
  for (int j = 0; j < outsize; j += 8) {
    __m256 a, c0, c1;
    // multiply filter
    a = _mm256_load_ps(_img); c0 = _mm256_mul_ps(a, f0);
    a = _mm256_load_ps(_img +  8); c1 = _mm256_mul_ps(a, f1); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 16); c1 = _mm256_mul_ps(a, f2); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 24); c1 = _mm256_mul_ps(a, f3); c0 = _mm256_add_ps(c0, c1);
    // add result to output
    a = _mm256_load_ps(out + j); 
    c0 = _mm256_add_ps(c0, a);
    _mm256_stream_ps(out + j, c0);
    _img += 32;
  }
  _mm256_zeroupper();
}
template <typename Dtype>
inline void dotsum_unwrapped_3x3(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  _mm256_zeroupper();

  const __m256 f0 = _mm256_broadcast_ss(&filter_ptr[0]); const __m256 f1 = _mm256_broadcast_ss(&filter_ptr[1]);
  const __m256 f2 = _mm256_broadcast_ss(&filter_ptr[2]); const __m256 f3 = _mm256_broadcast_ss(&filter_ptr[3]);
  const __m256 f4 = _mm256_broadcast_ss(&filter_ptr[4]); const __m256 f5 = _mm256_broadcast_ss(&filter_ptr[5]);
  const __m256 f6 = _mm256_broadcast_ss(&filter_ptr[6]); const __m256 f7 = _mm256_broadcast_ss(&filter_ptr[7]);
  const __m256 f8 = _mm256_broadcast_ss(&filter_ptr[8]);

  for (int j = 0; j < outsize; j += 8) {
    __m256 a, c0, c1;
    // multiply filter
    a = _mm256_load_ps(_img); c0 = _mm256_mul_ps(a, f0);
    a = _mm256_load_ps(_img +  8); c1 = _mm256_mul_ps(a, f1); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 16); c1 = _mm256_mul_ps(a, f2); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 24); c1 = _mm256_mul_ps(a, f3); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 32); c1 = _mm256_mul_ps(a, f4); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 40); c1 = _mm256_mul_ps(a, f5); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 48); c1 = _mm256_mul_ps(a, f6); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 56); c1 = _mm256_mul_ps(a, f7); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 64); c1 = _mm256_mul_ps(a, f8); c0 = _mm256_add_ps(c0, c1);
    // add result to output
    a = _mm256_load_ps(out + j); 
    c0 = _mm256_add_ps(c0, a);
    _mm256_stream_ps(out + j, c0);
    _img += 72;
  }
  _mm256_zeroupper();
}

template <typename Dtype>
inline void dotsum_unwrapped_4x4(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  _mm256_zeroupper();
  const __m256 f0 = _mm256_broadcast_ss(&filter_ptr[0]); const __m256 f1 = _mm256_broadcast_ss(&filter_ptr[1]);
  const __m256 f2 = _mm256_broadcast_ss(&filter_ptr[2]); const __m256 f3 = _mm256_broadcast_ss(&filter_ptr[3]);
  const __m256 f4 = _mm256_broadcast_ss(&filter_ptr[4]); const __m256 f5 = _mm256_broadcast_ss(&filter_ptr[5]);
  const __m256 f6 = _mm256_broadcast_ss(&filter_ptr[6]); const __m256 f7 = _mm256_broadcast_ss(&filter_ptr[7]);
  const __m256 f8 = _mm256_broadcast_ss(&filter_ptr[8]); const __m256 f9 = _mm256_broadcast_ss(&filter_ptr[9]);
  const __m256 f10 = _mm256_broadcast_ss(&filter_ptr[10]); const __m256 f11 = _mm256_broadcast_ss(&filter_ptr[11]);
  const __m256 f12 = _mm256_broadcast_ss(&filter_ptr[12]); const __m256 f13 = _mm256_broadcast_ss(&filter_ptr[13]);
  const __m256 f14 = _mm256_broadcast_ss(&filter_ptr[14]); const __m256 f15 = _mm256_broadcast_ss(&filter_ptr[15]);
  for (int j = 0; j < outsize; j += 8) {
    __m256 a, c0, c1;
    // multiply filter
    a = _mm256_load_ps(_img); c0 = _mm256_mul_ps(a, f0);
    a = _mm256_load_ps(_img + 8); c1 = _mm256_mul_ps(a, f1); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 16); c1 = _mm256_mul_ps(a, f2); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 24); c1 = _mm256_mul_ps(a, f3); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 32); c1 = _mm256_mul_ps(a, f4); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 40); c1 = _mm256_mul_ps(a, f5); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 48); c1 = _mm256_mul_ps(a, f6); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 56); c1 = _mm256_mul_ps(a, f7); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 64); c1 = _mm256_mul_ps(a, f8); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 72); c1 = _mm256_mul_ps(a, f9); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 80); c1 = _mm256_mul_ps(a, f10); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 88); c1 = _mm256_mul_ps(a, f11); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 96); c1 = _mm256_mul_ps(a, f12); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 104); c1 = _mm256_mul_ps(a, f13); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 112); c1 = _mm256_mul_ps(a, f14); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 120); c1 = _mm256_mul_ps(a, f15); c0 = _mm256_add_ps(c0, c1);
    // add result to output
    a = _mm256_load_ps(out + j); 
    c0 = _mm256_add_ps(c0, a);
    _mm256_stream_ps(out + j, c0);
    _img += 128;
  }
  _mm256_zeroupper();
}
template <typename Dtype>
inline void dotsum_unwrapped_5x5(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  _mm256_zeroupper();
  const __m256 f0 = _mm256_broadcast_ss(&filter_ptr[0]); const __m256 f1 = _mm256_broadcast_ss(&filter_ptr[1]);
  const __m256 f2 = _mm256_broadcast_ss(&filter_ptr[2]); const __m256 f3 = _mm256_broadcast_ss(&filter_ptr[3]);
  const __m256 f4 = _mm256_broadcast_ss(&filter_ptr[4]); const __m256 f5 = _mm256_broadcast_ss(&filter_ptr[5]);
  const __m256 f6 = _mm256_broadcast_ss(&filter_ptr[6]); const __m256 f7 = _mm256_broadcast_ss(&filter_ptr[7]);
  const __m256 f8 = _mm256_broadcast_ss(&filter_ptr[8]); const __m256 f9 = _mm256_broadcast_ss(&filter_ptr[9]);
  const __m256 f10 = _mm256_broadcast_ss(&filter_ptr[10]); const __m256 f11 = _mm256_broadcast_ss(&filter_ptr[11]);
  const __m256 f12 = _mm256_broadcast_ss(&filter_ptr[12]); const __m256 f13 = _mm256_broadcast_ss(&filter_ptr[13]);
  const __m256 f14 = _mm256_broadcast_ss(&filter_ptr[14]); const __m256 f15 = _mm256_broadcast_ss(&filter_ptr[15]);
  const __m256 f16 = _mm256_broadcast_ss(&filter_ptr[16]); const __m256 f17 = _mm256_broadcast_ss(&filter_ptr[17]);
  const __m256 f18 = _mm256_broadcast_ss(&filter_ptr[18]); const __m256 f19 = _mm256_broadcast_ss(&filter_ptr[19]);
  const __m256 f20 = _mm256_broadcast_ss(&filter_ptr[20]); const __m256 f21 = _mm256_broadcast_ss(&filter_ptr[21]);
  const __m256 f22 = _mm256_broadcast_ss(&filter_ptr[22]); const __m256 f23 = _mm256_broadcast_ss(&filter_ptr[23]);
  const __m256 f24 = _mm256_broadcast_ss(&filter_ptr[24]); 

  for (int j = 0; j < outsize; j += 8) {
    __m256 a, c0, c1;
    a = _mm256_load_ps(_img); c0 = _mm256_mul_ps(a, f0);
    a = _mm256_load_ps(_img +   8); c1 = _mm256_mul_ps(a,  f1); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  16); c1 = _mm256_mul_ps(a,  f2); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  24); c1 = _mm256_mul_ps(a,  f3); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  32); c1 = _mm256_mul_ps(a,  f4); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  40); c1 = _mm256_mul_ps(a,  f5); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  48); c1 = _mm256_mul_ps(a,  f6); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  56); c1 = _mm256_mul_ps(a,  f7); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  64); c1 = _mm256_mul_ps(a,  f8); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  72); c1 = _mm256_mul_ps(a,  f9); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  80); c1 = _mm256_mul_ps(a, f10); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  88); c1 = _mm256_mul_ps(a, f11); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img +  96); c1 = _mm256_mul_ps(a, f12); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 104); c1 = _mm256_mul_ps(a, f13); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 112); c1 = _mm256_mul_ps(a, f14); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 120); c1 = _mm256_mul_ps(a, f15); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 128); c1 = _mm256_mul_ps(a, f16); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 136); c1 = _mm256_mul_ps(a, f17); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 144); c1 = _mm256_mul_ps(a, f18); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 152); c1 = _mm256_mul_ps(a, f19); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 160); c1 = _mm256_mul_ps(a, f20); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 168); c1 = _mm256_mul_ps(a, f21); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 176); c1 = _mm256_mul_ps(a, f22); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 184); c1 = _mm256_mul_ps(a, f23); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(_img + 192); c1 = _mm256_mul_ps(a, f24); c0 = _mm256_add_ps(c0, c1);
    a = _mm256_load_ps(out + j);
    c0 = _mm256_add_ps(c0, a);
    _mm256_stream_ps(out + j, c0);
    _img += 200;
  }
  _mm256_zeroupper();
}

template <typename Dtype>
inline void dotsum_unwrapped_7x7(const Dtype* _img,  const Dtype* filter_ptr, Dtype* out, const int outsize) {
  _mm256_zeroupper();
  __m256 f[49];//=new __m256(s);
  for(int i=0; i<49; i++) {
    f[i]= _mm256_broadcast_ss(&filter_ptr[i]); 
  }
  for (int j = 0; j < outsize; j += 8) {
    __m256 a, c0, c1;
    a = _mm256_load_ps(_img);
    c0 = _mm256_mul_ps(a, f[0]);
    for(int i = 1; i < 49; i++) {
      a = _mm256_load_ps(_img + 8*i); c1 = _mm256_mul_ps(a, f[i]); c0 = _mm256_add_ps(c0, c1);
    }
    a = _mm256_load_ps(out + j);
    c0 = _mm256_add_ps(c0, a);
    _mm256_stream_ps(out + j, c0);
    _img += 49*8;
  }
  _mm256_zeroupper();
  //delete [] f;
}
#else // no AVX
template <typename Dtype>
inline void dotsum_unwrapped_2x2(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  dotsum_unwrapped_NxN<Dtype>(2, _img,  filter_ptr, out, outsize);
}
template <typename Dtype>
inline void dotsum_unwrapped_3x3(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  dotsum_unwrapped_NxN<Dtype>(3, _img,  filter_ptr, out, outsize);
}
template <typename Dtype>
inline void dotsum_unwrapped_4x4(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  dotsum_unwrapped_NxN<Dtype>(4, _img,  filter_ptr, out, outsize);
}
template <typename Dtype>
inline void dotsum_unwrapped_5x5(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  dotsum_unwrapped_NxN<Dtype>(5, _img,  filter_ptr, out, outsize);
}
template <typename Dtype>
inline void dotsum_unwrapped_7x7(const Dtype* _img, const Dtype* filter_ptr, Dtype* out, const int outsize) {
  dotsum_unwrapped_NxN<Dtype>(7, _img,  filter_ptr, out, outsize);
}
#endif

/**
 * Compute the cross product between vec_a and vec_b.
 * @param vec   Dtype*
 * @param vec_a Dtype*
 * @param vec_b Dtype*
 * @return void
 */
template <typename Dtype>
static inline void vec_cross_prod(Dtype* vec, Dtype* vec_a, Dtype* vec_b) {
  vec[0] = vec_a[1]*vec_b[2] - vec_a[2]*vec_b[1];
  vec[1] = vec_a[2]*vec_b[0] - vec_a[0]*vec_b[2];
  vec[2] = vec_a[0]*vec_b[1] - vec_a[1]*vec_b[0];
}

// TODO: delete
// l2-norm
template <typename Dtype> 
void vec_l2_norm(Dtype* val, vector<Dtype>& vec) {
  int n = vec.size();
  *val = 0;
  for(int i = 0; i < n; i++) { 
    *val += vec[i] * vec[i]; 
  }
  *val = sqrt(*val);
}
// l2-norm
template <typename Dtype> 
void vec_l2(Dtype* val, Dtype* vec, int n) {
  *val = 0;
  *val = vec_dot_prod(vec, vec, n);
  // for(int i = 0; i < n; i++) { *val += vec[i] * vec[i]; }
  // *val = sqrt(*val);
}
template <typename Dtype> 
void vec_l2_norm(Dtype* val, Dtype* vec, int n) {
  *val = 0;
  vec_l2<Dtype>(val, vec, n);
  *val = sqrt(*val);
}
template <typename Dtype> 
Dtype vec_l2_norm(Dtype* vec, int n) {
  Dtype val = 0;
  vec_l2<Dtype>(&val, vec, n);
  return sqrt(val);
}
// Create identity matrix
// static inline void eye(vector<vector<int> >& mat, int n) {
//   mat.resize(n, vector<int>(n, 0));
//   for(int i = 0; i < n; i++) {
//     mat[i][i] = 1;
//   }
// }
template <typename Dtype> 
void vec_sort(int *idx, Dtype* vec, int n)  { 
  vector<pair<Dtype, int> > vec_pair;
  vec_pair.resize(n);
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; ++i) {
    vec_pair[i] = make_pair(vec[i], i);
  }
  sort(vec_pair.begin(), vec_pair.end());
  reverse(vec_pair.begin(), vec_pair.end());
  for (int i = 0; i < n; i++) {
    vec[i] = vec_pair[i].first;
    idx[i] = vec_pair[i].second;
  }
  vec_pair.clear();
}
// TODO: delete
template <typename Dtype> // Ascending
void vect_idx_sort(vector<Dtype>& vec, vector<int>& idx) {
  iota(idx.begin(), idx.end(), 0);
  sort(idx.begin(), idx.end(), [&vec](size_t i, size_t j) { return vec[i] < vec[j]; });
  vector<Dtype> v = vec;
  for(int i = 0; i < idx.size(); i++) {
    vec[i] = v[idx[i]];
  }
}
/**
 * Normalizes the vector vec
 * @param vec Dtype*
 * @param n int
 * @return void
 */
template <typename Dtype>
void vec_normalize(Dtype* vec, int n) {
  Dtype sum = 0;
  vec_l2_norm(&sum, vec, n);
  sum = (Dtype)1.0 / (sum+DBL_EPSILON);
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; i++) {
    vec[i] = sum*vec[i];
  }
}

// template <typename Dtype>
// void vec_normalize_01(Dtype* vec, int n) {
//   // Dtype sum = 0;
//   // vec_l2_norm(&sum, vec, n);
//   // sum = (Dtype)1 / (sum+DBL_EPSILON);
//   #ifdef USE_OPENMP
//   #pragma omp parallel for
//   #endif
//   for (int i = 0; i < n; i++) {
//     vec[i] = logistic(vec[i]); // 1/(1+exp(-vec[i]));
//   }
// }
template <typename Dtype>
inline Dtype vec_dist(Dtype* vect_a, Dtype* vect_b, int n) {
  Dtype dist = 0;
  Dtype* vec = create_vec_malloc<Dtype>(n);
  vec_sub(vec, vect_a, vect_b, n);
  vec_l2_norm(&dist, vec, n);
  if (vec != NULL) { free_vec_malloc(vec); }
  return dist;
}

/**
 * Methods for matrix
 */

/**
 * Computes the mean of each row.
 */
template <typename Dtype> 
void mat_rowwise_mean(Dtype* vec, Dtype** mat, int m, int n) {
  Dtype sum = 0;
  Dtype tmp = 1/n;
  for (int i = 0; i < m; i++) {
    sum = 0;
    for (int j = 0; j < n; j++) {
      sum += mat[i][j];
    }
    vec[i] = sum * tmp;
  }
}

/**
 * Computes the mean of each column.
 */
template <typename Dtype> 
void mat_colwise_mean(Dtype* vec, Dtype** mat, int m, int n) {
  Dtype sum = 0;
  Dtype tmp = 1/m;
  for (int i = 0; i < n; i++) {
    sum = 0;
    for (int j = 0; j < m; j++) {
      sum += mat[j][i];
    }
    vec[i] = sum * tmp;
  }
}
/**
 * Computes the mean of a given matrix.
 */
template <typename Dtype> 
void mat_mean(Dtype* val, Dtype** mat, int m, int n) {
  Dtype sum = 0;
  // Dtype tmp = 1/(m*n);
  #ifdef USE_OPENMP
  #pragma omp parallel for reduction (+:sum)
  #endif
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < m; j++) {
      sum += mat[j][i];
    }
  }
  *val = sum / (m*n);
}
template <typename Dtype>
void zero_mean(Dtype** mat, int m, int n) {
  Dtype* vec = create_vec_malloc<Dtype>(n);
  mat_colwise_mean(vec, mat, m, n);
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] -= vec[j];
    }
  }
  if (vec != NULL) { free_vec_malloc(vec); }

  //  Compute data mean
  // Dtype* mean = (Dtype*) calloc(D, sizeof(Dtype));
  // if(mean == NULL) { fprintf(stderr, "Memory allocation failed!\n"); exit(1); }
  // int nD = 0;
  // for(int n = 0; n < N; n++) {
  //   for(int d = 0; d < D; d++) {
  //     mean[d] += X[nD + d];
  //   }
  //   nD += D;
  // }
  // for(int d = 0; d < D; d++) {
  //   mean[d] /= (Dtype) N;
  // }

  // Subtract data mean
  // nD = 0;
  // for(int n = 0; n < N; n++) {
  //   for(int d = 0; d < D; d++) {
  //     X[nD + d] -= mean[d];
  //   }
  //   nD += D;
  // }
  // free(mean); mean = NULL;
}
template <typename Dtype> 
void mat_init(Dtype** mat, int m, int n, Dtype val = 0, bool rand_flag = false) {
  for (int i = 0; i < m; i++) {
    vec_init(mat[i], n, val, rand_flag);
  }
}
template <typename Dtype> 
void mat_min_max_arg(Dtype* val, Dtype** mat, const int m, const int n) {
  int idx_max = 0;
  int idx_min = 0;
  Dtype arg_max = -1e6;
  Dtype arg_min = +1e6;
  for (int i = 0; i < m; i++) {
    idx_max = vec_max_arg(mat[i], n);
    idx_min = vec_min_arg(mat[i], n);
    if (arg_max < mat[i][idx_max]) {
      arg_max = mat[i][idx_max];
    }
    if (arg_min > mat[i][idx_min]) {
      arg_min = mat[i][idx_min];
    }
  }
  val[0] = arg_min;
  val[1] = arg_max;
}
//  between 0 and 1
template <typename Dtype> 
void mat_min_max_normalization(Dtype** mat, const int m, const int n) {
  Dtype* vect = create_vec_malloc<Dtype>(2);
  mat_min_max_arg(vect, mat, m, n);
  Dtype val = vect[1] - vect[0];
  if (val == 0) { val = 1e-6; }
  val = Dtype (1 / val);
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] = (mat[i][j] - vect[0]) * val;
    }
  }
  if (vect != NULL) { free_vec_malloc(vect); }
}
//  between 0 and 1
template <typename Dtype> 
void mat_min_max_normalization_rowcolwise(Dtype** mat, const int m, const int n, bool row = true) {
  int num1 = 0;
  int num2 = 0;
  if (row) { 
    num1 = m;
    num2 = n;
  } else { 
    num1 = n; 
    num2 = m; 
  }
  Dtype sum = 0.0;
  Dtype* vect = create_vec_malloc<Dtype>(num2);
  for (int i = 0; i < num1; i++) {
    sum = 0.0;
    for (int j = 0; j < num2; j++) {
      if (row) { 
        vect[j] = mat[i][j];
        sum += mat[i][j];
      } else {
        vect[j] = mat[j][i];
        sum += mat[j][i];
      }
    }
    sum = Dtype (1 / sum);
    int idx_max = vec_max_arg(vect, num2);
    int idx_min = vec_min_arg(vect, num2);
    Dtype val = vect[idx_max] - vect[idx_min];
    if (val == 0) { val = 1e-6; }
    val = Dtype (1 / val);
    for (int j = 0; j < num2; j++) {
      if (row) { 
        // mat[i][j] = (mat[i][j] - vect[idx_min]) * val;
        mat[i][j] *= sum;
      } else {
        // mat[j][i] = (mat[j][i] - vect[idx_min]) * val;;
        mat[j][i] *= sum;
      }
    }
  }
  // Dtype* vect = create_vec_malloc<Dtype>(2);
  // mat_min_max_arg(vect, mat, m, n);
  // Dtype val = vect[1] - vect[0];
  // if (val == 0) { val = 1e-6; }
  // val = Dtype (1 / val);
  // #ifdef USE_OPENMP
  // #pragma omp parallel for
  // #endif
  // for (int i = 0; i < m; i++) {
  //   for (int j = 0; j < n; j++) {
  //     mat[i][j] = (mat[i][j] - vect[0]) * val;
  //   }
  // }
  if (vect != NULL) { free_vec_malloc(vect); }
}
template <typename Dtype> 
void mat_mean_normalization(Dtype** mat, const int m, const int n) {
  Dtype val_avg = 0;
  mat_mean(&val_avg, mat, m, n);
  Dtype* vect = create_vec_malloc<Dtype>(2);
  mat_min_max_arg(vect, mat, m, n);
  Dtype val = vect[1] - vect[0];
  if (val == 0) { val = 1e-6; }
  val = Dtype (1 / val);
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] = (mat[i][j] - val_avg) * val;
    }
  }
  if (vect != NULL) { free_vec_malloc(vect); }
}


template <typename Dtype> 
void mat_standard_normalization(Dtype** mat, const int m, const int n) {
  // Standardization (Z-score Normalization)
  Dtype val_avg = 0;
  mat_mean(&val_avg, mat, m, n);
  Dtype val = 0;
  #ifdef USE_OPENMP
  #pragma omp parallel for reduction (+:val)
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      val += (mat[i][j] -  val_avg)*(mat[i][j] - val_avg);
    }
  }
  val /= (Dtype)(m*n);
  val  = sqrt(val);
  if (val == 0) { val = 1e-6; }
  val = Dtype (1 / val);

  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] = (mat[i][j] - val_avg) * val;
    }
  }
}

// template <typename Dtype> 
// void mat_normalization_01(Dtype** mat, const int m, const int n) {
//   #ifdef USE_OPENMP
//   #pragma omp parallel for
//   #endif
//   for (int i = 0; i < m; i++) {
//     for (int j = 0; j < n; j++) {
//       mat[i][j] = logistic(mat[i][j]); // 1/(1+exp(-mat[i][j]));;
//     }
//   }
// }


// template <typename Dtype> 
// void mat_colwise_mean(Dtype** mat, const int m, const int n) {
//   Dtype val = 0;
//   Dtype* vect = create_vec_malloc<Dtype>(n);
//   for (int i = 0; i < n; i++) {
//     val = 0;
//     for (int j = 0; j < m; j++) {
//       val += mat[j][i];
//     // vec_sum(&val, &(mat[i]), m);
//     // vect[i] = val;
//     }
//     vect[i] = val;
//   }
//   #ifdef USE_OPENMP
//   #pragma omp parallel for
//   #endif
//   for (int i = 0; i < m; i++) {
//     for (int j = 0; j < n; j++) {
//       mat[i][j] -= vect[j];
//     }
//   }
//   free_vec_malloc(vect);
// }
template <typename Dtype> 
void mat_transpose(Dtype** mtx, Dtype** mat, const int m, const int n) {
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mtx[j][i] = mat[i][j];
    }
  }
}
template <typename Dtype> 
void mat_scale(Dtype** mat, const int m, const int n, Dtype scale = 1, int type = 0) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      if (type == 0) {
        mat[i][j] *= scale;
      } else if (type == 1) { // tanh
        mat[i][j] = tanh(scale * mat[i][j]);
      } else if (type == 2) { 
        mat[i][j] = scale * (1-mat[i][j]*mat[i][j]);
      }
    }
  }
}
// template <typename Dtype> 
// void mat_normalize(Dtype** mat, const int m, const int n) {
//   Dtype val = 0;
//   for (int i = 0; i < m; i++) {
//     for (int j = 0; j < n; j++) {
//       val += mat[i][j]*mat[i][j];
//     }
//   }
//   val = 1/val;
//   #ifdef USE_OPENMP
//   #pragma omp parallel for
//   #endif
//   for (int i = 0; i < m; i++) {
//     for (int j = 0; j < n; j++) {
//       mat[i][j] *= val;
//     }
//   }
// }
// Normalized data with mean 0 and variance 1
template <typename Dtype> 
void mat_standard_scaler(Dtype** mat, const int m, const int n) {
  Dtype mu = 0;
  Dtype sigma = 0;
  Dtype val = 0;
  #ifdef USE_OPENMP
  #pragma omp parallel for reduction (+:val)
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      val += mat[i][j];
    }
  }
  mu = val/(m*n);
  val = 0;
  #ifdef USE_OPENMP
  #pragma omp parallel for reduction (+:val)
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      val += (mat[i][j] - mu)*(mat[i][j] - mu);
    }
  }
  sigma = (Dtype)val/(m*n);
  sigma = (Dtype)1/sqrt(sigma+DBL_EPSILON);

  #ifdef USE_OPENMP
  #pragma omp parallel for reduction (+:val)
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] = (mat[i][j] - mu)*sigma;
    }
  }
}
/**
 * Matrix subtraction. mat = mat_a - mat_b.
 */
template <typename Dtype> 
void mat_sub(Dtype** mat, Dtype** mat_a, Dtype** mat_b, int m, int n) {
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mat[i][j] = mat_a[i][j] - mat_b[i][j];
    }
  }
}
/**
 * Matrix multiplication. mat = mat_a * mat_b.
 */
template <typename Dtype> 
void mat_mult(Dtype** mat, Dtype** mat_a, int rows_a, int cols_a, Dtype** mat_b, int rows_b, int cols_b) {
  mat_init(mat, rows_a, cols_b);
  for(int i = 0; i < rows_a; i++ ) {
    for(int j = 0; j < cols_b; j++) {
      for(int k = 0; k < cols_a; k++) {
        mat[i][j] += mat_a[i][k] * mat_b[k][j];
      }
    }
  }
}
/**
 * Matrix multiplication. mat = mat_a * mat_b.
 */
template <typename Dtype> 
void mat_x_vec(Dtype* vec, Dtype** mat, Dtype* vect, int n) {
  for (int i = 0; i < n; i++) {
    vec[i] = vec_dot_prod(mat[i], vect, n);
  }
}
/**
 * Creates a diagonal matrix from vector v.
 */
template <typename Dtype> 
void mat_diag(Dtype** mat, int n, Dtype* vec) {
  mat_init(mat, n, n);
  for (int i = 0; i < n; i++) {
    mat[i][i] = vec[i];
  }
}
/**
 * Centers matrix M. (Subtracts the mean from every column)
 */
template <typename Dtype> 
void mat_center(Dtype** mat, int m, int n, Dtype* vect) {
  Dtype val = 1/m;
  for (int i = 0; i < m; i++) {
    for(int j = 0; j < n; j++) {
      vect[j] += mat[i][j];    
    }
  }
  for (int i = 0; i < n; i++) {
    vect[i] *= val; 
  }
  for (int i = 0; i < m; i++) {
    for(int j = 0; j < n; j++) {
      mat[i][j] -= vect[j];  
    }
  }
}

/**
 * De-center matrix mat. (Adds the vect to every column)
 */
template <typename Dtype> 
void mat_decenter(Dtype** mat, int m, int n, Dtype* vect) {
  for(int i = 0; i < m; i++) {
    for(int j = 0; j < n; j++) {
      mat[i][j] += vect[j]; 
    }
  }
}
/**
 * Compute inverse matrix. mtx = mat^(-1).
 */
template <typename Dtype> 
void mat_inverse(Dtype** mtx, Dtype** mat, int n) {
  Dtype** T =  create_mtx_malloc<Dtype>(n, 2*n);
  int maxrow;
  Dtype temp = 0;
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      T[i][j] = mat[i][j];
    }
  }
  for (int i = 0; i < n; i++) {
    for (int j = n; j < 2*n; j++) {
      if (i == j - n) {
        T[i][j] = 1;
      } else {
        T[i][j] = 0;
      }
    }
  }
  // Gauss-Jordan elimination
  for (int i = 0; i < n; i++) {
    maxrow = i;
    for (int j = i+1; j < n; j++) {
      if (abs(T[j][i]) > abs(T[maxrow][i])) {
        maxrow = j;
      }
    }
    for (int j = 0; j < 2*n; j++) {
      temp = T[i][j];
      T[i][j] = T[maxrow][j];
      T[maxrow][j] = temp; 
    }
    if (abs(T[i][i]) <= DBL_EPSILON) {
      printf( "\033[1;31m");
      printf( "Error: singular matrix cannot be inversed!");
      printf( "\033[0m\n");
      exit (EXIT_FAILURE);
    }
    for (int j = i+1; j < n; j++) {
      temp = T[j][i] / T[i][i];
      for (int k = i; k < 2*n; k++) {
        T[j][k] -= T[i][k] * temp;
      }
    }
  }
  for (int i = n-1; i >= 0; i--) {
    temp  = T[i][i];
    for (int j = 0; j < i; j++) {
      for (int k = 2*n-1; k >= i; k--) {
        T[j][k] -=  T[i][k] * T[j][i] / temp;
      }
    }
    T[i][i] /= temp;
    for (int j = n; j < 2*n; j++) {
      T[i][j] /= temp;
      mtx[i][j-n] = T[i][j];
    }
  }
  if (T != NULL) { free_mtx_malloc(T, n); }
}
template <typename Dtype>
void mat_copy(Dtype** mtx, Dtype** mat, int m, int n) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mtx[i][j] = mat[i][j];
    }
  }
}
template <typename Dtype>
void mat_copy(Dtype** mtx, const vector<vector<Dtype> >& mat, int m, int n) {
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < m; i++) {
    for (int j = 0; j < n; j++) {
      mtx[i][j] = mat[i][j];
    }
  }
}
template <typename Dtype>
void mat_copy(Dtype** mtx, const vector<vector<int> >& idx, const vector<Dtype>& val) {
  int n = val.size();
  #ifdef USE_OPENMP
  #pragma omp parallel for
  #endif
  for (int i = 0; i < n; i++) {
    mtx[idx[i][0]][idx[i][1]] = val[i];
    mtx[idx[i][1]][idx[i][0]] = val[i];
  }
}
template <typename Dtype> 
void mat_copy(Dtype** mtx, vector<vector<Dtype> >& mat, bool flag = false) {
  int m = mat.size();
  int n = mat[0].size();
  if (flag) {
    #ifdef USE_OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < m; i++) {
      int idx_1 = (int) mat[i][0];
      int idx_2 = (int) mat[i][1];
      mtx[idx_1][idx_2] = mat[i][2];
    }
  } else {
    #ifdef USE_OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < m; i++) {
      for (int j = 0; j < n; j++) {
        mtx[i][j] = mat[i][j];
      }
    }
  }
  mat.clear();
}

/**
 * Generates the components of the matrix mtx representing a rotation angle around 
 * an axis defined by the vector (vec). Angles must be given in degrees, 
 * and the rotation axis must be normalized.
 * @param mtx   Dtype**
 * @param m     int
 * @param n     int
 * @param angle Dtype
 * @param vec   Dtype*
 * @return void
 */

// static inline ErrorCode create_rotation_matrix(FloatValue *matxx, FloatValue *matxy, FloatValue *matxz, 
//                                                FloatValue *matyx, FloatValue *matyy, FloatValue *matyz, 
//                                                FloatValue *matzx, FloatValue *matzy, FloatValue *matzz, 
//                                                FloatValue angle, FloatValue axisx, FloatValue axisy, FloatValue axisz)
// {
template <typename Dtype> 
void mat_rotation_3x3(Dtype** mtx, Dtype angle, Dtype* vec) {
  Dtype cosine = cos(DEGREES_TO_RADIANS * angle);
  Dtype sine   = sin(DEGREES_TO_RADIANS * angle);
  Dtype val_1_consine = 1.0 - cosine;

  mtx[0][0] = vec[0] * vec[0] + (1 - vec[0] * vec[0])  * cosine;
  mtx[0][1] = vec[0] * vec[1] * val_1_consine + vec[2] * sine;
  mtx[0][2] = vec[0] * vec[2] * val_1_consine - vec[1] * sine;

  mtx[1][0] = vec[0] * vec[1] * val_1_consine - vec[2] * sine;
  mtx[1][1] = vec[1] * vec[1] + (1 - vec[1] * vec[1])  * cosine;
  // *matyz = vec[0] * vec[2] * val_1_consine + vec[0] * sine; // very likely error in previous version
  mtx[1][2] = vec[1] * vec[2] * val_1_consine + vec[0] * sine;    // very likely error in previous version

  mtx[2][0] = vec[0] * vec[2] * val_1_consine + vec[1] * sine;
  mtx[2][1] = vec[1] * vec[2] * val_1_consine - vec[0] * sine;
  mtx[2][2] = vec[2] * vec[2] + (1 - vec[2] * vec[2])  * cosine;
}



















/**
 *  GPU-based methods
 */
#ifdef USE_GPU
// template <typename Dtype>
// void two_nest_loop() {
//   dim3 grid(1, 1, 1);
//   dim3 block(4, 4, 1);
// }


template <typename Dtype>
void vec_add_gpu(Dtype* C, const Dtype* A, const Dtype* B, const int n, const Dtype a = 1, const Dtype b = 1);

// template <typename Dtype>
// void mat2vec_gpu(Dtype* vec, Dtype** mat, const int nrow, const int ncol);

#endif // USE_GPU

/* Kernel functions
 */

// #ifdef USE_GPU
// template <typename Dtype>
// __global__
// void dot_product_kernel(Dtype* dst, const int num, const Dtype* d_x, const Dtype* d_y) {
//   const int num_threads_x = blockDim.x * gridDim.x;
//   const int thread_id_x = blockIdx.x * blockDim.x + threadIdx.x;
//   Dtype sum = 0.0;
//   for(int i = thread_id_x; i < num; i += num_threads_x) {
//     sum += d_x[i]*d_y[i];
//   }
//   *dst = sum;
// }
// template <typename Dtype>
// __global__
// void scale_kernel(Dtype* dst, const int num, const Dtype* alpha) {
//   const int num_threads_x = blockDim.x * gridDim.x;
//   const int thread_id_x = blockIdx.x * blockDim.x + threadIdx.x;
//   for(int i = thread_id_x; i < num; i += num_threads_x) {
//     dst[i] *= *alpha;
//   }
// }
// #endif



































enum pad_type { zero = 0, edge = 1, median_edge = 2 }; // TODO:  delete


template <typename Dtype>
class LMatrix {
  void delete_x() { delete[] _x_mem; x = NULL;  _x_mem = NULL; }
  // 4 extra for alignment and 4 for 3 padding for SSE
  //Dtype *new_x(const int size) { _x_mem = new Dtype[size + 4+3];  x = (Dtype *)(((uintptr_t)_x_mem + 16) & ~(uintptr_t)0x0F); return x; }
  // avx mem aligment
  Dtype *new_x(const int size) { _x_mem = new Dtype[size + 8 + 7];
    x = (Dtype *)(((uintptr_t)_x_mem + 32) & ~(uintptr_t)0x1F); return x; }
public:
  string _name;
  int cols, rows, chans;
  int chan_aligned;
  int chan_stride;
  int _size;
  int _capacity;
  Dtype *x;
  Dtype *_x_mem;
  LMatrix()
  :cols(0),
   rows(0),
   chans(0),
   chan_aligned(0),
   chan_stride(0),
   _size(0),
   _capacity(0),
   x(NULL)
   /*, empty_chan(NULL)*/{}


  LMatrix(int _w, int _h, int _c = 1, const Dtype* data = NULL, int align_chan = 0)
  :cols(_w),
   rows(_h),
   chans(_c) {
    chan_aligned = align_chan;
    chan_stride = calc_chan_stride(cols, rows);
    _size = chan_stride*chans;
    _capacity =_size;
    x = new_x(_size);
    if(data != NULL) {
      memcpy(x,data,_size*sizeof(Dtype));
    }
  }
  // copy constructor - deep copy
  LMatrix(const LMatrix& m)
  :cols(m.cols),
   rows(m.rows),
   chans(m.chans),
   chan_aligned(m.chan_aligned),
   chan_stride(m.chan_stride),
   _size(m._size),
   _capacity(m._size) {
    x = new_x(_size);
    memcpy(x, m.x, sizeof(Dtype)*_size); /*empty_chan = new unsigned char[chans]; memcpy(empty_chan, m.empty_chan, chans);*/} // { v=m.v; x=(Dtype*)v.data();}
  // copy and pad constructor
  LMatrix(const LMatrix &m, int pad_cols, int pad_rows, leri::pad_type padding= leri::zero, int threads=1)
  :cols(m.cols),
   rows(m.rows),
   chans(m.chans),
   chan_aligned(m.chan_aligned),
   chan_stride(m.chan_stride),
   _size(m._size),
   _capacity(m._size) {
    x = new_x(_size);
    memcpy(x, m.x, sizeof(Dtype)*_size);
    *this = pad(pad_cols, pad_rows, padding, threads);
  }
  virtual ~LMatrix() {if(x) delete_x(); }
  // ~LMatrix() { if(x) delete_x(); }
  
  LMatrix get_chans(int start_channel, int num_chans = 1) const {
    return LMatrix(cols, rows, num_chans, &x[start_channel*chan_stride]);
  }
  // size must be divisible by 8 for AVX
  virtual int calc_chan_stride(int w, int h);

  // if edge_pad==0, then the padded area is just 0. 
  // if edge_pad==1 it fills with edge pixel colors
  // if edge_pad==2 it fills with median edge pixel color
  LMatrix pad(int dx, int dy, leri::pad_type edge_pad = leri::zero, int threads=1) const;
  LMatrix pad(int dx, int dy, int dx_right, int dy_bottom, leri::pad_type edge_pad = leri::zero, int threads=1) const;
  LMatrix crop(int dx, int dy, int w, int h, int threads=1) const;
  LMatrix shift(int dx, int dy, leri::pad_type edge_pad=leri::zero);
  LMatrix flip_cols();
  LMatrix flip_rows();

  void clip(Dtype min, Dtype max);
  void min_max(Dtype *min, Dtype *max, int *min_i=NULL, int *max_i=NULL);
  Dtype mean();
  Dtype remove_mean(int channel);
  Dtype remove_mean();
  void fill(Dtype val);
  void fill_random_uniform(Dtype range);
  void fill_random_normal(Dtype std);
  void resize(int _w, int _h, int _c, int align_chans = 0);
  int  size() const {return _size;} 
  // deep copy
  inline LMatrix& operator =(const LMatrix &m)
  {
    resize(m.cols, m.rows, m.chans, m.chan_aligned);
    memcpy(x,m.x,sizeof(Dtype)*_size);
//    memcpy(empty_chan, m.empty_chan, chans);
    return *this;
  }

  
  // dot vector to 2d mat
  inline LMatrix dot_1dx2d(const LMatrix<Dtype>& m_2d) const {
    LMatrix<Dtype> v(m_2d.rows, 1, 1);
    for(int j=0; j<m_2d.rows; j++)  v.x[j] = vec_dot_prod(x,&m_2d.x[j*m_2d.cols], _size);
    return v;
  }

  // +=
  inline LMatrix& operator+=(const LMatrix &m2){
    for(int i = 0; i < _size; i++) x[i] += m2.x[i];
    return *this;
  }
  // -=
  inline LMatrix& operator-=(const LMatrix &m2) {
    for (int i = 0; i < _size; i++) x[i] -= m2.x[i];
    return *this;
  }
#ifdef USE_AVX
  // *= float
  inline LMatrix operator *=(const Dtype v) {
    for (int i = 0; i < _size; i++) x[i] = x[i] * v;
    return *this;
  }


// #else
//   inline LMatrix operator *=(const Dtype v) {
//     __m128  b;
//     b = _mm_set_ps(v, v, v, v);
//     for (int j = 0; j < _size; j += 4)
//       _mm_store_ps(x + j, _mm_mul_ps(_mm_load_ps(x + j), b));
//     return *this;
//   }
#endif
  // *= LMatrix
  inline LMatrix operator *=(const LMatrix &v) {
    for (int i = 0; i < _size; i++) x[i] = x[i] * v.x[i];
    return *this;
  }
  inline LMatrix operator *(const LMatrix &v) {
    LMatrix T(cols, rows, chans);
    for (int i = 0; i < _size; i++) T.x[i] = x[i] * v.x[i];
    return T;
  }
  // * float
  inline LMatrix operator *(const Dtype v) {
    LMatrix T(cols, rows, chans);
    for (int i = 0; i < _size; i++) T.x[i] = x[i] * v;
    return T;
  }

  // + float
  inline LMatrix operator +(const Dtype v) {
    LMatrix T(cols, rows, chans);
    for (int i = 0; i < _size; i++) T.x[i] = x[i] + v;
    return T;
  }

  // +
  inline LMatrix operator +(LMatrix m2)
  {
    LMatrix T(cols,rows,chans);
    for(int i = 0; i < _size; i++) T.x[i] = x[i] + m2.x[i]; 
    return T;
  }
};

















// template <typename Dtype>
// class Lmatrix {
//   int _size;
//   int _capacity;
//   float* _x_mem;
//   void delete_x() { delete[] _x_mem; x = NULL;  _x_mem = NULL; }
//   // 4 extra for alignment and 4 for 3 padding for SSE
//   //float *new_x(const int size) { _x_mem = new float[size + 4+3];  x = (float *)(((uintptr_t)_x_mem + 16) & ~(uintptr_t)0x0F); return x; }
//   // avx mem aligment
//   float* new_x(const int size) { _x_mem = new float[size + 8 + 7];  x = (float *)(((uintptr_t)_x_mem + 32) & ~(uintptr_t)0x1F); return x; }
// public:
//   Lmatrix(): cols(0), rows(0), chans(0), _size(0), _capacity(0), chan_stride(0), x(NULL), chan_aligned(0)/*, empty_chan(NULL)*/{}
//   Lmatrix(int _w, int _h, int _c=1, const float *data=NULL, int align_chan=0)
//   :cols(_w), 
//    rows(_h),
//    chans(_c) {
//     chan_aligned = align_chan;
//     chan_stride = calc_chan_stride(cols, rows);
//     _size= chan_stride*chans; _capacity=_size; x = new_x(_size);
//   }
//   // copy constructor - deep copy
//   Lmatrix(const Lmatrix &m)
//   :cols(m.cols),
//    rows(m.rows),
//    chan_aligned(m.chan_aligned),
//    chans(m.chans),
//    chan_stride(m.chan_stride),
//    _size(m._size), _capacity(m._size) {
//     x = new_x(_size);
//     memcpy(x,m.x,sizeof(float)*_size); /*empty_chan = new unsigned char[chans]; memcpy(empty_chan, m.empty_chan, chans);*/} // { v=m.v; x=(float*)v.data();}
//   // copy and pad constructor
//   Lmatrix(const Lmatrix &m, int pad_cols, int pad_rows, leri::pad_type padding= leri::zero, int threads=1)
//   :cols(m.cols), 
//    rows(m.rows),
//    chans(m.chans),
//    chan_aligned(m.chan_aligned),
//    chan_stride(m.chan_stride),
//    _size(m._size),
//    _capacity(m._size) {
//     x = new_x(_size);
//     memcpy(x, m.x, sizeof(float)*_size);
//     *this = pad(pad_cols, pad_rows, padding, threads);
//   }

//   ~Lmatrix() { if(x) { delete_x(); } }
//   string _name;
//   int cols, rows, chans;
//   int chan_stride;
//   int chan_aligned;
//   float *x;

//   // size must be divisible by 8 for AVX
//   virtual int calc_chan_stride(int w, int h)
//   {
//     if (chan_aligned)
//     {
//       int s = w*h;
//       const int remainder = s % 8;
//       if (remainder > 0) s += 8 - remainder;
//       return s;
//     }
//     else return w*h;
//   }
  
//   Lmatrix get_chans(int start_channel, int num_chans = 1) const
//   {
//     return Lmatrix(cols,rows,num_chans,&x[start_channel*chan_stride]);
//   }


//   // if edge_pad==0, then the padded area is just 0. 
//   // if edge_pad==1 it fills with edge pixel colors
//   // if edge_pad==2 it fills with median edge pixel color
//   Lmatrix pad(int dx, int dy, leri::pad_type edge_pad = leri::zero, int threads=1) const
//   {
//     return pad(dx, dy, dx, dy, edge_pad, threads);
//   }
//   Lmatrix pad(int dx, int dy, int dx_right, int dy_bottom, leri::pad_type edge_pad = leri::zero, int threads=1) const
//   {
//     Lmatrix v(cols+dx+dx_right,rows+dy+dy_bottom,chans);//,NULL,this->chan_aligned);
//     v.fill(0);
  
//     //float *new_x = new float[chans*w*h]; 
// #pragma omp parallel for num_threads(threads)
//     for(int k=0; k<chans; k++)
//     {
//       const int v_chan_offset=k*v.chan_stride;
//       const int chan_offset=k*chan_stride;
//       // find median color of perimeter
//       float median = 0.f;
//       if (edge_pad == leri::median_edge)
//       {
//         int perimeter = 2 * (cols + rows - 2);
//         std::vector<float> d(perimeter);
//         for (int i = 0; i < cols; i++)
//         {
//           d[i] = x[i+ chan_offset]; d[i + cols] = x[i + cols*(rows - 1)+ chan_offset];
//         }
//         for (int i = 1; i < (rows - 1); i++)
//         {
//           d[i + cols * 2] = x[cols*i+ chan_offset];
//           // file from back so i dont need to cal index
//           d[perimeter - i] = x[cols - 1 + cols*i+ chan_offset];
//         }

//         std::nth_element(d.begin(), d.begin() + perimeter / 2, d.end());
//         median = d[perimeter / 2];
//         //for (int i = 0; i < v.rows*v.cols; i++) v.x[v_chan_offset + i] = solid_fill;
//       }

//       for(int j=0; j<rows; j++)
//       {
//         memcpy(&v.x[dx+(j+dy)*v.cols+v_chan_offset], &x[j*cols+chan_offset], sizeof(float)*cols);
//         if(edge_pad== leri::edge)
//         {
//           // do left/right side
//           for(int i=0; i<dx; i++) v.x[i+(j+dy)*v.cols+v_chan_offset]=x[0+j*cols+chan_offset];
//           for (int i = 0; i<dx_right; i++) v.x[i + dx + cols + (j + dy)*v.cols + v_chan_offset] = x[(cols - 1) + j*cols + chan_offset];
//         }
//         else if (edge_pad == leri::median_edge)
//         {
//           for (int i = 0; i < dx; i++) v.x[i + (j + dy)*v.cols + v_chan_offset] = median;
//           for (int i = 0; i < dx_right; i++) v.x[i + dx + cols + (j + dy)*v.cols + v_chan_offset] = median;
//         }
//       }
//       // top bottom pad
//       if(edge_pad== leri::edge)
//       {
//         for(int j=0; j<dy; j++) memcpy(&v.x[(j)*v.cols+v_chan_offset],&v.x[(dy)*v.cols+v_chan_offset], sizeof(float)*v.cols);
//         for (int j = 0; j<dy_bottom; j++) memcpy(&v.x[(j + dy + rows)*v.cols + v_chan_offset], &v.x[(rows - 1 + dy)*v.cols + v_chan_offset], sizeof(float)*v.cols);
//       }
//       if (edge_pad == leri::median_edge)
//       {
//         for (int j = 0; j<dy; j++)  
//           for (int i = 0; i<v.cols; i++) 
//             v.x[i + j*v.cols + v_chan_offset] = median;
//         for (int j = 0; j<dy_bottom; j++) 
//           for (int i = 0; i<v.cols; i++) 
//             v.x[i + (j + dy + rows)*v.cols + v_chan_offset] = median;
//       }
//     }

//     return v;
//   }

//   Lmatrix crop(int dx, int dy, int w, int h, int threads=1) const
//   {
//     Lmatrix v(w,h,chans);

// #pragma omp parallel for num_threads(threads)
//     for(int k=0; k<chans; k++)
//     {
//       for(int j=0; j<h; j++)
//       {
//         memcpy(&v.x[j*w+k*v.chan_stride], &x[dx+(j+dy)*cols+k*chan_stride], sizeof(float)*w);
//       }
//     }

//     return v;
//   }

//   leri::Lmatrix shift(int dx, int dy, leri::pad_type edge_pad=leri::zero)
//   {
//     int orig_cols=cols;
//     int orig_rows=rows;
//     int off_x=abs(dx);
//     int off_y=abs(dy);

//     leri::Lmatrix shifted= pad(off_x, off_y, edge_pad);

//     return shifted.crop(off_x-dx, off_y-dy,orig_cols,orig_rows);
//   }

//   leri::Lmatrix flip_cols()
//   {
//     leri::Lmatrix v(cols,rows,chans);
//     for(int k=0; k<chans; k++)
//       for(int j=0; j<rows; j++)
//         for(int i=0; i<cols; i++)
//           v.x[i+j*cols+k*chan_stride]=x[(cols-i-1)+j*cols+k*chan_stride];

//     return v;
//   }
//   leri::Lmatrix flip_rows()
//   {
//     leri::Lmatrix v(cols, rows, chans);
    
//     for (int k = 0; k<chans; k++)
//       for (int j = 0; j<rows; j++)
//         memcpy(&v.x[(rows-1-j)*cols + k*chan_stride],&x[j*cols + k*chan_stride], cols*sizeof(float));

//     return v;
//   }

//   void clip(float min, float max)
//   {
//     int s = chan_stride*chans;
//     for (int i = 0; i < s; i++)
//     {
//       if (x[i] < min) x[i] = min;
//       if (x[i] > max) x[i]=max;
//     }
//   }


//   void min_max(float *min, float *max, int *min_i=NULL, int *max_i=NULL)
//   {
//     int s = rows*cols;
//     int mini = 0;
//     int maxi = 0; 
//     for (int c = 0; c < chans; c++)
//     {
//       const int t = chan_stride*c;
//       for (int i = t; i < t+s; i++)
//     {
//       if (x[i] < x[mini]) mini = i;
//       if (x[i] > x[maxi]) maxi = i;
//     }
//     }
//     *min = x[mini];
//     *max = x[maxi];
//     if (min_i) *min_i = mini;
//     if (max_i) *max_i = maxi;
//   }

//   float mean()
//   {
//     const int s = rows*cols;
//     int cnt = 0;// channel*s;
//     float average = 0;
//     for (int c = 0; c < chans; c++)
//     {
//       const int t = chan_stride*c;
//       for (int i = 0; i < s; i++)
//         average += x[i + t];
//     }
//     average = average / (float)(s*chans);
//     return average;
//   }
//   float remove_mean(int channel)
//   {
//     int s = rows*cols;
//     int offset = channel*chan_stride;
//     float average=0;
//     for(int i=0; i<s; i++) average+=x[i+offset];    
//     average= average/(float)s;
//     for(int i=0; i<s; i++) x[i+offset]-=average;    
//     return average;
//   }

//   float remove_mean()
//   {
//     float m=mean();
//     int s = chan_stride*chans;
//     //int offset = channel*s;
//     for(int i=0; i<s; i++) x[i]-=m;   
//     return m;
//   }
//   void fill(float val) { for(int i=0; i<_size; i++) x[i]=val; 
//   }
//   void fill_random_uniform(float range)
//   {
//     std::mt19937 gen(0);
//     std::uniform_real_distribution<float> dst(-range, range);
//     for (int i = 0; i<_size; i++) x[i] = dst(gen);
//   }
//   void fill_random_normal(float std)
//   {
//     std::mt19937 gen(0);
//     std::normal_distribution<float> dst(0, std);
//     for (int i = 0; i<_size; i++) x[i] = dst(gen);
//   }


//   // deep copy
//   inline Lmatrix& operator =(const Lmatrix &m)
//   {
//     resize(m.cols, m.rows, m.chans, m.chan_aligned);
//     memcpy(x,m.x,sizeof(float)*_size);
// //    memcpy(empty_chan, m.empty_chan, chans);
//     return *this;
//   }

//   int  size() const {return _size;} 
  
//   void resize(int _w, int _h, int _c, int align_chans=0) { 
//     chan_aligned = align_chans;
//     int new_stride = calc_chan_stride(_w,_h);
//     int s = new_stride*_c;
//     if(s>_capacity) 
//     { 
//       if(_capacity>0) delete_x(); _size = s; _capacity=_size; x = new_x(_size);
//     }
//     cols = _w; rows = _h; chans = _c; _size = s; chan_stride = new_stride; 
//   } 
  
//   // dot vector to 2d mat
//   inline Lmatrix dot_1dx2d(const Lmatrix &m_2d) const
//   {
//     leri::Lmatrix v(m_2d.rows, 1, 1);
//     for(int j=0; j<m_2d.rows; j++)  v.x[j]=vec_dot_prod(x,&m_2d.x[j*m_2d.cols],_size);
//     return v;
//   }

//   // +=
//   inline Lmatrix& operator+=(const Lmatrix &m2){
//     for(int i = 0; i < _size; i++) x[i] += m2.x[i];
//     return *this;
//   }
//   // -=
//   inline Lmatrix& operator-=(const Lmatrix &m2) {
//     for (int i = 0; i < _size; i++) x[i] -= m2.x[i];
//     return *this;
//   }
// #ifndef MOJO_AVX
//   // *= float
//   inline Lmatrix operator *=(const float v) {
//     for (int i = 0; i < _size; i++) x[i] = x[i] * v;
//     return *this;
//   }


// #else
//   inline Lmatrix operator *=(const float v) {
//     __m128  b;
//     b = _mm_set_ps(v, v, v, v);
//     for (int j = 0; j < _size; j += 4)
//       _mm_store_ps(x + j, _mm_mul_ps(_mm_load_ps(x + j), b));
//     return *this;
//   }
// #endif
//   // *= Lmatrix
//   inline Lmatrix operator *=(const Lmatrix &v) {
//     for (int i = 0; i < _size; i++) x[i] = x[i] * v.x[i];
//     return *this;
//   }
//   inline Lmatrix operator *(const Lmatrix &v) {
//     Lmatrix T(cols, rows, chans);
//     for (int i = 0; i < _size; i++) T.x[i] = x[i] * v.x[i];
//     return T;
//   }
//   // * float
//   inline Lmatrix operator *(const float v) {
//     Lmatrix T(cols, rows, chans);
//     for (int i = 0; i < _size; i++) T.x[i] = x[i] * v;
//     return T;
//   }

//   // + float
//   inline Lmatrix operator +(const float v) {
//     Lmatrix T(cols, rows, chans);
//     for (int i = 0; i < _size; i++) T.x[i] = x[i] + v;
//     return T;
//   }

//   // +
//   inline Lmatrix operator +(Lmatrix m2)
//   {
//     Lmatrix T(cols,rows,chans);
//     for(int i = 0; i < _size; i++) T.x[i] = x[i] + m2.x[i]; 
//     return T;
//   }
// };


} // namespace leri
#endif // LERI_MATRIX_H_




