/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_RANDOM_H_
#define LERI_RANDOM_H_

#include <math.h>
#include <string>
#include <stdlib.h>     /* srand, rand */
#include <time.h>       /* time */

#include "boost/random/mersenne_twister.hpp"
#include "boost/random/uniform_int.hpp"

namespace leri {
typedef boost::mt19937 rng_t;
/**
 * Implementation of a Mersenne Twister Random Number Generator. 
 * This random number generator is described in the article by
 * M. Matsumoto & T. Nishimura, in:
 * ACM Transactions on Modeling and Computer Simulation,
 * vol. 8, no. 1, 1998, pp. 3-30.
 * It is very well suited for Monte Carlo simulations
 * This code has been ported to C from the original C++ code by Agner Fog. This code can be
 * found at http://www.agner.org.
 * For more information on the Mersenne Twister, check prof. M. Matsumoto's website:
 * http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
 *
 */

// Define integer types with known size: int32_t, uint32_t, int64_t, uint64_t.
// If this doesn't work then insert compiler-specific definitions here:
#if defined(__GNUC__)
  // Compilers supporting C99 or C++0x have inttypes.h defining these integer types
  #include <inttypes.h>
  #define INT64_SUPPORTED // Remove this if the compiler doesn't support 64-bit integers
#elif defined(_WIN16) || defined(__MSDOS__) || defined(_MSDOS) 
  // 16 bit systems use long int for 32 bit integer
  typedef   signed long int int32_t;
  typedef unsigned long int uint32_t;
#elif defined(_MSC_VER)
  // Microsoft have their own definition
  typedef   signed __int32  int32_t;
  typedef unsigned __int32 uint32_t;
  typedef   signed __int64  int64_t;
  typedef unsigned __int64 uint64_t;
  #define INT64_SUPPORTED // Remove this if the compiler doesn't support 64-bit integers
#else
  // This works with most compilers
  typedef signed int          int32_t;
  typedef unsigned int       uint32_t;
  typedef long long           int64_t;
  typedef unsigned long long uint64_t;
  #define INT64_SUPPORTED // Remove this if the compiler doesn't support 64-bit integers
#endif

// #define MT11213A // Uses the Mersenne Twister 11213A (otherwise 19937 is used).
// Choose which version of Mersenne Twister you want:
#ifdef MT11213A
  // Define constants for type MT11213A:
  #define MERS_N   351
  #define MERS_M   175
  #define MERS_R   19
  #define MERS_U   11
  #define MERS_S   7
  #define MERS_T   15
  #define MERS_L   17
  #define MERS_A   0xE4BD75F5
  #define MERS_B   0x655E5280
  #define MERS_C   0xFFD58000
#else
  // or constants for type MT19937:
  #define MERS_N   624
  #define MERS_M   397
  #define MERS_R   31
  #define MERS_U   11
  #define MERS_S   7
  #define MERS_T   15
  #define MERS_L   18
  #define MERS_A   0x9908B0DF
  #define MERS_B   0x9D2C5680
  #define MERS_C   0xEFC60000
#endif



/**
 * Basic initialization procedure.
 * @param seed int
 */
void init_mersenne(int seed);

/**
 * Initialize and seed generator.
 * @param seed int
 */
void set_seed(int seed);

/**
 * Generates 32 random bits,
 */
unsigned int get_random_bits();

/**
 * Generates random float number in the interval 0 <= x < 1
 */
template<typename Dtype>
Dtype get_frandom() {
  // Multiply by 2^(-32)
  return (Dtype) get_random_bits() * (1.0/(65536.0*65536.0));
}
/**
 * Generates random integer in the interval min <= x <= max
 * Relative error on frequencies < 2^-32
 * @param min int
 * @param max int
 */
// int get_int_random(int min, int max);
/**
 * Generates random integer in the interval min <= x < max
 * Relative error on frequencies < 2^-32
 * @param min int
 * @param max int
 */

/**
 * Output random integer in the interval min <= x <= max
 * Each output value has exactly the same probability.
 * This is obtained by rejecting certain bit values so that the number
 * of possible bit values is divisible by the interval length
 * @param min int
 * @param max int
 */
int get_irandom(int min, int max);     // Output random integer, exact

void get_log_random(double* r);
// template <typename Dtype>
// void get_normal_random(Dtype* r);
// rd_rng_normal
// Box muller method

template <typename Dtype>
void get_normal_random(Dtype* rnd) {
  Dtype n0, n1, s;
  /* initialize random seed: */
  srand (time(NULL));
  do {
    n0 = 2.0 * rand()/RAND_MAX - 1.0;
    n1 = 2.0 * rand()/RAND_MAX - 1.0;
    s = n0 * n0 + n1 * n1;
    // printf("%f\n", s);
  } while (s >= 1.0 || s == 0.0);
  *rnd = n0 * sqrt(-2.0 * log(s) / s);
}
template <typename Dtype>
Dtype get_normal_random() {
  Dtype n0, n1, s, rnd;
  /* initialize random seed: */
  srand (time(NULL));
  do {
    n0 = 2.0 * rand()/RAND_MAX - 1.0;
    n1 = 2.0 * rand()/RAND_MAX - 1.0;
    s = n0 * n0 + n1 * n1;
    // printf("%f\n", s);
  } while (s >= 1.0 || s == 0.0);
  rnd = n0 * sqrt(-2.0 * log(s) / s);
  return rnd;
}
template <typename Dtype>
Dtype get_random_h() {
  return pow(10., 1.5 - 6.*get_frandom<Dtype>())*get_normal_random<Dtype>();
}

// For random numbers
#define IM1 2147483563
#define IM2 2147483399
#define IMM1 (IM1 - 1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1 + IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0 - EPS)
template <typename Dtype>
Dtype rnd_uni(long *idum) {
  long j, k;
  static long idum2 = 123456789;
  static long iy = 0;
  static long iv[NTAB];
  Dtype temp;

  if (*idum <= 0) {
    if (-(*idum) < 1) { *idum = 1; } 
    else { *idum = -(*idum); }
    idum2 = (*idum);
    for (j=NTAB+7;j >= 0;j--) {
      k = (*idum)/IQ1;
      *idum = IA1*(*idum - k*IQ1) - k*IR1;
      if (*idum < 0) { *idum += IM1; }
      if (j < NTAB) { iv[j] = *idum; }
    }
    iy = iv[0];
  }
  k = (*idum)/IQ1;
  *idum = IA1*(*idum - k*IQ1) - k*IR1;
  if (*idum < 0) { *idum += IM1; }
  k = idum2/IQ2;
  idum2 = IA2*(idum2 - k*IQ2) - k*IR2;
  if (idum2 < 0) { idum2 += IM2; }
  j = iy/NDIV;
  iy = iv[j] - idum2;
  iv[j] = *idum;
  if (iy < 1) { iy += IMM1; }
  Dtype AM = (1.0/IM1);
  temp = (Dtype) AM*iy;
  if (temp > RNMX) { return (Dtype) RNMX; } 
  else { return temp; }
}






// Returns a string of random alphabets of length n. 
inline std::string str_random(int n, int m = 62) {
  std::string str = "";
  char alphabet[62] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K',
                        'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
                        'U', 'V', 'W', 'X', 'Y', 'Z',
                        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                        'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
                        'u', 'v', 'w', 'x', 'y', 'z'
                    }; 

  for (int i = 0; i < n; i++) {
    set_seed(time(0)+i);
    int r = get_irandom(0, m-1);
    str += alphabet[r]; 
  }
  return str; 
} 

}// namespace leri
#endif // LERI_RANDOM_H_
