/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_ATOM_H_
#define LERI_ATOM_H_

#include "leri/common.h"
#include "leri/protein/bioutil.h"

namespace leri {

template <typename Dtype>
class Atom {
public:
  explicit Atom() { active_ = 0; }
  explicit Atom(string name, int parent_1, int parent_2, int parent_3, Dtype dist, Dtype angle, Dtype torsion, string iscore, string chitype)
  :name_(name),
   parent_1_(parent_1),
   parent_2_(parent_2),
   parent_3_(parent_3),
   dist_(dist),
   angle_(angle),
   torsion_(torsion),
   iscore_(iscore),
   chitype_(chitype) { active_ = 0; }
  virtual ~Atom() {}

  inline void set_index(int index)            { idx_  = index; }
  inline void set_name(string str)            { name_ = str; }
  inline void set_isomer(string str)          { isomer_ = str; }
  inline void set_amino_acid_name(string str) { amino_acid_name_ = str; }
  inline void set_chain_name(string str)      { chain_name_ = str; }
  inline void set_amino_acid_idx(int indx)    { amino_acid_indx_ = indx; }
  inline void set_coord(Dtype x, Dtype y, Dtype z) {
    coord_[0] = x;
    coord_[1] = y;
    coord_[2] = z;
    // coord_[3] = w;
  }
  inline void set_occupancy(Dtype occupancy)       { occupancy_ = occupancy; }
  inline void set_tfactor(Dtype tfactor)           { tfactor_   = tfactor; }
  inline void set_symbol(string symbol)            { symbol_    = symbol; }
  inline void set_iscore(string str)               { iscore_    = str; }
  inline void set_extra(string str)                { extra_     = str; }

  inline int get_indx() { return idx_; }
  inline string get_atom_name() { return name_; }
  inline string get_amino_acid_name() { return amino_acid_name_; }
  inline int get_amino_acid_indx() { return amino_acid_indx_; }
  inline string get_chain_name() { return chain_name_; }

public:
  /**
   * Atom type (Carbon, Hydrogen, Oxygen, etc).
   */
  char type_;
  int  idx_;
  string name_;

  int parent_1_;  // first parent
  int parent_2_;  // second parent
  int parent_3_;  // third parent
  
  Dtype dist_;    // distance from first parent
  Dtype angle_;   // angle made with first and second parents
  Dtype torsion_; // torsion made withe first, second and third parents
  string iscore_;
  string chitype_;
  Dtype coord_[3];

  string isomer_;
  string amino_acid_name_;
  string chain_name_;
  int amino_acid_indx_;
  int amino_acid_indx_sorted_; // TODO: 






  // TODO: for folding, from struct.h 
  /**
   * Pointers to the xyzw coordinates of the atom.
   */
  // Dtype xyzw[4];
  
  // Dtype *x;
  // Dtype *y;
  // Dtype *z;
  // Dtype *w;

  /**
   * Residue and chain where this atom resides.
   */
  // int res;
  // int chain;

  /**
   * Arrays to hold custom integer or float properties for each atom.
   * This arrays are always NULL when the atom record is initialized. 
   */
  int niprop_;
  // int *iproperties;
  // int nfprop;
  // double *fproperties;

  Dtype  occupancy_; // Occupancy
  Dtype  tfactor_;   // Temperature factor
  string symbol_;    // Element symbol
  string extra_;
private:
  bool active_;
protected:
  // DISABLE_COPY_AND_ASSIGN(Atom);
};  // class Atom

}  // namespace leri

#endif  // LERI_ATOM_H_


// # Atom types currently recognized by the FoldingSimulator object.
// supported_atoms = {"N"  : 0,
//                    "HN" : 1, "H"  : 2,
//                    "CA" : 3,
//                    "CB" : 4,
//                    "CG" : 5, "CG1": 6, "CG2": 7, "OG" : 8, "OG1": 9, "SG" :10,
//                    "CD" :11, "CD1":12, "CD2":13, "OD" :14, "OD1":15, "OD2":16, "ND" :17, "ND1":18, "ND2":19, "SD" :20,
//                    "CE" :21, "CE1":22, "CE2":23, "CE3":24, "OE" :25, "OE1":26, "OE2":27, "NE" :28, "NE1":29, "NE2":30,
//                    "CZ" :31, "CZ2":32, "CZ3":33, "NZ" :34,
//                    "CH2":35, "OH" :36, "NH" :37, "NH1":38, "NH2":39,
//                    "HA1":40, "1HA":41, "HA" :42,
//                    "HA2":43, "2HA":44,
//                    "C"  :45, "O"  :46}

// # Dictionary that specify the  order in which the atoms should be passed to the underlying C library in the case of not using full side-chains.
// aa_nosc_atoms = {"N":0, "HN":1, "CA":2, "CB":3, "HA2":4, "HA1":5, "C":6, "O":7}

// # These dictionaries specify the order in which the atoms should be passed to the underlying C library for each amino-acid. This ordering is important to properly
// # construct the protein connectivity.
// ala_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "HA1":4, "C"  :5, "O"  :6}
// arg_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD" :5, "NE" :6, "CZ" :7, "NH1":8, "NH2":9, "HA1":10, "C"  :11,  "O" :12}
// asn_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "OD1":5, "ND2":6, "HA1":7, "C"  :8, "O"  :9}
// asp_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "OD1":5, "OD2":6, "HA1":7, "C"  :8, "O"  :9}
// cys_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "SG" :4, "HA1":5, "C"  :6, "O"  :7}
// gln_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD" :5, "OE1":6, "NE2":7, "HA1":8, "C"  :9, "O"  :10}
// glu_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD" :5, "OE1":6, "OE2":7, "HA1":8, "C"  :9, "O"  :10}
// gly_atoms = {"N":0, "HN":1, "CA":2, "HA2":3, "HA1":4, "C"  :5, "O"  :6}
// his_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "ND1":5, "CD2":6, "CE1":7, "NE2":8, "HA1":9, "C"  :10, "O"  :11}
// ile_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG1":4, "CD1":5, "CG2":6, "HA1":7, "C"  :8, "O"  :9}
// leu_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD1":5, "CD2":6, "HA1":7, "C"  :8, "O"  :9}
// lys_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD" :5, "CE" :6, "NZ" :7, "HA1":8, "C"  :9, "O"  :10}
// met_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "SD" :5, "CE" :6, "HA1":7, "C"  :8, "O"  :9}
// phe_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD1":5, "CD2":6, "CE1":7, "CE2":8, "CZ" :9, "HA1":10, "C"  :11,  "O" :12}
// pro_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "HA1":5, "C"  :6, "O"  :7}
// ser_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "OG" :4, "HA1":5, "C"  :6, "O"  :7}
// thr_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "OG1":4, "CG2":5, "HA1":6, "C"  :7, "O"  :8}
// trp_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD1":5, "CD2":6, "NE1":7, "CE2":8, "CE3":9, "CZ2":10, "CZ3":11, "CH2":12, "HA1":13, "C" :14,  "O" :15}
// tyr_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG" :4, "CD1":5, "CD2":6, "CE1":7, "CE2":8, "CZ" :9, "OH" :10, "HA1":11, "C"  :12,  "O" :13}
// val_atoms = {"N":0, "HN":1, "CA":2, "CB" :3, "CG1":4, "CG2":5, "HA1":6, "C"  :7, "O"  :8}

// aa_atoms = [ala_atoms, arg_atoms, asn_atoms, asp_atoms, cys_atoms, gln_atoms, glu_atoms, gly_atoms, his_atoms, ile_atoms,
//             leu_atoms, lys_atoms, met_atoms, phe_atoms, pro_atoms, ser_atoms, thr_atoms, trp_atoms, tyr_atoms, val_atoms]

