/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */
 
#ifndef LERI_PROTEIN_UTIL_H_
#define LERI_PROTEIN_UTIL_H_

#include "leri/common.h"

namespace leri {

// #define char2digit(c)    ((47 < c) && (c < 58))

static const char PROTEIN  = 1; // Type
#define NUM_AA (20+1) // Number of common residues + gap
#define NHLength 1    // angstrom

#define PDB_FORMAT_0 "ATOM  %5d  %-4s%-4s%1s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f  %12s\n"
#define PDB_FORMAT_1 "ATOM  %5d %-4s %3s %1s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n"
#define PDB_FORMAT_2 "ATOM  %5d %-4s %-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n"
#define PDB_FORMAT_3 "ATOM  %5i  %-3s %s %c %3i     %7.3lf %7.3lf %7.3lf  1.00  0.00\n" // Used in DSA

const string AA_STR = "ARNDCQEGHILKMFPSTWYV";
// H: Alpha helix; G: 3-10 helix; I: PI-helix; E: Extended conformation
// B or b: Isolated bridge; T: Turn; C: Coil (none of the above)
const string AA_SEC    = "HGIEBTC";
const int gap_code = -2, gap_extend_code = -8;

#define H_BOND 0.88
#define O_BOND 1.24

// Dictionary that specify the  order in which the atoms should be passed 
// to the underlying C++ library in the case of not using full side-chains.
static vector<string> ATOMS_NO_SC = {"N", "HN", "CA", "CB", "HA2", "HA1", "C", "O"};
static vector<string> ATOMS_BACKBONE = {"N", "CA", "C", "O"};
static vector<string> ATOMS_BACKBONE_CB = {"N", "CA", "C", "O", "CB"};
// static map<string, int> ATOMS_MAP_NO_SC = {{"N", 0}, {"HN", 1}, {"CA", 2}, {"CB", 3}, {"HA2", 4}, {"HA1", 5}, {"C", 6}, {"O", 7}};
int atom2idx(string atom, const string aa, int idx_aa, bool use_SC = false);


// Charged (side chains often form salt bridges): "+""
// • Arginine - Arg - R
// • Lysine   - Lys - K
// • Aspartic acid - Asp - D
// • Glutamic acid - Glu - E

// Polar (form hydrogen bonds as proton donors or acceptors): "†" or ">"
// • Glutamine  - Gln - Q
// • Asparagine - Asn - N
// • Histidine  - His - H
// • Serine     - Ser - S
// • Threonine  - Thr - T
// • Tyrosine   - Tyr - Y
// • Cysteine   - Cys - C

// Amphipathic (often found at the surface of proteins or lipid membranes, sometimes also classified as polar): "‡" or "#"
// • Tryptophan - Trp - W
// • Tyrosine   - Tyr - Y
// • Methionine - Met - M (may function as a ligand to metal ions)

// Hydrophobic (normally buried inside the protein core): "*"
// • Alanine    - Ala - A
// • Isoleucine - Ile - I
// • Leucine    - Leu - L
// • Methionine - Met - M
// • Phenylalanine - Phe - F
// • Valine  - Val - V
// • Proline - Pro - P
// • Glycine - Gly - G

// static const string aa_property_note = "The symbol before each amino acid is illustrated as follows, +: charged (side chains often form salt bridges), >: polar (form hydrogen bonds as proton donors or acceptors), *: hydrophobic (normally buried inside the protein core), and #: amphipathic (often found at the surface of proteins or lipid membranes, sometimes also classified as polar).";
static const string aa_property_note = "";
// static map<char, const string> aa_property = {
//   {'A', "*A"}, {'R', "+R"}, {'N', "†N"},  {'D', "+D"},  {'C', "†C"}, 
//   {'Q', "†Q"}, {'E', "+E"}, {'G', "*G"},  {'H', "†H"},  {'I', "*I"}, 
//   {'L', "*L"}, {'K', "+K"}, {'M', "*‡M"}, {'F', "*F"},  {'P', "*P"}, 
//   {'S', "†S"}, {'T', "†T"}, {'W', "‡W"},  {'Y', "†‡Y"}, {'V', "*V"}
// };
// static map<char, const string> aa_property = {
//   {'A', "*A"}, {'R', "+R"}, {'N', ">N"},  {'D', "+D"},  {'C', ">C"}, 
//   {'Q', ">Q"}, {'E', "+E"}, {'G', "*G"},  {'H', ">H"},  {'I', "*I"}, 
//   {'L', "*L"}, {'K', "+K"}, {'M', "*#M"}, {'F', "*F"},  {'P', "*P"}, 
//   {'S', ">S"}, {'T', ">T"}, {'W', "#W"},  {'Y', ">#Y"}, {'V', "*V"}
// };
static map<char, const string> aa_property = {
  {'A', "Hydrophobic"}, {'R', "Positive"}, {'N', "Polar uncharged"}, {'D', "Negative"}, {'C', "Special"}, 
  {'Q', "Polar uncharged"}, {'E', "Negative"}, {'G', "Special"}, {'H', "Positive"}, {'I', "Hydrophobic"}, 
  {'L', "Hydrophobic"}, {'K', "Positive"}, {'M', "Hydrophobic"}, {'F', "Hydrophobic"}, {'P', "Special"}, 
  {'S', "Polar uncharged"}, {'T', "Polar uncharged"}, {'W', "Hydrophobic"}, {'Y', "Hydrophobic"}, {'V', "Hydrophobic"}
};

// Constants for Amino Acids
static map<string, int> AA2INT = {
  {"ALA",  1}, {"ARG",  2}, {"ASN",  3}, {"ASP",  4}, {"CYS",  5},
  {"GLN",  6}, {"GLU",  7}, {"GLY",  8}, {"HIS",  9}, {"ILE", 10},
  {"LEU", 11}, {"LYS", 12}, {"MET", 13}, {"PHE", 14}, {"PRO", 15},
  {"SER", 16}, {"THR", 17}, {"TRP", 18}, {"TYR", 19}, {"VAL", 20},
  {"UNK", 21}
};
static map<char, string> AA2FUL = {
  {'A',  "Alanine"}, {'R',  "Arginine"}, {'N',  "Asparagine"}, {'D',  "Aspartic acid"}, {'C',  "Cysteine"},
  {'Q',  "Glutamine"}, {'E',  "Glutamic acid"}, {'G',  "Glycine"}, {'H',  "Histidine"}, {'I', "Isoleucine"},
  {'L', "Leucine"}, {'K', "Lysine"}, {'M', "Methionine"}, {'F', "Phenylalanine"}, {'P', "Proline"},
  {'S', "Serine"}, {'T', "Threonine"}, {'W', "Tryptophan"}, {'Y', "Tyrosine"}, {'V', "Valine"},
  {'U', "Unknow"}
};
static map<string, string> AA32FUL = {
  {"ALA",  "Alanine"}, {"ARG",  "Arginine"}, {"ASN",  "Asparagine"}, {"ASP",  "Aspartic acid"}, {"CYS",  "Cysteine"},
  {"GLN",  "Glutamine"}, {"GLU",  "Glutamic acid"}, {"GLY",  "Glycine"}, {"HIS",  "Histidine"}, {"ILE", "Isoleucine"},
  {"LEU", "Leucine"}, {"LYS", "Lysine"}, {"MET", "Methionine"}, {"PHE", "Phenylalanine"}, {"PRO", "Proline"},
  {"SER", "Serine"}, {"THR", "Threonine"}, {"TRP", "Tryptophan"}, {"TYR", "Tyrosine"}, {"VAL", "Valine"},
  {"UNK", "Unknow"}
};

int aa12idx(char aa);      // one letters to index
void aa32idx(int* hb, string aa_str); // three letters to index
// void idx2aa(unsigned char* aa,  int _idx);
void idx2aa(unsigned char* aa,  int idx);
string idx2sec(int idx);
// void aa123(string* aa_3L, unsigned char aa_1L);
string aa123(unsigned char aa_1L);
// void aa321(unsigned char* aa_1L, string aa_3L);
string aa321(string aa_3L);
void pair_index(int** mat, int m, int n);

/**
 * Divides an atom name into its component parts and put them into the characters *at, *id, *n1 and *n2. 
 * For instance, for a string like "HA1", *at is set 'H', *id to 'A', *n1 to '1' and *n2 to ' '.
 * @param atname const char*
 * @param type char*
 * @param id char*
 * @param n1 char*
 * @param n2 char*
 * @return void
 */
void get_atom_name(char type, char *atname);
/**
 * Converts the type code value of an atom into a pdb name.
 * @param type const char
 * @param aaname char*
 * @return void
 */
void get_aminoacid_name(const char type, char *aaname);


const int BLOSUM50[][20] = {
  { 5, -2, -1, -2, -1, -1, -1,  0, -2, -1, -2, -1, -1, -3, -1,  1,  0, -3, -2,  0}, 
  {-2,  7, -1, -2, -4,  1,  0, -3,  0, -4, -3,  3, -2, -3, -3, -1, -1, -3, -1, -3}, 
  {-1, -1,  7,  2, -2,  0,  0,  0,  1, -3, -4,  0, -2, -4, -2,  1,  0, -4, -2, -3}, 
  {-2, -2,  2,  8, -4,  0,  2, -1, -1, -4, -4, -1, -4, -5, -1,  0, -1, -5, -3, -4}, 
  {-1, -4, -2, -4, 13, -3, -3, -3, -3, -2, -2, -3, -2, -2, -4, -1, -1, -5, -3, -1}, 
  {-1,  1,  0,  0, -3,  7,  2, -2,  1, -3, -2,  2,  0, -4, -1,  0, -1, -1, -1, -3}, 
  {-1,  0,  0,  2, -3,  2,  6, -3,  0, -4, -3,  1, -2, -3, -1, -1, -1, -3, -2, -3}, 
  { 0, -3,  0, -1, -3, -2, -3,  8, -2, -4, -4, -2, -3, -4, -2,  0, -2, -3, -3, -4}, 
  {-2,  0,  1, -1, -3,  1,  0, -2, 10, -4, -3,  0, -1, -1, -2, -1, -2, -3,  2, -4}, 
  {-1, -4, -3, -4, -2, -3, -4, -4, -4,  5,  2, -3,  2,  0, -3, -3, -1, -3, -1,  4}, 
  {-2, -3, -4, -4, -2, -2, -3, -4, -3,  2,  5, -3,  3,  1, -4, -3, -1, -2, -1,  1}, 
  {-1,  3,  0, -1, -3,  2,  1, -2,  0, -3, -3,  6, -2, -4, -1,  0, -1, -3, -2, -3}, 
  {-1, -2, -2, -4, -2,  0, -2, -3, -1,  2,  3, -2,  7,  0, -3, -2, -1, -1,  0,  1}, 
  {-3, -3, -4, -5, -2, -4, -3, -4, -1,  0,  1, -4,  0,  8, -4, -3, -2,  1,  4, -1}, 
  {-1, -3, -2, -1, -4, -1, -1, -2, -2, -3, -4, -1, -3, -4, 10, -1, -1, -4, -3, -3}, 
  { 1, -1,  1,  0, -1,  0, -1,  0, -1, -3, -3,  0, -2, -3, -1,  5,  2, -4, -2, -2}, 
  { 0, -1,  0, -1, -1, -1, -1, -2, -2, -1, -1, -1, -1, -2, -1,  2,  5, -3, -2,  0}, 
  {-3, -3, -4, -5, -5, -1, -3, -3, -3, -3, -2, -3, -1,  1, -4, -4, -4, 15,  2, -3}, 
  {-2, -1, -2, -3, -3, -1, -2, -3,  2, -1, -1, -2,  0,  4, -3, -2, -2,  2,  8, -1}, 
  { 0, -3, -3, -4, -1, -3, -3, -4, -4,  4,  1, -3,  1, -1, -3, -2,  0, -3, -1,  5}
};


inline void substitution_matrix(string *aa_str, string aa) {
  if      (aa == "A") { *aa_str = "A"; }
  else if (aa == "C") { *aa_str = "CIVQRK"; }
  else if (aa == "D") { *aa_str = "D"; }
  else if (aa == "E") { *aa_str = "E"; }
  else if (aa == "F") { *aa_str = "VIF"; }
  else if (aa == "G") { *aa_str = "G"; }
  else if (aa == "H") { *aa_str = "HKR"; }
  else if (aa == "I") { *aa_str = "VIF"; }
  else if (aa == "K") { *aa_str = "RQK"; }
  else if (aa == "L") { *aa_str = "LM"; }
  else if (aa == "M") { *aa_str = "LM"; }
  else if (aa == "N") { *aa_str = "N"; }
  else if (aa == "P") { *aa_str = "P"; }
  else if (aa == "Q") { *aa_str = "QRK"; }
  else if (aa == "R") { *aa_str = "KQR"; }
  else if (aa == "S") { *aa_str = "S"; }
  else if (aa == "T") { *aa_str = "T"; }
  else if (aa == "V") { *aa_str = "VIF"; }
  else if (aa == "W") { *aa_str = "WYF"; }
  else if (aa == "Y") { *aa_str = "WYF"; }
}
inline void secseq2int(int*idx, char sec) {
  if      (sec == 'A') { *idx = 0; }
  else if (sec == 'O') { *idx = 1; }
  else if (sec == 'Q') { *idx = 2; }
  else if (sec == 'H') { *idx = 3; }
  else if (sec == 'C') { *idx = 4; }
  else if (sec == 'E') { *idx = 5; }
}
inline int basex2dec(int* vect, int n, int baseX) {
  int num = 0;
  int base = 0;
  for (int i = 0; i < n; i++) {
    base = n-i-1;
    num += vect[i]*pow(baseX, base);
  }
  return num;
}

static const char NOT_SET = 0;

/**
 * Amino-acid types.
 */
/*
static const char ALA = 0;
static const char ARG = 1;
static const char ASN = 2;
static const char ASP = 3;
static const char CYS = 4;
static const char GLN = 5;
static const char GLU = 6;
static const char GLY = 7;
static const char HIS = 8;
static const char ILE = 9;
static const char LEU = 10;
static const char LYS = 11;
static const char MET = 12;
static const char PHE = 13;
static const char PRO = 14;
static const char SER = 15;
static const char THR = 16;
static const char TRP = 17;
static const char TYR = 18;
static const char VAL = 19;
static const char UNK = 20;
*/
// comment out the above and uncomment this to use 1-numbered amino acids
static const char ALA =  1;
static const char ARG =  2;
static const char ASN =  3;
static const char ASP =  4;
static const char CYS =  5;
static const char GLN =  6;
static const char GLU =  7;
static const char GLY =  8;
static const char HIS =  9;
static const char ILE = 10;
static const char LEU = 11;
static const char LYS = 12;
static const char MET = 13;
static const char PHE = 14;
static const char PRO = 15;
static const char SER = 16;
static const char THR = 17;
static const char TRP = 18;
static const char TYR = 19;
static const char VAL = 20;
static const char UNK = 21;

/**
 * Atom types in a molecule chain.
 */
static const char MC_N   =  1;
static const char MC_CA  =  2;
static const char MC_C   =  3;
static const char MC_O   =  4;
static const char MC_HN  =  5;
static const char MC_HA1 =  6;
static const char MC_CAP =  7;

static const char SC_CB  =  8;
static const char SC_HA2 =  9;

static const char SC_CG  = 10;
static const char SC_CG1 = 11;
static const char SC_CG2 = 12;
static const char SC_OG  = 13;
static const char SC_OG1 = 14;
static const char SC_SG  = 15;

static const char SC_CD  = 16;
static const char SC_CD1 = 17;
static const char SC_CD2 = 18;
static const char SC_OD  = 19;
static const char SC_OD1 = 20;
static const char SC_OD2 = 21;
static const char SC_ND  = 22;
static const char SC_ND1 = 23;
static const char SC_ND2 = 24;
static const char SC_SD  = 25;

static const char SC_CE  = 26;
static const char SC_CE1 = 27;
static const char SC_CE2 = 28;
static const char SC_CE3 = 29;
static const char SC_OE  = 30;
static const char SC_OE1 = 31;
static const char SC_OE2 = 32;
static const char SC_NE  = 33;
static const char SC_NE1 = 34;
static const char SC_NE2 = 35;

static const char SC_CZ  = 36;
static const char SC_CZ2 = 37;
static const char SC_CZ3 = 38;
static const char SC_NZ  = 39;

static const char SC_CH2 = 40;
static const char SC_OH  = 41;
static const char SC_NH  = 42;
static const char SC_NH1 = 43; 
static const char SC_NH2 = 44;

static const char SC_H   = 45;
static const char SC_X   = 46;

/**
 * Bond types in a protein. For an abuse of notation, bonds are named after the torsional dihedral defined by them.
 */
static const char PHI_BOND       =  1;
static const char PSI_BOND       =  2;
static const char OMEGA_BOND     =  3;
static const char NH_BOND        =  4;
static const char CO_BOND        =  5;
static const char CAH_BOND       =  6;
static const char CHI1_BOND      =  7;
static const char CHI2_BOND      =  8;
static const char CHI3_BOND      =  9;
static const char CHI4_BOND      = 10;
static const char CHI5_BOND      = 11;
static const char UNDEFINED_BOND = 12;

/**
 * Indexes for protein bonds. For an abuse of notation, bonds are named after the torsional dihedral defined by them.
 */
static const int PHI   = 0;
static const int PSI   = 1;
static const int OMEGA = 2;
static const int CHI1  = 3;
static const int CHI2  = 4;
static const int CHI3  = 5;
static const int CHI4  = 6;
static const int CHI5  = 7;



// Rotamer

const float BONDDIST = 2.1;
#define ROT_PROB_CUT_MIN  0.01
#define ROT_PROB_CUT_ACC  0.97

#define WGT_HBOND       1.0
#define WGT_SSBOND      6.0
#define CACB_DIST_CUT   2.35
#define RESI_DIST_CUT   4.25
#define SEC_CUT         15.0
// VDW parameters, 18 atom types in CHARMM19
//                               Radius, depth
static const double VDW[][2] = {{1.78, 0.25}, // C1
                                {1.40, 0.14}, // C2
                                {2.30, 0.30}, // C3
                                {1.91, 0.37}, // C4
                                {1.96, 0.48}, // C5
                                {1.73, 0.38}, // C6
                                {1.43, 0.07}, // C7
                                {1.99, 0.38}, // C8
                                {1.48, 0.22}, // O1
                                {1.44, 0.27}, // O2
                                {1.40, 0.07}, // O3
                                {1.43, 0.12}, // O4
                                {1.42, 0.08}, // N1
                                {1.69, 0.24}, // N2
                                {1.56, 0.46}, // N3
                                {1.70, 0.48}, // N4
                                {2.15, 0.44}, // S1
                                {1.74, 0.40}  // S2
                               };
// Threshold
// VDW
#define DSTAR_MIN_CUT   0.015
#define DSTAR_MAX_CUT   1.90
#define VDW_REP_CUT     10.0
// Hbond energy
#define OPT_HBOND_DIST  2.8
#define MIN_HBOND_DIST  2.6
#define MAX_HBOND_DIST  3.2
#define MIN_HBOND_THETA 90.0
#define MIN_HBOND_PHI   90.0
// SSbond energy
#define OPT_SSBOND_DIST 2.03
#define MIN_SSBOND_DIST 1.73
#define MAX_SSBOND_DIST 2.53
#define OPT_SSBOND_ANGL 105.0
#define MIN_SSBOND_ANGL 75.0
#define MAX_SSBOND_ANGL 135.0

} // namespace leri
#endif  // LERI_PROTEIN_UTIL_H_
