/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

/**
 * This file contains the declaration of basic data structure types to hold molecule information.
 * There is no assumption made here as to the specific type of molecules that will be represented
 * with these structures. There are only two assumptions made with regard of the structure fo the
 * data that will be stored in these structures: 
 * 1) the hierarchy of chain->residue->atom, although chain could be protein, rna, dna, etc. For protein 
 * chains, residues should be amino-acids, nucleic bases for rna/dna, etc.
 * 2) the ordering of the atoms of a branched molecule in the array of AtomData structures. The numbering
 * of the atoms within a given branch should be consecutive, and the atoms of the branches starting at a 
 * given atom should occur after this atom. In order words, if i is the array index of a branching atom,
 * and branches B0, B1, B2, etc, start at b, then the branches should be all the atoms within an index interval
 * [j00, j01] for B0, [j10, j11] for B1 and so on, Furthermore:
 * i, j00 = i + 1, ..., j01, j10 = j01 + 1, ..., j11...
 * In this way, torsional rotations can be defined and implemented in a very convenient way.
 *
 */

#ifndef LERI_DATA_TYPES_H__
#define LERI_DATA_TYPES_H__

#include "leri/common.h"
#include "leri/protein/bioutil.h"

// TODO:  TODO
namespace leri {

// /**
//  * Record of atom data.
//  */
// struct AtomData
// {
//     /**
//             * Atom type (Carbon, Hydrogen, Oxygen, etc).
//             */
//     char type;

//     /**
//             * Pointers to the xyzw coordinates of the atom.
//             */
//     // Dtype *x;
//     // Dtype *y;
//     // Dtype *z;
//     // Dtype *w;

//     Dtype xyzw[4];

//     /**
//             * Residue and chain where this atom resides.
//             */
//     int res;
//     int chain;

//     /**
//             * Arrays to hold custom integer or float properties for each atom.
//             * This arrays are always NULL when the atom record is initialized. 
//             */
//     int niprop;
//     // int *iproperties;
//     // int nfprop;
//     // Dtype *fproperties;
// };

// /**
//  * Record of bond data. A bond is defined by four consecutive atoms in a chain, and 
//  * have a torsional or dihedral angle associated to it. By rotating this angle, the atoms
//  * that occur down the chain starting from index atom3 will change their position 
//  * in space. 
//  */
// struct BondData
// {
//     /**
//             * Bond type.
//             */
//     char type;

//     /**
//             * Indexes of the atoms defining the bond. It should be atom0 < atom1 < atom2 < atom3,
//             * but not necessarily atom1 = atom0 + 1, etc.
//             */
//     int atom0; 
//     int atom1; 
//     int atom2;
//     int atom3;

//     /**
//             * Index of the last atom in the chain that is affected by a torsional rotation at this bond.
//             * The existence of such an index requires the atom ordering described at beginning of
//             * this file, this is, i, j00 = i + 1, ..., j01, j10 = j01 + 1, ..., j11... for branches defined at i.
//             */
//     int lastatom;

//     /**
//             * Current value for the torsional angle of this bond.
//             */
//     Dtype angle;
//     /**
//             * Used to store a new value for the torsional angle.
//             */
//     Dtype newangle;
// };

// /**
//  * Record of residue data.
//  */
// struct ResidueData
// {
//     /**
//             * Residue type. This should represent amino-acids in the case of protein chains, nucleic acids
//             * for rna, etc.
//             */
//     char type;

//     /**
//             * First and last atom of this residue.
//             */
//     int firstatom;
//     int lastatom;

//     /**
//             * Chain containing this residue.
//             */     
//     int chain; 

//     /**
//             * Numer of bonds that start at atoms within this residue.
//             */
//     int nbonds;
//     /**
//             * List of bonds that start at atoms within this residue.
//             */
//     struct BondData *bonds;

//     /**
//             * Arrays to hold custom integer or float properties for each residue.
//             * This arrays are always NULL when the residue record is initialized. 
//             */
//     // int niprop;
//     // int *iproperties;
//     // int nfprop;
//     // Dtype *fproperties;
// };

// /**
//  * Record of chain data.
//  */
// struct ChainData
// {
//     /**
//             * Chain type.
//             */
//     char type;

//     /**
//             * First and last residue of this chain.
//             */
//     int firstres;
//     int lastres;
// };

// // Some useful defines to access the components of an atom in a 1D array where the coordinates are stored one
// // after another.
// // #define xcoord_index(atom)	(atom * 4 + 0)
// // #define ycoord_index(atom)	(atom * 4 + 1)
// // #define zcoord_index(atom)	(atom * 4 + 2)
// // #define wcoord_index(atom)	(atom * 4 + 3)

// // Similarly, a convenience define to access the elements of a 2D  distance matrix stored as a 1D array.
// // #define distance_index(atom0, atom1, natoms)	(atom0 * natoms + atom1)


// Single atom
template <typename Dtype>
class AtomStrut {
public:
  string* residue; //  Residue index (ResidueEnum)
  int* resseq;     //  Resseq index
  string* atom;    //  Atom string
  string* altloc;  //  Altloc string
  int* real_ss;    //  Secondary structure index (real)
  int* pred_ss;    //  Secondary structure index (predicted)
  Dtype* pred_prob_ss_C; //  Secondary structure propensity - C
  Dtype* pred_prob_ss_H; //  Secondary structure propensity - H
  Dtype* pred_prob_ss_E; //  Secondary structure propensity - E
  Dtype* coordinates; //  3D coordinates
  Dtype* occupancy_;
  Dtype* tfactor_;
  void init();

  //  Constructor
  AtomStrut() { init(); }

  //  Copy constructor
  //  \param other Source object from which copy is made.
  AtomStrut(const AtomStrut &other) {
    this->init();
    if (other.residue)
      set_residue(*other.residue);
    if (other.resseq)
      set_resseq(*other.resseq);
    if (other.atom)
      set_atom(*other.atom);
    if (other.altloc)
      set_altloc(*other.altloc);
    if (other.coordinates)
      set_coordinates(other.coordinates[0], other.coordinates[1], other.coordinates[2]);
    if (other.real_ss)
      set_real_ss(*other.real_ss);
    if (other.pred_ss)
      set_pred_ss(*other.pred_ss);
    if (other.pred_prob_ss_C)
      set_pred_prob_ss_C(*other.pred_prob_ss_C);
    if (other.pred_prob_ss_H)
      set_pred_prob_ss_H(*other.pred_prob_ss_H);
    if (other.pred_prob_ss_E)
      set_pred_prob_ss_E(*other.pred_prob_ss_E);
  }

  //  Destructor
  ~AtomStrut();
  //  Set residue value
  void set_residue(string residue) { this->residue = new string(residue); }

  //  Set resseq value
  void set_resseq(int resseq) { this->resseq = new int(resseq); }

  //  Set atom value
  void set_atom(string atom) { this->atom = new string(atom); }

  //  Set altloc value
  //  \param altloc Altloc identification string
  void set_altloc(string altloc) {
    string modified_altloc = altloc;
    if (altloc == string(" ")) { modified_altloc = "_"; }
    this->altloc = new string(modified_altloc);
  }

  //  Set coordinate values
  void set_coordinates(Dtype x, Dtype y, Dtype z) {
    this->coordinates = new Dtype[3];
    this->coordinates[0] = x;
    this->coordinates[1] = y;
    this->coordinates[2] = z;
  }

  //  Set secondary structure value (real)
  //  \param ss Secondary structure index (SecondaryStructureEnum)
  void set_real_ss(int ss) {
    this->real_ss = new int(ss);
  }

  //  Set secondary structure value (predicted)
  //  \param ss Secondary structure index (SecondaryStructureEnum)
  void set_pred_ss(int ss) {
    this->pred_ss = new int(ss);
  }

  //  Set secondary structure C propersity value
  // 
  //  \param ss_prob Secondary structure propensity
  void set_pred_prob_ss_C(Dtype ss_prob) {
    this->pred_prob_ss_C = new Dtype(ss_prob);
  }

  //  Set secondary structure H propensity value
  // 
  //  \param ss_prob Secondary structure propensity
  void set_pred_prob_ss_H(Dtype ss_prob) {
    this->pred_prob_ss_H = new Dtype(ss_prob);
  }

  //  Set secondary structure E propensity value
  // 
  //  \param ss_prob Secondary structure propensity
  void set_pred_prob_ss_E(Dtype ss_prob) {
    this->pred_prob_ss_E = new Dtype(ss_prob);
  }

  void set_occupancy(Dtype occupancy) {
    this->occupancy_ = new Dtype(occupancy);
  }
  void set_tfactor(Dtype tfactor) {
    this->occupancy_ = new Dtype(tfactor);
  }


  
  //  Overload input operator for AtomStrut
  friend ifstream &operator>>(ifstream &in, AtomStrut &data) {
    vector<string> words;
    if (in.good()) {
      string line;
      if (getline(in, line)) {
        boost::trim(line);
        vector<string> words;
        boost::split(words, line, boost::is_any_of("\t "), boost::token_compress_on);
      
        for (unsigned int i=0; i<words.size(); i++) {
          string word = words[i];
          if (word == "RES") {
            data.set_residue(words[++i].data());
          } else if (word == "RESSEQ") {
            data.set_resseq(atoi(words[++i].c_str()));
          } else if (word == "ATOM") {
            data.set_atom(words[++i]);
          } else if (word == "ALTLOC") {
            data.set_altloc(words[++i]);
          } else if (word == "COORD") {
            Dtype x = atof(words[++i].data());
            Dtype y = atof(words[++i].data());
            Dtype z = atof(words[++i].data());
            data.set_coordinates(x, y, z);
          } else if (word == "REALSS") {
            data.set_real_ss(atoi(words[++i].data()));
          } else if (word == "PREDSS") {
            data.set_pred_ss(atoi(words[++i].data()));
          } else if (word == "PREDPROBSS_C") {
            data.set_pred_prob_ss_C(atof(words[++i].data()));
          } else if (word == "PREDPROBSS_H") {
            data.set_pred_prob_ss_H(atof(words[++i].data()));
          } else if (word == "PREDPROBSS_E") {
            data.set_pred_prob_ss_E(atof(words[++i].data()));
          }
        }
      }
    }
    return in;
  }

  //  Overload output operator for AtomStrut
  friend ostream & operator<<(ostream &o, AtomStrut &data) {
    char output[200];
    if (data.residue) {
      sprintf(output, "RES %4s ", data.residue->c_str());
      o << output;
    }

    if (data.resseq) {
      sprintf(output, "RESSEQ %4d ", *data.resseq);
      o << output;
    }

    if (data.atom) {
      sprintf(output, "ATOM %3s ", data.atom->data());
      o << output;
    }

    if (data.altloc) {
      sprintf(output, "ALTLOC %1s ", data.altloc->data());
      o << output;
    }

    if (data.coordinates) {
      sprintf(output, "COORD %20.15f %20.15f %20.15f ",
           data.coordinates[0], data.coordinates[1], data.coordinates[2]);
      o << output;
    }

    if (data.real_ss) {
      sprintf(output, "REALSS %2d ", *data.real_ss);
      o << output;
    }

    if (data.pred_ss) {
      sprintf(output, "PREDSS %2d ", *data.pred_ss);
      o << output;
    }
    
    if (data.pred_prob_ss_C) {
      sprintf(output, "PREDPROBSS_C %20.15f ", *data.pred_prob_ss_C);
      o << output;
    }

    if (data.pred_prob_ss_H) {
      sprintf(output, "PREDPROBSS_H %20.15f ", *data.pred_prob_ss_H);
      o << output;
    }

    if (data.pred_prob_ss_E) {
      sprintf(output, "PREDPROBSS_E %20.15f ", *data.pred_prob_ss_E);
      o << output;
    }
    
    return o;
  }    
};


//  Input data for entire protein 
template <typename Dtype>
class ProteinData {
public:
  //  Internal data container.
  //  data: vector<polypeptide>
  //  polypeptide: vector<residue>
  //  residue: vector<atoms>
  vector<vector<vector<AtomStrut<Dtype>* > > > data;
  //  Constructor (default)
  ProteinData(){};
  //  Constructor
  //  \param filename Pdb input filename
  ProteinData(string filename);
  //  Copy Constructor
  //  \param other Source object from which copy is made.
  // ProteinData(const ProteinData &other) {
  //   for (int i=0; i<other.get_num_polypeptides(); i++) {
  //     data.push_back(vector<vector<AtomStrut<Dtype>* > >());
  //     for (int j=0; j<other.get_num_amino_acids(i); j++) {
  //       data[i].push_back(vector<AtomStrut<Dtype>* >());
  //       for (int k=0; k<other.get_num_atoms(i,j); k++) {
  //         data[i][j].push_back(new AtomStrut<Dtype>(*other.data[i][j][k]));
  //       }
  //     }
  //   }
  // }

  //  Destructor
  ~ProteinData();

  //  Overload boolean evaluation operator
  operator bool() const { return !data.empty(); }

  //  Overload assignment operator
  // 
  //  \param other Source object from which assignment is made.
  //  \return Current object (*this)
  // ProteinData &operator=(const ProteinData &other) {
  //   if (this == &other) { return *this; }
  //   for (int i=0; i<other.get_num_polypeptides(); i++) {
  //     data.push_back(vector<vector<AtomStrut<Dtype>* > >());
  //     for (int j=0; j<other.get_num_amino_acids(i); j++) {
  //       data[i].push_back(vector<AtomStrut<Dtype>* >());
  //       for (int k=0; k<other.get_num_atoms(i,j); k++) {
  //         data[i][j].push_back(new AtomStrut<Dtype>(*other.data[i][j][k]));
  //       }
  //     }
  //   }  
  //   return *this;
  // }
  
  //  Return number of polypeptides
  //  \return Number of polypeptides in data object
  int get_num_polypeptides() const { return data.size(); }

  //  Return number of residues in given polypeptide
  //  \param pp_index Polypeptide index
  //  \return Number of residues in polypeptide
  int get_num_amino_acids(int pp_index) const { return data[pp_index].size(); }

  //  Return number of atoms in given residue
  // 
  //  \param pp_index Polypeptide index
  //  \param res_index Residue index
  //  \return Number of atoms in residue
  int get_num_atoms(int pp_index, int res_index) const { return data[pp_index][res_index].size(); }  
  // Get sequence of amino acids for each polypeptide
  // void get_sequence(vector<vector<char> >& seq) const {
    // bool flag = false;
    // unsigned char aa;
    // for (int i = 0; i < get_num_polypeptides(); i++) {
    //   seq.push_back(vector<char>());
    //   for (int j=0; j< get_num_amino_acids(i); j++) {
    //     // Just use information from first atom
    //     if (data[i][j][0]->residue) {
    //         aa321(&aa, *data[i][j][0]->residue);
    //         seq[i].push_back(aa);
    //       flag = true;
    //     }
    //   }
    // }
    // if (!flag) {
    //   printf( "--  \033[1;31m");
    //   printf( "WARNING: There is not sequence for the PDB file!\n");
    //   printf( "\033[0m\n");
    //   // exit (EXIT_FAILURE);
    // }
  // }
    //  Split polypeptides into smaller pieces if chainbreaks are detected - full atom version
  // 
  //  \param radius Maximum distance between consecutive atoms
  void split_by_chain_breaks_full_atom(Dtype radius = 1.8);


  //  Split polypeptides into smaller pieces if chainbreaks are detected - CA-only version
  // 
  //  \param radius Maximum distance between consecutive atoms
  void split_by_chain_breaks_ca(Dtype radius = 4.3);
  

  //  Split polypeptides into smaller pieces if chainbreaks are detected
  // 
  //  \param CA_only Determines whether to use full atom or CA_only version
  void split_by_chain_breaks(bool CA_only) {
    if (CA_only) {
      split_by_chain_breaks_ca();
    } else {
      split_by_chain_breaks_full_atom();
    }

  }






  //  Create new polypeptide
  void new_polypeptide() { data.push_back(vector<vector<AtomStrut<Dtype>* > >()); }

  //  Create new residue
  // 
  //  \param pp_index Index of polypeptide in which residue should be placed
  void new_residue(int pp_index) {
    while (pp_index >= get_num_polypeptides())
      new_polypeptide();
    data[pp_index].push_back(vector<AtomStrut<Dtype>* >());
  }

  //  Create new atom
  // 
  //  \param pp_index Index of polypeptide in which atom should be placed
  //  \param res_index Index of residue in which atom should be placed
  void new_atom(int pp_index, int res_index) {
    while (pp_index >= get_num_polypeptides()) {
      new_polypeptide();
    }
    while (res_index >= get_num_amino_acids(pp_index)) {
      new_residue(pp_index);
    }
    data[pp_index][res_index].push_back(new AtomStrut<Dtype>());
  }

  //  Return last added atom
  // 
  //  \return Pointer to atom
  AtomStrut<Dtype>* current_atom() {
    int lastPP = data.size()-1;
    int lastRes = data[lastPP].size()-1;
    int lastAtom = data[lastPP][lastRes].size()-1;
    return data[lastPP][lastRes][lastAtom];
  }



  //  Return chain size for each polypeptide
  // 
  //  \return Integer for each polypeptide
  // vector<int> size(){
  //   vector<int> v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(get_num_amino_acids(i));
  //   }
  //   return v;
  // }
  



  //  Return position sequences for each polypeptide
  // 
  //  \return vector of 3D coordinates for each polypeptide
  // vector<vector<vector<Vector_3D> > > get_positions() const {
  //   bool status = false;
  //   vector<vector<vector<Vector_3D> > > v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(vector<vector<Vector_3D> >());
  //     for (int j=0; j<get_num_amino_acids(i); j++) {
  //       v[i].push_back(vector<Vector_3D>());
  //       for (int k=0; k<get_num_atoms(i,j); k++) {       
  //         v[i][j].push_back(Vector_3D(data[i][j][k]->coordinates));
  //         status=true;
  //       }
  //     }
  //   }
  //   if (!status)
  //     fprintf(stderr, "Warning: No position data found\n");
    
  //   return v;
  // }


  //  Return (phi, psi) sequences for each polypeptide
  // 
  //  \param include_omega Whether omega angle should be included
  //  \return vector of phi,psi pairs for each polypeptide
  // vector<vector<vector<Dtype> > > get_phi_psi(bool include_omega=false) const {
  //   bool status = false;

  //   vector<vector<vector<Vector_3D> > > positions = this->get_positions();

  //   vector<vector<vector <Dtype> > > v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(vector<vector<Dtype> >());
  //     for (int j=0; j<get_num_amino_acids(i); j++) {
  //       v[i].push_back(vector<Dtype>());

  //       Dtype phi;
  //       if (j==0) {
  //         phi = numeric_limits<Dtype>::quiet_NaN();

  //         // Set first phi based on terminal Hydrogen position is available
  //         for (unsigned int k=0; k<positions[i][j].size(); ++k) {
  //           if ((*data[i][j][k]->atom) == "H" || (*data[i][j][k]->atom) == "H1" || (*data[i][j][k]->atom) == "1H") {
  //             phi = calc_dihedral(positions[i][j][k], positions[i][j][0], positions[i][j][1], positions[i][j][2]);
  //             break;
  //           }
  //         }
  //       } else {
  //         phi = calc_dihedral(positions[i][j-1][2], positions[i][j][0], positions[i][j][1], positions[i][j][2]);
  //       }
        
  //       Dtype psi;
  //       if (j==get_num_amino_acids(i)-1) {
  //         psi = numeric_limits<Dtype>::quiet_NaN();

  //         // Set first phi based on terminal Hydrogen position is available
  //         for (unsigned int k=0; k<positions[i][j].size(); ++k) {
  //           if ((*data[i][j][k]->atom) == "O" || (*data[i][j][k]->atom) == "OXT") {
  //             psi = calc_dihedral(positions[i][j][0], positions[i][j][1], positions[i][j][2], positions[i][j][k]);
  //             break;
  //           }
  //         }

  //       } else {
  //         psi = calc_dihedral(positions[i][j][0], positions[i][j][1], positions[i][j][2], positions[i][j+1][0]);
  //       }

  //       Dtype omega;
  //       if (include_omega) {
  //         if (j==0) {
  //           omega = numeric_limits<Dtype>::quiet_NaN();
  //         } else {
  //           omega = calc_dihedral(positions[i][j-1][1], positions[i][j-1][2], positions[i][j][0], positions[i][j][1]);
  //         }
  //       }
        
  //       v[i][j].push_back(phi);
  //       v[i][j].push_back(psi);
  //       if (include_omega) {
  //         v[i][j].push_back(omega);
  //       }
        
  //       status=true;
  //     }
  //   }
  //   if (!status)
  //     fprintf(stderr, "Warning: No position data found\n");
    
  //   return v;
  // }
  

  //  Return sequence of omega angles for each polypeptide
  // 
  //  \return vector of omega angles for each polypeptide
  // vector<vector<Dtype> > get_omega() const {
  //   bool status = false;
    
  //   vector<vector<vector <Dtype> > > angles = get_phi_psi(true);

  //   vector<vector<Dtype> > v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(vector<Dtype>());
  //     for (int j=0; j<get_num_amino_acids(i); j++) {
  //       Dtype omega = angles[i][j][2];
  //       v[i].push_back(omega);
  //       status = true;
  //     }
  //   }
  //   if (!status)
  //     fprintf(stderr, "Warning: No cis data found\n");
    
  //   return v;
  // }
  

  //  Return sequence of cis/trans states for each polypeptide
  // 
  //  \return vector of cis/trans states for each polypeptide
  // vector<vector<int> > get_cis() const {
  //   bool status = false;
    
  //   vector<vector<vector <Dtype> > > angles = get_phi_psi(true);

  //   vector<vector<int> > v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(vector<int>());
  //     for (int j=0; j<get_num_amino_acids(i); j++) {
  //       Dtype omega = angles[i][j][2];
  //       int cis = 0;
  //       if (isnan(omega)) {
  //         set_initial_value(cis);
  //       } else if (fabs(omega) < M_PI/4) {
  //         cis = 1;
  //       }
  //       v[i].push_back(cis);
  //       status = true;
  //     }
  //   }
  //   if (!status)
  //     fprintf(stderr, "Warning: No cis data found\n");
    
  //   return v;
  // }
  
  //  Return (theta, tau) sequences for each polypeptide
  // 
  //  \return vector of theta,tau pairs for each polypeptide
  // vector<vector<vector <Dtype> > > get_theta_tau() const {
  //   bool status = false;

  //   vector<vector<vector<Vector_3D> > > positions = this->get_positions();

  //   vector<vector<vector <Dtype> > > v;
  //   for (int i=0; i<get_num_polypeptides(); i++) {
  //     v.push_back(vector<vector<Dtype> >());
  //     for (int j=0; j<get_num_amino_acids(i); j++) {
  //       v[i].push_back(vector<Dtype>());

  //       Dtype theta;
  //       if ((j==0) || (j==get_num_amino_acids(i)-1)) {
  //         theta = numeric_limits<Dtype>::quiet_NaN();
  //       } else {
  //         theta = calc_angle(positions[i][j-1][1], positions[i][j][1], positions[i][j+1][1]);
  //       }
        
  //       Dtype tau;
  //       if ((j==0) || (j==1) || j==get_num_amino_acids(i)-1) {
  //         tau = numeric_limits<Dtype>::quiet_NaN();
  //       } else {
  //         tau = calc_dihedral(positions[i][j-2][1], positions[i][j-1][1], positions[i][j][1], positions[i][j+1][1]);
  //       }

  //       v[i][j].push_back(theta);
  //       v[i][j].push_back(tau);
        
  //       status=true;
  //     }
  //   }
  //   if (!status)
  //     fprintf(stderr, "Warning: No position data found\n");
    
  //   return v;
  // }

  //  Get resseq sequence for each polypeptide
  // 
  //  \return Vector of resseq vectors
  vector<vector<int> > get_resseq() const {
    vector<vector<int> > v;
    for (int i=0; i<get_num_polypeptides(); i++) {
      v.push_back(vector<int>());
      for (int j=0; j<get_num_amino_acids(i); j++) {
        v.back().push_back(*data[i][j][0]->resseq);
      }
    }
    return v;
  }
  
  
  //  Get first resseq of each polypeptide
  // 
  //  \return Vector of resseq start indices
  vector<int> get_resseq_offset() const {
    vector<int> v;
    for (int i=0; i<get_num_polypeptides(); i++) {
      if (data[i][0][0]->resseq) {
        v.push_back(*data[i][0][0]->resseq);
      }
    }
    return v;
  }

  
  
  //  Return real secondary structure sequences for each polypeptide
  // 
  //  \return Vector of secondary structure indices for each polypeptide
  vector<vector<int> > get_real_ss() const {
    bool status = false;
    vector<vector<int> > v;
    for (int i=0; i<get_num_polypeptides(); i++) {
      v.push_back(vector<int>());
      for (int j=0; j<get_num_amino_acids(i); j++) {
        // Just use information from first atom
        if (data[i][j][0]->real_ss) {
          v[i].push_back(*data[i][j][0]->real_ss);
          status=true;
        }
      }
    }
    if (!status)
      fprintf(stderr, "Warning: No realSS data found\n");
      
    return v;
  }

  //  Return predicted secondary sequences for each polypeptide
  // 
  //  \return Vector of secondary structure indices for each polypeptide
  vector<vector<int> > get_pred_ss() const {
    bool status = false;
    vector<vector<int> > v;
    for (int i=0; i<get_num_polypeptides(); i++) {
      v.push_back(vector<int>());
      for (int j=0; j<get_num_amino_acids(i); j++) {
        // Just use information from first atom
        if (data[i][j][0]->pred_ss) {
          v[i].push_back(*data[i][j][0]->pred_ss);
          status=true;
        }
      }
    }
    if (!status)
      fprintf(stderr, "Warning: No predSS data found\n");

    return v;
  }

  //  Return predicted secondary structure propensity sequences for each polypeptide
  // 
  //  \return vector of H,E,C propensities for each polypeptide
  vector<vector<vector<Dtype> > > get_pred_prob_ss() const {
    bool status = false;
    vector<vector<vector<Dtype> > > v;
    for (int i=0; i<get_num_polypeptides(); i++) {
      v.push_back(vector<vector<Dtype> >());
      for (int j=0; j<get_num_amino_acids(i); j++) {
        vector<Dtype> probs;
        // Just use information from first atom
        if (data[i][j][0]->pred_prob_ss_C &&
            data[i][j][0]->pred_prob_ss_H &&
            data[i][j][0]->pred_prob_ss_E ) {

          probs.push_back(*data[i][j][0]->pred_prob_ss_H);
          probs.push_back(*data[i][j][0]->pred_prob_ss_E);
          probs.push_back(*data[i][j][0]->pred_prob_ss_C);
          status=true;
        }
        v[i].push_back(probs);
      }
    }

    if (!status)
      fprintf(stderr, "Warning: No predProbSS data found\n");
    return v;
  }

  //  Overload [] indexing operator for ProteinData
  // 
  //  \param pp_index polypeptide index
  //  \return Vector(residues) of vector of AtomStrut
  vector<vector<AtomStrut<Dtype>* > > operator[](int pp_index) {
    return data[pp_index];
  }

  //  Save data to file
  // 
  //  \param filename Destination filename
  void save(string filename) {
    ofstream out(filename.data());
    if (!out.good()) {
      cerr << "Couldn't open file " << filename.data() << "\n";
      return;
    }
    out << *this;
    out.close();
  }

  //  Load data from file
  // 
  //  \param filename Source filename
  void load(string filename) {
    ifstream in(filename.data());
    if (!in.good()) {
      cerr << "Couldn't open file " << filename.data() << "\n";
      return;
    }
    string pp_index_label;
    string pp_string;
    string res_index_label;
    string res_string;
    while (in >> pp_index_label >> pp_string >> res_index_label >> res_string) {
      int pp_index;
      stringstream(pp_string) >> pp_index;
      int res_index;
      stringstream(res_string) >> res_index;
      new_atom(pp_index, res_index);
      in >> *current_atom();
    }
  }

  //  Overload output operator for ProteinData
  friend ostream & operator<<(ostream &o, ProteinData &data) {
    for (int i=0; i<data.get_num_polypeptides(); i++) {
      for (int j=0; j<data.get_num_amino_acids(i); j++) {
        for (int k=0; k<data.get_num_atoms(i,j); k++) {
          // o << "PPINDEX " << setw(2) <<  i << " " << "RESINDEX " << setw(3) << j << " " << *data[i][j][k] << "\n";
        }
      }
    }    
    return o;
}

};

} // namespace leri
#endif  // LERI_DATA_TYPES_H__
