/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_ROTAMER_H_
#define LERI_ROTAMER_H_

#include "leri/common.h"
#include "leri/util/strutil.h"
#include "leri/util/fileutil.h"
#include "leri/util/malloc.h"
#include "leri/util/matrix.h"
#include "leri/util/math.h"
#include "leri/protein/bioutil.h"
#include "leri/protein/aminoacid.h"
#include "leri/graph/graph.h"
#include "leri/graph/bag.h"
#include "leri/graph/tree.h"

namespace leri {

template <typename Dtype>
class Rotamer {
public:
  Rotamer(int n, string s, string output_dirt = "", string job_name = "")
  :output_dirt_(output_dirt),
   job_name_(job_name),
   num_of_amino_acids_(n),
   seq_(s) {
    base_name_ = output_dirt_ + job_name_;
   }
  ~Rotamer() {}
  
  void init();
  void actuate(vector<AminoAcid<Dtype> >& aa);
  void finish();

  void set_quiet(bool q = false) { quiet_ = q; }
private:
  void build_sidechain();
  void calc_site_energy();
  void calc_couple_energy();
  void optimization();

  void set_phi_psi(vector<AminoAcid<Dtype> > aa);
  void load_rotamer_lib(); // 2010
  void set_sidechain();
  void set_backbone(int pos);
  void calc_sidechain(int pos, int rot, vector<vector<Dtype> >& rxyz);

  void set_vdw();
  void calc_aa_contact();

  Dtype get_vdw_dist(int a, int b, Dtype rij, Dtype dist);
  Dtype get_atom_vdw_energy(Dtype ddash, Dtype eij);

  void calc_energy_backbond_sidechain(int pos, vector<Dtype>& eval);
  void calc_energy_rotamer_and_fixed_sidechain(int pos, vector<Dtype>& eval);
  Dtype calc_energy_polar_sidechain(int pos1, int rot1, int pos2, int rot2);
  Dtype calc_energy_atom_hbond(Dtype* DBxyz, Dtype* Dxyz, Dtype* Axyz, Dtype* ABxyz, Dtype Dangle, Dtype Aangle);
  Dtype calc_atom_ssbond_energy(Dtype* CA1xyz, Dtype* CB1xyz, Dtype* SG1xyz, Dtype* SG2xyz, Dtype* CB2xyz, Dtype* CA2xyz);
  Dtype calc_energy_polar_sidechain_backbone(int pos1, int rot1, int pos2);
  Dtype calc_energy_favor_rotamer(int pos, int rot);
  bool calc_energy_rotamer_sidechain_x2(int pos1, int pos2, Dtype**& tab);

  void calc_adj_mtx(int nunfix, vector<vector<int> >& mat_adj);
  void search_subgraph_dfs(Graph& graph, int u, vector<int>& visited, vector<vector<int> >& mat_adj, vector<vector<int> >& mat_mask, stack<int>& vertices);
  void build_subgraphs(int nunfix, vector<int>& visited, vector<vector<int> >& mat_adj, vector<vector<int> >& mat_mask);
  void graph_edge_decomposition(vector<vector<int> >& mat_adj, Dtype threshold);
  // void show_graphs();

  void check_subset(vector<int>& leaf_site_left, vector<int>& parent_site_total, vector<int>& idx);
  void bag_deploy_sites(Bag<Dtype>& bag_leaf);
  void root_bag_find_gmec(Bag<Dtype>& bag_root);
  void calc_energy_leaf_bag(Bag<Dtype>& bag_leaf, vector<vector<int> >& Rlcom);
  void get_left_bag_rotamer_combination(Bag<Dtype>& bag_leaf,int depth, vector<int>& Rtmp, vector<vector<int> >& Rlcom);
  void merge_child2parent_bag(Bag<Dtype>& bag_leaf, Bag<Dtype>& parbag, vector<vector<int> >& Rclcom);
  void calc_energy_tree_decomposition_btm2top();
  void calc_energy_left_bag_rotamer_combination(Bag<Dtype>& bag_root, int depth, Dtype& Etmp, vector<int>& Rtmp, vector<Dtype>& Elcom, vector<vector<int> >& Rlcom);
  void calc_energy_right_bag_rotamer_combination(Bag<Dtype>& bag_leaf, int depth, Dtype& Etmp, vector<int>& Rtmp, vector<Dtype>& Ercom, vector<vector<int> >& Rrcom);
  void tree_decomposition_top2btm_rotamer(Bag<Dtype>& parbag, Bag<Dtype>& childbag);
  void free_tree_decomposition();

  void choose_sidechain(int pos, int rot);
  bool actuate_DEE_generalization(vector<int>& pos, int cf);
  int actuate_DEE_goldstein(vector<int>& pos);
  int actuate_DEE_split(vector<int>& pos);
  bool actuate_DEE(vector<int>& pos);

private:
  void create_coord(int n1, int n2);
  void free_coord(int n1, int n2);
  void copy_coord(int n1, int n2, int idx1, int idx2, vector<int> p, bool mc = 0);

private:
  string output_dirt_;    // output directory
  string job_name_;       // job name
  string base_name_;      // basic name

  bool quiet_;
  
  int num_of_amino_acids_;
  string seq_;
  vector<Dtype> phi_;
  vector<Dtype> psi_;
  vector<AminoAcid<Dtype> > amino_acids_;
  map<char, Dtype> wgt_rotamer_;
  map<char, AminoAcid<Dtype> > AA_SIDECHAINS_;

  vector<vector<vector<Dtype> > > chi_;
  vector<vector<Dtype> > prob_rotamer_;
  vector<Dtype> max_prob_;
  vector<int> num_rotamers_;
  vector<int> sub_stat_;
  vector<vector<vector<vector<Dtype> > > > side_chains_; // TODO: 
  // Dtype**** sidechains_;

  vector<int> unfixres_;
  vector<int> rotamer_best_so_far_;
  vector<vector<int> > contact_map_;
  vector<vector<int> > idx_atom_;
  vector<vector<Dtype> > radius_;
  vector<vector<Dtype> > depth_;

  vector<vector<Dtype> > e_site_mtx_; // TODO:
  Dtype**** e_cupl_mtx_;
  
  vector< Graph > graphs_;
  Tree<Dtype> tree_;

  Dtype** coord_1_;
  Dtype** coord_2_;
protected:
  DISABLE_COPY_AND_ASSIGN(Rotamer);
}; // class Rotamer


} // namespace leri

#endif // LERI_ROTAMER_H_