/**
 * The C++ Package of LERI (Learning Engine to Recognize Life).
 * Copyright 2015-present (C) Leri Ltd. All Rights Reserved.
 * Contact: Dr Yaan J. Jang (yaan.jang 'AT' gmail.com)
 *
 */

#ifndef LERI_SEQUENCE_H_
#define LERI_SEQUENCE_H_

#include "leri/common.h"
#include "leri/util/mt19937ar.h"
#include "leri/util/math.h"
#include "leri/util/malloc.h"
#include "leri/util/fileutil.h"
#include "leri/gio/json.h"
#include "leri/gio/pyplot.h"
#include "leri/gio/javas.h"
#include "leri/gio/figure.h"
#include "leri/gio/gnuplot.h"
#include "leri/gio/pymol.h"
#include "leri/gio/ngl.h"
#include "leri/gio/rscript.h"
#include "leri/protein/ribosome.h"
#include "leri/protein/bioutil.h"
#include "leri/protein/atom.h"
#include "leri/protein/aminoacid.h"
#include "leri/protein/chain.h"
#include "leri/protein/protein.h"
#include "leri/solvers/lbfgs.h"

#include "leri/stats/tsne.h"

namespace leri {

#define xHi(i, Ai)             xvar[i + num_site * (Ai)]
#define xEij(i, j, Ai, Aj)     xvar[num_site * NUM_AA + (i < j ? (((j) * (j - 1)/2 + i) * NUM_AA * NUM_AA + (Aj) * NUM_AA + Ai) : (((i)*(i - 1)/2 + j) * NUM_AA * NUM_AA + (Ai) * NUM_AA + Aj))]
#define dEij(i, j, Ai, Aj)     gvar[num_site * NUM_AA + (i < j ? (((j) * (j - 1)/2 + i) * NUM_AA * NUM_AA + (Aj) * NUM_AA + Ai) : (((i)*(i - 1)/2 + j) * NUM_AA * NUM_AA + (Ai) * NUM_AA + Aj))]

#define siteE(j, ai, aj)       Xi[ai + NUM_AA * (aj + NUM_AA * (j))]
#define siteDH(i, a)           Di[a  + NUM_AA * (a  + NUM_AA * (i))]
#define siteDE(j, ai, aj)      Di[ai + NUM_AA * (aj + NUM_AA * (j))]

/**
 * Protein Sequence Class
 */ 
template <typename Dtype>
class LSequence {
public:
  LSequence(string output_dirt = "", string job_name = "", int offset = 0)
  :output_dirt_(output_dirt),
   job_name_(job_name) {
   // pos_offset_(offset+1) {
    if (offset == 0) { pos_offset_ = 1; }
    else { pos_offset_ = offset; }
    base_name_ = output_dirt_ + job_name_;
    max_iter_    = 2e3;
    temperature_ = 1.0;
    ratio_       = 0.99;
  }
  ~LSequence() {}
  void set_job_info(string output_dirt, string job_name);
  void set_seed(MT19937<Dtype>* rand_gen);
  void set_max_iter(int n) { max_iter_ = n; }
  void set_gap_frac(Dtype v) { gap_frac_ = v; }
  void set_site_gap_frac(Dtype v) { site_gap_frac_ = v; }
  void set_temperature(Dtype t) { temperature_ = t; }
  void set_threshold(Dtype t) { threshold_ = t; }
  void set_idx_mutant(string s);

  int get_num_of_seq()  { return num_seqs_; }
  int get_num_of_aa() { return num_amino_acids_; }
  int get_num_of_site() { return num_site_; }
  string get_ref_seq() { return sequence_; }
  int get_num_of_amino_acids() { return num_amino_acids_; };

  void init(string pdb = "");
  void read_fasta(const string fasta_file);
  void fasta_large2small(bool combine = false);
  void jackhmmer2fasta(const string file);
  void blast2fasta(const string file);
  void a3m_to_a2m(const string msa_a3m);
  void sequence_mix(const string msa_file_1, const string msa_file_2);
  // void sequence_trim(const string msa_file, Dtype gap_frac = 0.3, Dtype site_gap_frac = 0.5, bool site_conserved = false);
  void sequence_trim(const string msa_file, string pdb_file = "", bool tlog = false, bool site_conserved = false);
  void calc_sequence_similarity(const string msa_file, int num_scale = 1e3);
  void calc_site_conservation();
  void sequence_reweight(Dtype theta = 0.2, Dtype scale = 1.0);

  void calc_couplings(const string msa_file, bool rc = true, string pdb_file = "", string mat = "", bool conserved = false);  // Evolutionary coupling analysis from FASTA or PDB

  void calc_sequences_energy(const string msa_file, const string potential_mtx);
  void calc_multimutant_energy(const string fasta_file, const string potential_mtx);
  void selectivity_sector_analysis(const string msa_file, Dtype gap_frac = 0.2, Dtype site_gap_frac = 0.3, bool site_conserved = false);

  void sequence_design(const string wt_fasta, const string potential_mtx, int ncupl = 1);
  void point_mutation(const string wt_fasta, const string potential_mtx);
  void residue_contact(const string msa_file, Dtype threshold = 0.1);
  void calc_aa_contact(const string msa_file = "", Dtype threshold = 0);

  void fasta2pdb(string sequence, string cfg_phsi = "", bool random_angles = false, string pdb_name = "L.pdb");
  void ribosome2coord(Chain<Dtype>* chain, vector<string> ribarray);
  /* Protein folding */
  void sequence2sec(string str = "C", string prev = "");
  void sequence2rama(string secseq_prev = "");
  void set_moveset(int n);
  /* Protein folding */
  void add_hn(Chain<Dtype>* chain); // add H, TODO: remove, as the same one is located in chain.h/cc
  void ribosome2aalist(vector<string>& amino_acids_str, vector<string> ribarray, bool random_angles = false);

  void finish();
  void finish_gpu();  


  /* FASTA to distance potential */
  // void fasta2distp() {
  //   string s = "ACDEFG";
  //   int n = s.length();
  //   for (int i = 1; i < n-1; i++) {
  //     for (int j = 0; j < i; j++) {
  //       for (int k = i+1; k < n; k++) {
  //         LOG(INFO)<<s[j]<<s[i]<<s[k];
  //         LOG(INFO)<<j-i<<":"<<k-i;
  //       }
  //     }
  //   }
  // }



  // TODO: delete
  // void sequence_site_conservation_trim(const string& aln, Dtype threshold); // TODO merge with sequence_trim
  // void calc_site_conserved_couplings(const string msa_file, string file_name, Dtype threshold = 0.75); // TODO: merge with calc_evolutionary_couplings

private:
  void create_malloc(); // TODO: create malloc for all variables
  void read_msa(const string msa_file, bool tlog = false);
  void sequence_pairwise_alignment(); // align two sequences
  void sequence_filter_site_gaps(Dtype threshold);
  void sequence_filter_seqs(Dtype threshold);
  void is_sequence_valid();
  void sequence_reweight_gpu(Dtype theta = 0.2, Dtype scale = 1.0);
  // void sequence_reweight(Dtype theta = 0.2, Dtype scale = 1.0);
  void calc_aa_prob_from_msa();
  void calc_aa_prob_from_msa_gpu();
  void sequence_entropy(Dtype& dglo, vector<Dtype> frequency, vector<Dtype> prob_ref, Dtype s = 1.0);
  void optimize_coupling_parameters();
  // Pseudo-likelihood maximization 
  static Dtype plm_negative_log_posterior(const int n, const Dtype* xvar, Dtype* gvar, void *instance);
  static Dtype plm_negative_log_posterior_gpu(const int n, const Dtype* xvar, Dtype* gvar, void *instance);
  void calc_residue_communities(const string pdb_file = "", bool flag = true);
  void calc_evolutionary_couplings(string file_name = "", bool wflag = true, bool site_conserved = false);
  
  void get_aa_idx_in_pdb(string pdb);
  // void sequence_logo();
  void index2sequence(string& _sequence, vector<int>& _index);
  void sequence2index(vector<int>& index, string& sequence);
  void create_point_sites_mtx_();
  void random_sample_position(int* pos, MT19937<Dtype>* rand_gen, int length);
  void random_sample_state(int* _state, MT19937<Dtype>* rand_gen, vector<int>& curpos);
  void calc_mutant_energy(Dtype* energy, int pos,int state, vector<int>& index);
  void calc_energy(Dtype* energy, Dtype* energy_one, Dtype* energy_two, vector<int> &conformation);
  int  actuate_DEE_condition_single(int i, int i_aa, int j_aa);
  int  actuate_DEE_condition();
  void get_minimal_value(Dtype* min_val, int i_pos, int j_pos, int i_aa, int j_aa);
  int  monte_carlo_simulation(vector<int>& conformation, Dtype& energy_change, MT19937<Dtype>* rand_gen);
  // I/O
  void sequence_write(string file_name);
  void sequence_potential_write();
  /* FASTA to PDB */
  Atom<Dtype> build_atom_from_ribosome(string line);
  void normalize_indices(Chain<Dtype>* chain);
  // void internal2cartesian(Dtype* coord, Dtype* p1, Dtype dist, Dtype* p2, Dtype angle, Dtype* p3, Dtype torsion);
  void write_chain2pdb(Chain<Dtype> chain, string pdb_name = "L.pdb");
  void build_peptide(Chain<Dtype>* chain);
  void make_chain_from_ribosome(Chain<Dtype>* chain, vector<string> ribarray, Chain<Dtype> chain_db);
  void read_parameters_from_ribosome(Chain<Dtype>* chain);
  /* FASTA to PDB (end) */ 

// private:
  // Utils
  // static string read_lines(const string str, void* instance);

private:
  string output_dirt_;    // output directory
  string job_name_;       // job name
  string base_name_;      // basic name
  string fasta_name_;
  // string id_name_;
  string pdb_;

  int max_iter_;
  int num_amino_acids_;
  int num_seqs_;          // Number of sequences
  int num_seqs_trim_;     // Number of sequences after trimmed
  int num_seqs_valid_;    // Number of valid sequences
  int num_site_;          // Number of sites (columns) in an MSA
  int num_site_trim_;     // Number of sites after trimmed
  int num_vald_mutant_;   // Number of valid mutants
  int num_aa_pairs_trim_;
  int num_param_;
  int pos_offset_; 
  vector<int> aa_idx_;

  string sequence_;
  string sequence_trim_;
  string sequence_sec_;
  vector<string> seqs_nam_;
  vector<string> seqs_aln_;
  // vector<string> aa_wt_;
  string aa_wt_;
  // vector<string> aa_mt_;
  string aa_mt_;

  vector<vector<int> > point_sites_mtx_;
  vector<int> idx_seqs_trim_;
  vector<int> idx_site_;
  vector<int> idx_site_trim_;
  int* idx_site_trim_real_;
  int* aa_pair_idx_pdb_;
  int** msa_;
  int** msa_trim_;
  int** aa_pair_idx_trim_;
  int** point_site_mask_;

  Dtype gap_frac_;
  Dtype site_gap_frac_;
  Dtype num_eff_;
  Dtype temperature_;
  Dtype ratio_;
  Dtype threshold_;
  Dtype lambda_h_;            // regularization parameters
  Dtype lambda_e_;
  // Dtype* lambdas_;
  Dtype* xvar_;
  Dtype* wgt_;
  Dtype* entropy_min_max_;
  Dtype** relative_entropy_;
  Dtype** entropy_matrix_;
  Dtype** coupling_mtx_;      // coupling between each pair amino acids
  Dtype** epistasis_mtx_;     // reduced matrix of couplings 
  Dtype* freq_site_;          // first-order, P_i(A_i)
  Dtype* freq_cupl_;          // second-order, P_{ij}(A_i, A_j)
  Dtype**   energy_site_mtx_;
  Dtype**** energy_cupl_mtx_;
  bool ec_show_;

  // FASTA to STRUCTURE
  int allh_;
  Dtype** phsi_;
  vector<string> amino_acids_;
  // Plotting
  int*    plot_info_; // size of data for xaxis, yaxis, legends, number of specific data points
  string* plot_ttxy_; // type (heatmap ...), title, xlabel, ylabel
  string* plot_legd_; // legends
  string* plot_xtic_; // xtickers
  string* plot_ytic_; // ytickers
  Dtype*  plot_marg_; // margin: width, height, top, right, bottom, left, rotation angle (x), rotation angle (y)
  Dtype*  plot_data_; // specific data points
  string* plot_text_; // texts for the specific data points
protected:
  DISABLE_COPY_AND_ASSIGN(LSequence);
}; // class LSequence

} // namespace leri
#endif // LERI_SEQUENCE_H_

